package com.bocloud.tse.booter.config;

import java.util.Map;

import com.bocloud.tse.scheduler.worker.listener.TaskResultListener;
import org.springframework.amqp.core.AcknowledgeMode;
import org.springframework.amqp.core.Binding;
import org.springframework.amqp.core.BindingBuilder;
import org.springframework.amqp.core.DirectExchange;
import org.springframework.amqp.core.Queue;
import org.springframework.amqp.core.TopicExchange;
import org.springframework.amqp.rabbit.connection.ConnectionFactory;
import org.springframework.amqp.rabbit.core.RabbitAdmin;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.amqp.rabbit.listener.SimpleMessageListenerContainer;
import org.springframework.amqp.support.converter.Jackson2JsonMessageConverter;
import org.springframework.amqp.support.converter.MessageConverter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Scope;

import com.bocloud.tse.scheduler.listener.TaskActionListener;
import com.megatron.common.utils.MapTools;
import com.megatron.framework.core.CurrentService;

/**
 * rabbitmq相关配置
 */
@Configuration
@ConfigurationProperties(prefix = "spring.rabbitmq")
public class RabbitMQConfiguration {

    @Autowired
    private ConnectionFactory connectionFactory;
    private Map<String, Object> arguments = MapTools.simpleMap("x-ha-policy", "all");

    /**
     * 声明任务动作队列
     */
    @Bean
    public Queue declareTaskActionQueue(CurrentService service) {
        return new Queue("task.action.queue.tse." + service.getService().getHost() + "_" + service.getService().getPort(), true, false, true, arguments);
    }

    /**
     * 定义direct exchange
     */
    @Bean
    public DirectExchange defineDirectExchange() {
        return new DirectExchange("bocloud.direct.exchange", true, true, arguments);
    }

    @Bean
    public Queue declareTaskResultQueue(CurrentService service) {
        return new Queue("task.result.queue." + service.getService().getHost() + "_" + service.getService().getPort(),
                true, false, true, arguments);
    }

    /**
     * 定义topic exchange
     */
    @Bean
    public TopicExchange defineTopicExchange() {
        return new TopicExchange("bocloud.topic.exchange", true, true, arguments);
    }

    /**
     * 任务动作队列绑定bocloud.topic.exchange
     */
    @Bean
    public Binding bindingTaskActionQueue(CurrentService service) {
        return BindingBuilder.bind(declareTaskActionQueue(service)).to(defineTopicExchange())
            .with("topic.task.action.#");
    }

    /**
     * 绑定任务执行结果队列
     *
     * @param service
     * @return
     */
    @Bean
    public Binding bindingTaskResultQueue(CurrentService service) {
        return BindingBuilder.bind(declareTaskResultQueue(service)).to(defineDirectExchange()).withQueueName();
    }
    @Bean
    public SimpleMessageListenerContainer autoActionListener(CurrentService service,
        TaskActionListener taskActionListener) {
        SimpleMessageListenerContainer container = new SimpleMessageListenerContainer(connectionFactory);
        container.setQueues(declareTaskActionQueue(service));
        container.setAcknowledgeMode(AcknowledgeMode.AUTO); // 设置确认模式自动确认
        container.setMessageListener(taskActionListener);
        return container;
    }

    @Bean
    public SimpleMessageListenerContainer autoTaskResultListener(CurrentService service,
                                                                 TaskResultListener taskResultListener) {
        SimpleMessageListenerContainer container = new SimpleMessageListenerContainer(connectionFactory);
        container.setQueues(declareTaskResultQueue(service));
        container.setAcknowledgeMode(AcknowledgeMode.AUTO); // 设置确认模式自动确认
        container.setMessageListener(taskResultListener);
        return container;
    }
    @Bean
    public MessageConverter messageConverter() {
        return new Jackson2JsonMessageConverter();
    }

    @Bean
    @Scope(ConfigurableBeanFactory.SCOPE_PROTOTYPE)
    public RabbitTemplate amqpTemplate() {
        RabbitTemplate amqpTemplate = new RabbitTemplate(connectionFactory);
        amqpTemplate.setMessageConverter(messageConverter());
        amqpTemplate.setExchange(defineDirectExchange().getName());
        return amqpTemplate;
    }

    @Bean
    @Scope(ConfigurableBeanFactory.SCOPE_PROTOTYPE)
    public RabbitAdmin rabbitAdmin() {
        RabbitAdmin rabbitAdmin = new RabbitAdmin(connectionFactory);
        return rabbitAdmin;
    }

}
