package com.bocloud.tse.core.utils;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import org.springframework.scheduling.support.CronExpression;
import org.springframework.util.StringUtils;

import com.google.common.collect.Lists;
import com.megatron.common.model.GeneralResult;

import lombok.extern.slf4j.Slf4j;

@Slf4j
public class CronUtils {

    /**
     * 校验cron表达式是否合法
     */
    public static GeneralResult<Void> isValid(String expression) {
        String message = "保存任务失败：Cron表达式有误或者执行时间是过去的时间";
        try {
            String[] fields = StringUtils.tokenizeToStringArray(expression, " ");
            List<Integer> years = Lists.newArrayList();
            if (null == fields || fields.length < 6 || fields.length > 7) {
                return new GeneralResult<>(false, message);
            } else if (fields.length == 7) {
                if (fields[6].contains("/")) {
                    return new GeneralResult<>(false, message);
                }
                String[] values = StringUtils.delimitedListToStringArray(fields[6], ",");
                for (String value : values) {
                    int[] range = getRange(value, 0, 10000);
                    for (int i : range) {
                        years.add(i);
                    }
                }
                int currentYear = Calendar.getInstance().get(Calendar.YEAR);
                boolean flag = false;
                for (Integer year : years) {
                    if (year >= currentYear) {
                        flag = true;
                    }
                }
                if (!flag) {
                    return new GeneralResult<>(false, message);
                }
                expression =
                    fields[0] + " " + fields[1] + " " + fields[2] + " " + fields[3] + " " + fields[4] + " " + fields[5];
            }
            CronExpression generator = CronExpression.parse(expression);
            LocalDateTime time = generator.next(LocalDateTime.now());
            if (null == time) {
                return new GeneralResult<>(false, message);
            }
            if (fields.length == 7) {
                Calendar calendar = Calendar.getInstance();
                calendar.setTime(Date.from(time.atZone(ZoneId.systemDefault()).toInstant()));
                int execYear = calendar.get(Calendar.YEAR);
                for (Integer year : years) {
                    if (execYear > year) {
                        return new GeneralResult<>(false, message);
                    }
                }
            }
            return new GeneralResult<>(true, "success");
        } catch (Exception e) {
            log.warn("Cron expression is invalid");
            return new GeneralResult<>(false, message);
        }
    }

    private static int[] getRange(String field, int min, int max) {
        int[] result = new int[2];
        if (!field.contains("-")) {
            result[0] = result[1] = Integer.valueOf(field);
        } else {
            String[] split = StringUtils.delimitedListToStringArray(field, "-");
            if (split.length > 2) {
                throw new IllegalArgumentException("Range has more than two fields: '" + field + "' in expression");
            }
            result[0] = Integer.valueOf(split[0]);
            result[1] = Integer.valueOf(split[1]);
        }
        if (result[0] >= max || result[1] >= max) {
            throw new IllegalArgumentException("Range exceeds maximum (" + max + "): '" + field + "' in expression");
        }
        if (result[0] < min || result[1] < min) {
            throw new IllegalArgumentException("Range less than minimum (" + min + "): '" + field + "' in expression");
        }
        if (result[0] > result[1]) {
            throw new IllegalArgumentException("Invalid inverted range: '" + field + "' in expression");
        }
        return result;
    }

}
