package com.bocloud.tse.scheduler.core;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Component;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.intf.NodeModuleService;
import com.bocloud.tse.intf.TaskInstanceService;
import com.bocloud.tse.intf.TaskNodeService;
import com.bocloud.tse.intf.TaskService;
import com.bocloud.tse.model.Job;
import com.bocloud.tse.model.TaskResult;
import com.bocloud.tse.scheduler.jobclient.JobClient;
import com.bocloud.tse.scheduler.listener.TaskListener;
import com.bocloud.tse.scheduler.utils.TaskProgressPublisher;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.megatron.common.model.Result;
import com.megatron.framework.core.CurrentService;

import lombok.extern.slf4j.Slf4j;

/**
 * 任务调度器,
 */
@Slf4j
@Component
public class JobScheduler extends AbstractScheduler implements InitializingBean, TaskListener {

    /**
     * 编排任务执行线程池
     */
    private ExecutorService taskExecutorPool = null;
    /**
     * 任务节点执行线程池
     */
    private ThreadPoolExecutor nodeExecutorPool = null;
    @Autowired
    private CurrentService service;
    @Autowired
    private JobClient jobClient;
    @Autowired
    private JobChannel channel;
    @Autowired
    private TaskProgressPublisher publisher;
    @Autowired
    private TaskService taskService;
    @Autowired
    private TaskNodeService nodeService;
    @Autowired
    private TaskInstanceService instanceService;
    @Autowired
    private NodeModuleService moduleService;
    @Autowired
    private AmqpTemplate amqpTemplate;
    @Autowired
    private StringRedisTemplate redisTemplate;
    private SchedulerDaemon daemon = null;

    @Override
    public void onCompleted(TaskResult taskResult) {
        String taskId = taskResult.getTaskId();
        log.info("Task [{}] are executed :{}", taskResult.getTaskId(), taskResult.isSuccess());
        TaskCache.instance().removeTaskExecutor(taskId);
    }

    /**
     * 移除定时任务
     */
    public Result removeCronTask(String taskId) {
        Result response = jobClient.cancelJob(taskId);
        log.info(JSONObject.toJSONString(response));
        return new Result(response.isSuccess(), response.isSuccess() ? "定时任务移除成功！" : "定时任务移除失败！");
    }

    /**
     * 重置定时任务
     */
    public Result resetCronTask(String taskId, Job job) {
        Result response = jobClient.cancelJob(taskId);
        if (response.isSuccess()) {
            boolean result = channel.insertTask(job);
            return new Result(result, result ? "任务重置成功" : "任务重置失败！");
        }
        return new Result(response.isSuccess(), "任务重置失败！");
    }

    /**
     * 关闭任务引擎
     */
    public void switchOff() {
        this.daemon.swithOff();
        log.info("任务引擎关闭成功");
    }

    /**
     * 激活任务引擎
     */
    public void switchOn() {
        if (!isRunning()) {
            daemon = new SchedulerDaemon(service, this, publisher, jobClient, amqpTemplate, taskService, nodeService,
                instanceService, taskExecutorPool, nodeExecutorPool, moduleService, redisTemplate);
            this.daemon.start();
        }
        log.info("任务引擎开启成功");
    }

    public boolean isRunning() {
        return this.daemon.isRunning();
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        log.info("====== 启动 SchedulerDaemon ======");
        // 编排任务调度线程池,默认是采用可用CPU2倍的线程数
        int processors = Runtime.getRuntime().availableProcessors();
        taskExecutorPool = new ThreadPoolExecutor(processors * 2, Integer.MAX_VALUE, 60L, TimeUnit.SECONDS,
            new SynchronousQueue<Runnable>(), new ThreadFactoryBuilder().build());
        // 编排任务节点调度线程池,默认是采用可用CPU5倍的线程数
        nodeExecutorPool = new ThreadPoolExecutor(processors * 5, Integer.MAX_VALUE, 60L, TimeUnit.SECONDS,
            new SynchronousQueue<Runnable>(), new ThreadFactoryBuilder().build());
        nodeExecutorPool.allowCoreThreadTimeOut(true);
        daemon = new SchedulerDaemon(service, this, publisher, jobClient, amqpTemplate, taskService, nodeService,
            instanceService, taskExecutorPool, nodeExecutorPool, moduleService, redisTemplate);
        daemon.start();
        log.info("====== Start SchedulerDaemon Success ======");
    }

    @Override
    void destory() {
        this.switchOff();
        this.taskExecutorPool.shutdown();
        this.nodeExecutorPool.shutdown();
    }

}
