package com.bocloud.tse.scheduler.core;

import com.bocloud.tse.enums.TaskParam;
import com.bocloud.tse.intf.NodeModuleService;
import com.bocloud.tse.intf.TaskInstanceService;
import com.bocloud.tse.intf.TaskNodeService;
import com.bocloud.tse.intf.TaskService;
import com.bocloud.tse.model.Job;
import com.bocloud.tse.scheduler.executor.TaskExecutor;
import com.bocloud.tse.scheduler.jobclient.JobClient;
import com.bocloud.tse.scheduler.utils.TaskProgressPublisher;
import com.megatron.common.utils.Common;
import com.megatron.framework.core.CurrentService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.data.redis.core.StringRedisTemplate;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;

/**
 * 调度器守护进程，用来持续从内存队列中获取到实时的编排任务并执行。
 */
@Slf4j
public class SchedulerDaemon extends Thread {

    private TaskCache taskDataCache = TaskCache.instance();
    private boolean running = true;
    private JobScheduler scheduler;
    private TaskService taskService;
    private TaskNodeService nodeService;
    private TaskProgressPublisher publisher;
    private TaskInstanceService instanceService;
    private JobClient jobClient;
    private ExecutorService taskExecutorPool;
    private ExecutorService nodeExecutorPool;
    private AmqpTemplate amqpTemplate;
    private CurrentService service;
    private NodeModuleService nodeModuleService;
    private StringRedisTemplate redisTemplate;
    private List<Job> jobs;

    @Override
    public void run() {
        log.info("-------开启任务执行守护进程--------");
        String instance = service.getInstance().split(":")[0];
        int processors = Runtime.getRuntime().availableProcessors() * 10;
        while (running) {
            try {
                String num = redisTemplate.opsForValue().get("TaskProcessors");
                processors = StringUtils.isEmpty(num) ? processors : Integer.valueOf(num);
            } catch (Exception e) {
            }
            try {
                if (jobs.isEmpty()) {
                    String executeEnabled = redisTemplate.opsForValue().get("TaskExecuteEnabled");
                    boolean enabled = StringUtils.isEmpty(executeEnabled) ? true : Boolean.valueOf(executeEnabled);
                    if (!enabled) {
                        log.info("任务执行开关已关闭，不接受新任务入池");
                        Thread.sleep(3000);
                        continue;
                    }
                    String excludeServices = redisTemplate.opsForValue().get("TaskExcludeServices");
                    if (!StringUtils.isEmpty(excludeServices)) {
                        List<String> instances = java.util.Arrays.asList(excludeServices.split(","));
                        if (instances.contains(instance)) {
                            log.info("当前服务已排除在任务引擎之外，不接受新任务入池");
                            Thread.sleep(3000);
                            continue;
                        }
                    }
                    long start = System.currentTimeMillis();
                    jobs.addAll(instanceService.getJobs());
                    if (!jobs.isEmpty()) {
                        log.info("获取任务实例，用时：" + (System.currentTimeMillis() - start));
                    }
                    if (jobs.isEmpty()) {
                        // log.info("暂无可执行任务，等待两秒");
                        Thread.sleep(2000);
                        continue;
                    }
                }
                if (taskDataCache.getTaskPoolSize() >= processors) {
                    log.info("正在执行的任务数已达到线程池最大值【{}】，不接受新任务入池", processors);
                    Thread.sleep(3000);
                    continue;
                }
                // 拿出job，并移除job
                Job job = jobs.get(0);
                TaskExecutor executor =
                    new TaskExecutor(job, publisher, jobClient, taskService, nodeService, instanceService, scheduler,
                        nodeExecutorPool, amqpTemplate, service, nodeModuleService, redisTemplate);
                taskExecutorPool.submit(executor);
                if (StringUtils.isEmpty(job.getParam(TaskParam.TASKPJOB.name()))) {
                    taskDataCache.putTaskExecutor(job.getTaskId(), executor);
                }
                jobs.remove(0);
                log.info("任务【{}】 放入任务线程池", job.getParam(Common.NAME));
            } catch (InterruptedException e) {
                log.error("SchedulerDaemon get InterruptedException", e);
            }
        }
    }

    public SchedulerDaemon(CurrentService service, JobScheduler scheduler, TaskProgressPublisher publisher,
        JobClient jobClient, AmqpTemplate amqpTemplate, TaskService taskService, TaskNodeService nodeService,
        TaskInstanceService instanceService, ExecutorService taskExecutorPool, ExecutorService nodeExecutorPool,
        NodeModuleService nodeModuleService, StringRedisTemplate redisTemplate) {
        super();
        this.service = service;
        this.scheduler = scheduler;
        this.publisher = publisher;
        this.jobClient = jobClient;
        this.amqpTemplate = amqpTemplate;
        this.taskService = taskService;
        this.nodeService = nodeService;
        this.instanceService = instanceService;
        this.taskExecutorPool = taskExecutorPool;
        this.nodeExecutorPool = nodeExecutorPool;
        this.nodeModuleService = nodeModuleService;
        this.redisTemplate = redisTemplate;
        this.jobs = new ArrayList<>();
    }

    public void swithOff() {
        this.running = false;
    }

    public boolean isRunning() {
        return this.running;
    }

}
