package com.bocloud.tse.scheduler.core;

import java.util.concurrent.ConcurrentHashMap;

import com.bocloud.tse.model.TaskResult;
import com.bocloud.tse.scheduler.executor.NodeExecutor;
import com.bocloud.tse.scheduler.executor.TaskExecutor;

/**
 * 任务数据缓存器，用来存放任务调度执行过程中的各种数据
 */
public class TaskCache {

    /**
     * 任务执行引擎池
     */
    private ConcurrentHashMap<String, TaskExecutor> taskExecutorPool =
        new ConcurrentHashMap<String, TaskExecutor>(50, 0.75f);
    /**
     * 任务执行引擎池
     */
    private ConcurrentHashMap<String, NodeExecutor> nodeExecutorPool =
        new ConcurrentHashMap<String, NodeExecutor>(50, 0.75f);

    public ConcurrentHashMap<String, TaskExecutor> getTaskExecutorPool() {
        return taskExecutorPool;
    }

    public ConcurrentHashMap<String, NodeExecutor> getNodeExecutorPool() {
        return nodeExecutorPool;
    }

    /**
     * 当前服务器是否存在此任务节点执行器
     */
    public boolean hasNodeExecutor(String taskId) {
        return this.nodeExecutorPool.containsKey(taskId);
    }

    /**
     * 将节点任务执行器加入到缓存中
     */
    public void putNodeExecutor(String taskId, NodeExecutor executor) {
        this.nodeExecutorPool.putIfAbsent(taskId, executor);
    }

    /**
     * 将节点任务执行器从缓存中移除
     */
    public void removeNodeExecutor(String taskId) {
        this.nodeExecutorPool.remove(taskId);
    }

    /**
     * 通知指定的任务节点执行器相应的结果
     */
    public void noticeNodeExecutor(String taskId, TaskResult taskResult) {
        this.nodeExecutorPool.get(taskId).onCompleted(taskResult);
    }

    /**
     * 当前服务器是否存在该任务的执行引擎
     */
    public boolean hasTaskExecutor(String taskId) {
        return this.taskExecutorPool.containsKey(taskId);
    }

    /**
     * 放入任务执行引擎
     */
    public void putTaskExecutor(String taskId, TaskExecutor taskEngine) {
        this.taskExecutorPool.putIfAbsent(taskId, taskEngine);
    }

    /**
     * 获取任务执行引擎
     */
    public TaskExecutor getTaskExecutor(String taskId) {
        return this.taskExecutorPool.get(taskId);
    }

    /**
     * 移除已经执行完成的任务执行引擎
     */
    public void removeTaskExecutor(String taskId) {
        this.taskExecutorPool.remove(taskId);
    }

    public int getTaskPoolSize() {
        return this.taskExecutorPool.size();
    }

    public int getNodePoolSize() {
        return this.nodeExecutorPool.size();
    }

    /**
     * 私有构造方法
     */
    private TaskCache() {
        if (null != TaskCache.instance()) {
            // 防止利用反射机制实现多个对象的漏洞
            throw new RuntimeException("IDFactory正在遭受反射攻击！");
        }
    }

    /**
     * 获取ID工厂实例方法
     */
    public static TaskCache instance() {
        return CacheContainer.cache;
    }

    /**
     * 缓存工厂容器，利用静态内部类的方式实现单例模式，屏蔽多线程干扰
     */
    private static class CacheContainer {
        private static TaskCache cache = new TaskCache();
    }

    /**
     * 防止反序列化获取多个对象的漏洞
     */
    private Object readResolve() {
        return CacheContainer.cache;
    }
}
