package com.bocloud.tse.scheduler.executor;

import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.lang3.StringUtils;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.entity.TaskNode;
import com.bocloud.tse.enums.Catalog;
import com.bocloud.tse.enums.Catalog.Node;
import com.bocloud.tse.enums.Category;
import com.bocloud.tse.enums.TaskParam;
import com.bocloud.tse.model.Job;
import com.bocloud.tse.model.JobResult;
import com.bocloud.tse.model.NodeControl;
import com.bocloud.tse.model.TaskResult;
import com.bocloud.tse.scheduler.core.TaskCache;
import com.bocloud.tse.scheduler.jobclient.JobClient;
import com.bocloud.tse.scheduler.listener.TaskListener;
import com.bocloud.tse.scheduler.utils.NodeDependencyHandler;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Result;
import com.megatron.common.utils.Common;

import lombok.extern.slf4j.Slf4j;

/**
 * 单一小任务执行器，配合scheduler中的jobExecutor将任务分发下去，这里的任务都是实时任务中的小任务。 主要解决将小任务提交到jobTracker中调度TaskTracker并等待结果返回的逻辑。有等待超时机制。
 */
@Slf4j
public class NodeExecutor implements TaskListener, Runnable {

    private int executeTimes = 1;
    /**
     * 默认失败之间的重试间隔为10秒
     */
    private long interval = 10;
    private long timeout = 10 * 60;
    private Job job;
    private TaskNode taskNode;
    private NodeControl control;
    private TaskExecutor taskExecutor;
    private boolean finished = false;
    private TaskResult taskResult = null;
    private JobClient jobClient;
    private Map<String, List<String>> preLinks;
    private ConcurrentHashMap<String, TaskResult> resultMap;

    @Override
    public void run() {
        TaskCache.instance().putNodeExecutor(taskNode.getUuid(), this);
        JobResult jobResult = null;
        try {
            GeneralResult<Node> result = Catalog.Node.hasValue(taskNode.getCatalog());
            if (result.isFailed()) {
                log.error("任务【{}】的节点任务【{}】分类校验失败：【{}】", taskNode.getInstanceName(), taskNode.getName(),
                    result.getMessage());
                buildJob();
                jobResult =
                    new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "任务节点分类参数格式错误")), job);
                taskResult = new TaskResult(false, "任务节点分类参数格式错误！", jobResult, job.getTaskId(), taskNode);
            } else {
                switch (result.getData()) {
                    case CTRL:
                        handleCtrlNode();
                        break;
                    default:
                        handleTaskNode();
                        break;
                }
            }
        } catch (Exception e) {
            taskResult = new TaskResult(false, e.getMessage(),
                new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, e.getMessage())), job),
                job.getTaskId(), taskNode);
        }
        TaskCache.instance().removeNodeExecutor(taskNode.getUuid());
        this.taskExecutor.onCompleted(taskResult);
    }

    /**
     * 处理开始节点任务
     */
    private void handleCtrlNode() {
        if (Category.Node.JUDGE.name().equalsIgnoreCase(taskNode.getCategory())) {
            handleJudgeNode();
        } else {
            handleSimpleNode();
        }
    }

    /**
     * 处理任务节点
     */
    private void handleTaskNode() {
        log.info("收到任务【{}】的节点任务【{}】", taskNode.getInstanceName(), taskNode.getName());
        GeneralResult<Category.Node> result = Category.Node.hasValue(taskNode.getCategory());
        if (result.isFailed()) {
            log.info("任务【{}】的节点任务【{}】类型不存在", taskNode.getInstanceName(), taskNode.getName());
            buildJob();
            JobResult jobResult =
                new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "任务节点类型参数格式错误")), job);
            taskResult = new TaskResult(false, "任务节点类型参数格式错误", jobResult, job.getTaskId(), taskNode);
            return;
        }
        handldFailover();
        for (int executedTime = 1; executedTime <= executeTimes; executedTime++) {
            // log.info("节点任务【{}】第【{}】次执行", taskNode.getName(), executedTime);
            //节点任务执行
            execute();
            taskResult.getJobResult().getJob().setParam(TaskParam.RETRYTIMES.name(),
                new Integer(executedTime - 1).toString());
            if (taskResult.isSuccess()) {
                // log.info("当前节点【{}】执行成功！执行次数【{}】", taskNode.getName(), executedTime);
                break;
            }
            log.error("任务【{}】的节点任务【{}】第【{}】次执行失败，错误结果【{}】", taskNode.getInstanceName(), taskNode.getName(),
                executedTime, JSONObject.toJSONString(taskResult));
            if (executedTime < executeTimes) {
                this.sleep(interval * 1000L);
            }
        }
        taskResult.setTaskNode(taskNode);
        log.info("任务【{}】的节点任务【{}】执行结束", taskNode.getInstanceName(), taskNode.getName());
    }

    /**
     * 处理判断节点任务节点
     */
    private void handleJudgeNode() {
        buildJob();
        JobResult jobResult = new JobResult();
        jobResult.setJob(job);
        Result result = new NodeDependencyHandler(taskNode, resultMap).getJudgeResult();
        if (result.isFailed()) {
            jobResult.setMsg(JSONObject.toJSONString(new GeneralResult<>(false, "参数有误，导致无法解析")));
            taskResult = new TaskResult(false, "failed", jobResult, job.getTaskId(), taskNode);
        } else {
            jobResult.setSuccess(true);
            jobResult.setMsg(JSONObject.toJSONString(new GeneralResult<>(true, "success")));
            taskNode.setDicision(Boolean.valueOf(result.getMessage()));
            taskResult = new TaskResult(true, "success", jobResult, job.getTaskId(), taskNode);
        }
    }

    /**
     * 处理简单任务节点
     */
    private void handleSimpleNode() {
        buildJob();
        JobResult result = new JobResult(true, JSONObject.toJSONString(new GeneralResult<>(true, "success")), job);
        taskResult = new TaskResult(true, "success", result, job.getTaskId(), taskNode);
    }

    /**
     * 处理重试次数和间隔逻辑
     */
    private void handldFailover() {
        if (control.isFailover()) {
            // 失败后重试
            if (0 == control.getRetryTime()) {
                // 重试次数为空，默认三次
                executeTimes = 1;
            } else {
                // 执行次数为重试次数+1
                executeTimes = executeTimes + control.getRetryTime();
            }
            if (0 != control.getIntervals() && control.getIntervals() > 0) {
                // 重试时间为不为空且时间不小于0s,否则采用默认的10s间隔
                interval = control.getIntervals();
            }
        } else {
            executeTimes = 1;
        }
        // log.info("节点【{}】需要进行失败后重试【{}】次，重试间隔为:【{}】毫秒", taskNode.getName(), executeTimes - 1, interval);
    }

    /**
     * 任务节点执行完成之后的动作
     */
    @Override
    public void onCompleted(TaskResult taskResult) {
        this.finished = true;
        this.taskResult = taskResult;
    }

    /**
     * 处理节点间的依赖传值
     */
    private void handleDependency() {
        // log.info("开始处理节点任务【{}】前置依赖关系", taskNode.getName());
        List<String> preNodes = preLinks.get(taskNode.getNodeId());
        // log.info("获取节点任务【{}】前置依赖节点信息", taskNode.getName());
        new NodeDependencyHandler(taskNode, resultMap, preNodes).handle();
        // log.info("结束处理节点任务【{}】前置依赖关系", taskNode.getName());
    }

    /**
     * 节点任务执行
     */
    private void execute() {
        // 处理节点之间的依赖关系
        handleDependency();
        // 将任务提交到JobTracker进行调度
        this.buildJob();
        //提交节点任务
        Result response = jobClient.submitJob(job);
        if (response.isSuccess()) {
            // 任务提交成功，等待执行结果
            log.info("任务【{}】的节点任务【{}】提交成功", taskNode.getInstanceName(), taskNode.getName());
            finished = false;
            waitForResult(timeout);
        } else {
            // 任务提交失败，直接表示该任务执行失败
            log.error("任务【{}】的节点任务【{}】提交失败", taskNode.getInstanceName(), taskNode.getName());
            JobResult jobResult =
                new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "任务提交失败")), job);
            taskResult = new TaskResult(false, "任务节点提交失败！", jobResult, job.getTaskId(), taskNode);
        }
    }

    /**
     * 等待任务执行结果
     */
    private void waitForResult(long timeout) {
        long start = System.currentTimeMillis();
        long cost = 0L;
        // 这里提供5秒钟的超时经验值时间，用来缓解由于网络延迟造成的部分超时任务问题
        timeout = (timeout + 60 + 60) * 1000;
        while (cost < timeout) {
            if (finished) {
                return;
            }
            sleep(5000);
            cost = System.currentTimeMillis() - start;
        }
        log.error("任务【{}】的节点任务【{}】 执行超时！", taskNode.getInstanceName(), taskNode.getName());
        JobResult jobResult = new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "任务执行超时")), job);
        taskResult = new TaskResult(false, "任务执行超时", jobResult, job.getTaskId(), taskNode);
    }

    /**
     * 线程睡眠
     */
    private void sleep(long time) {
        try {
            Thread.sleep(time);
        } catch (InterruptedException e) {
            log.error("Thread [{}] InterruptedException", Thread.currentThread().getName());
            Thread.currentThread().interrupt();
        }
    }

    /**
     * 构造方法
     */
    public NodeExecutor(TaskNode taskNode, TaskExecutor taskExecutor, JobClient jobClient,
        Map<String, List<String>> preLinks, ConcurrentHashMap<String, TaskResult> resultMap) {
        super();
        this.taskNode = taskNode;
        this.taskExecutor = taskExecutor;
        this.jobClient = jobClient;
        this.taskNode.setGmtStart(new Date());
        this.preLinks = preLinks;
        this.resultMap = resultMap;
        this.control = JSONObject.parseObject(taskNode.getControl(), NodeControl.class);
        if (null != control) {
            this.timeout = control.getTimeout();
        } else {
            control = new NodeControl();
        }
    }

    /**
     * 构建任务
     */
    private void buildJob() {
        job = new Job();
        job.setPriority(1);
        job.setParam(TYPE, getCode(taskNode.getModuleType()));
        String data = taskNode.getParams();
        if (StringUtils.isEmpty(data)) {
            data = taskNode.getInput();
        }
        job.setParam(DATA, data);
        job.setParam(MODULETYPE, taskNode.getModuleType());
        job.setParam(Common.NAME, taskNode.getName());
        job.setParam(TIME, String.valueOf(timeout));
        job.setMaxRetryTimes(0);
        job.setTaskId(taskNode.getUuid());
        job.setParam(CODE, taskNode.getCode());
        job.setParam(INSTANCE, taskNode.getInstanceName());
    }

    private String getCode(String code) {
        String[] values = code.split("\\.");
        return values[1];
    }

}
