package com.bocloud.tse.scheduler.executor;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.cmp.boot.model.BoCloudExchange;
import com.bocloud.tse.core.utils.CronUtils;
import com.bocloud.tse.entity.Task;
import com.bocloud.tse.entity.TaskInstance;
import com.bocloud.tse.entity.TaskNode;
import com.bocloud.tse.enums.Catalog;
import com.bocloud.tse.enums.Category;
import com.bocloud.tse.enums.Status;
import com.bocloud.tse.intf.NodeModuleService;
import com.bocloud.tse.intf.TaskInstanceService;
import com.bocloud.tse.intf.TaskNodeService;
import com.bocloud.tse.intf.TaskService;
import com.bocloud.tse.model.*;
import com.bocloud.tse.scheduler.core.JobScheduler;
import com.bocloud.tse.scheduler.core.TaskCache;
import com.bocloud.tse.scheduler.jobclient.JobClient;
import com.bocloud.tse.scheduler.listener.TaskListener;
import com.bocloud.tse.scheduler.selector.NodeSelector;
import com.bocloud.tse.scheduler.selector.TaskNodeSelector;
import com.bocloud.tse.scheduler.utils.TaskProgressPublisher;
import com.bocloud.tse.utils.NodeHelper;
import com.google.common.collect.Maps;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.OperateResult;
import com.megatron.common.model.RequestContext;
import com.megatron.common.model.Result;
import com.megatron.common.utils.Common;
import com.megatron.common.utils.ListTool;
import com.megatron.framework.core.CurrentService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.data.redis.core.StringRedisTemplate;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;

/**
 * 编排任务执行器
 */
@Slf4j
public class TaskExecutor implements TaskListener, Runnable {

    /**
     * 任务等待执行节点数量
     */
    private int waitingNode = 0;
    /**
     * 执行中的任务节点数量
     */
    private int runningNode = 0;
    /**
     * 任务执行实例是否取消
     */
    private boolean cancel = false;
    /**
     * 任务是否执行结束
     */
    private boolean finished = false;
    /**
     * 整体任务实例是否暂停
     */
    private boolean suspendable = false;
    /**
     * 整体任务实例是否失败
     */
    private boolean failed = false;
    /**
     * 暂停中的任务节点列表
     */
    private List<TaskNode> suspendingNodes = new ArrayList<>();
    /**
     * 节点任务执行线程池
     */
    private ExecutorService nodeEexecutorPool;
    private Result response = null;
    private Job job;
    private TaskInstance instance = null;
    private TaskResult taskResult;
    private JobScheduler scheduler;
    private TaskProgressPublisher publisher;
    private TaskService taskService;
    private TaskNodeService nodeService;
    private TaskInstanceService instanceService;
    private NodeSelector<TaskNode> nodeSelector;
    private JobClient jobClient;
    private AmqpTemplate amqpTemplate;
    private CurrentService service;
    private Map<String, List<String>> preLinks;
    private Map<String, List<TaskEdge>> nextEdges;
    private Map<String, Long> nodeMap;
    private Map<String, String> ingoreNodes;
    private Map<String, List<TaskNode>> nextNodes;
    private boolean hasAction = false;
    private ConcurrentHashMap<String, TaskResult> resultMap = new ConcurrentHashMap<String, TaskResult>(50, 0.75f);

    private List<TaskNode> taskNodes = null;
    private boolean noEdges = false;
    private boolean single = false;
    private boolean monolayer = false;
    private StringRedisTemplate redisTemplate;

    @Override
    public void run() {
        log.info("开始执行任务 【{}】 ", job.getParam(Common.NAME));
        // 任务初始化
        init();
        log.info("任务 【{}】初始化完毕 ", job.getParam(Common.NAME));
        String type = job.getParam(TYPE);
        GeneralResult<Catalog.Task> result = Catalog.Task.hasValue(type);
        if (result.isFailed()) {
            log.error("任务【 {}】 类型 【{}】不支持！", instance.getName(), type);
            handleException();
            return;
        }
        switch (result.getData()) {
            case STANDARD:
                executeRealTask();
                break;
            case SCHEDULED:
                executeCronTask();
                break;
            default:
                log.error("收到不支持的任务类型：【{}】", type);
                taskResult = new TaskResult(false, "Unsupported Task!!", job.getTaskId());
                break;
        }
        log.info("任务 【{}】执行结束 ", job.getParam(Common.NAME));
        handleFinish();
    }

    /**
     * 处理判断节点不满足条件的后续节点
     */
    private void handlerRelation(String nodeId, List<String> nodes) {
        nodes.add(nodeId);
        List<TaskEdge> edges = nextEdges.get(nodeId);
        if (ListTool.isEmpty(edges)) {
            return;
        }
        for (TaskEdge edge : edges) {
            String target = edge.getTarget();
            target = ingoreNodes.containsKey(target) ? ingoreNodes.get(target) : target;
            List<String> preNodes = preLinks.get(target);
            if (ListTool.isEmpty(preNodes)) {
                continue;
            }
            if (preNodes.size() > 1) {
                continue;
            }
            nodes.add(target);
            handlerRelation(target, nodes);
        }
    }

    @Override
    public void onCompleted(TaskResult taskResult) {
        TaskNode taskNode = taskResult.getTaskNode();
        // 缓存节点结果信息
        resultMap.put(taskNode.getNodeId(), taskResult);
        // 更新TaskNode的数据库信息
        NodeControl control = JSONObject.parseObject(taskNode.getControl(), NodeControl.class);
        if (null == control) {
            control = new NodeControl();
        }
        taskNode.setGmtFinish(new Date());
        BigDecimal cost = new BigDecimal(taskNode.getGmtFinish().getTime() - taskNode.getGmtStart().getTime())
            .divide(new BigDecimal(1000), 3, RoundingMode.HALF_UP);
        taskNode.setCost(cost.doubleValue());
        // 检测返回数据大小，超过5M的数据拒绝入库
        checkDataSize(taskNode, taskResult);
        taskNode.setOutput(taskResult.getJobResult().getMsg());
        if (taskResult.isSuccess()) {
            taskNode.setStatus(control.isSuspensible() ? Status.Node.SUSPENDED.name() : Status.Node.SUCCESS.name());
        } else {
            taskNode.setStatus(control.isSkippable() ? Status.Node.SKIPPED.name() : Status.Node.FAILED.name());
        }
        nodeService.modify(taskNode);
        String status = control.isSkippable() ? Status.Node.SUCCESS.name() : taskNode.getStatus();
        updateRelation(instance.getId(), taskNode.getId(), taskNode.getNodeId(), status);
        // 发送结果消息
        sendMessage(taskNode, taskResult);
        taskResult.setTaskNode(taskNode);
        if (single) {
            handleSingle(taskResult, control.isSkippable());
        } else if (noEdges) {
            handleNoEdges(taskResult, control.isSkippable());
        } else {
            boolean flag = isExecNext(taskResult);
            if (flag) {
                log.info("准备执行任务【{}】剩余其他节点", instance.getName());
                synchronized (this) {
                    execNext(taskNode);
                }
            }
        }
        // 发送结果消息
        if (taskResult.isFailed()) {
            this.sendMessage(taskNode, taskResult);
        }
        // log.info("任务节点【{}】结束响应完成！", taskNode.getName());
    }

    private static String div(String str1) {
        BigDecimal source = new BigDecimal(str1);
        BigDecimal target = new BigDecimal(1024);
        BigDecimal divide = source.divide(target, 2, RoundingMode.HALF_UP);
        return divide.toString();
    }

    private void checkDataSize(TaskNode taskNode, TaskResult taskResult) {
        try {
            if (null == taskResult.getJobResult().getMsg()) {
                taskNode.setOutput(taskResult.getJobResult().getMsg());
                return;
            }
            String kbStr = div(String.valueOf(taskResult.getJobResult().getMsg().length()));
            String mbStr = div(kbStr);
            if (5 >= Double.parseDouble(mbStr)) {
                taskNode.setOutput(taskResult.getJobResult().getMsg());
            } else {
                JSONObject resultData = JSONObject.parseObject(taskResult.getJobResult().getMsg());
                JSONObject tooLarge = new JSONObject();
                tooLarge.put("success", resultData.getBoolean("success"));
                tooLarge.put("failed", resultData.getBoolean("failed"));
                tooLarge.put("message", resultData.getString("message"));
                JSONObject data = new JSONObject();
                data.put("message", "任务返回的数据内容过大，无法进行详细查看!");
                JSONArray nodes = new JSONArray();
                nodes.add(data);
                tooLarge.put("data", nodes);
                taskNode.setOutput(tooLarge.toJSONString());
            }
        } catch (NumberFormatException e) {
            log.error("Check Data Size Error", e);
        }
    }

    /**
     * 处理仅仅只有一个任务节点
     */
    private void handleSingle(TaskResult taskResult, boolean skippable) {
        runningNode = 0;
        waitingNode = 0;
        if (taskResult.isFailed() && !skippable) {
            failed = true;
            handleFinish(Status.Node.FAILED, null);
        } else {
            handleFinish(Status.Node.SUCCESS, null);
        }
    }

    /**
     * 处理没有连线的任务节点
     */
    private synchronized void handleNoEdges(TaskResult taskResult, boolean skippable) {
        runningNode = runningNode - 1;
        waitingNode = waitingNode - 1;
        if (taskResult.isFailed() && !skippable) {
            failed = true;
            handleFinish(Status.Node.FAILED, null);
        } else {
            handleFinish(Status.Node.SUCCESS, null);
        }
    }

    /**
     * 判断是否需要执行next，
     */
    private synchronized boolean isExecNext(TaskResult taskResult) {
        TaskNode taskNode = taskResult.getTaskNode();
        // 处理判断节点
        if (Category.Node.JUDGE.name().equalsIgnoreCase(taskNode.getCategory())) {
            if (null != taskNode.getDicision()) {
                List<TaskEdge> edges = nextEdges.get(taskNode.getNodeId());
                List<String> nodeIds = new ArrayList<>();
                for (TaskEdge edge : edges) {
                    if (edge.isDicision() == taskNode.getDicision()) {
                        continue;
                    }
                    String target = edge.getTarget();
                    target = ingoreNodes.containsKey(target) ? ingoreNodes.get(target) : target;
                    handlerRelation(target, nodeIds);
                }
                waitingNode = waitingNode - nodeIds.size();
                for (String nodeId : nodeIds) {
                    nodeService.updateStatus(nodeMap.get(nodeId), Status.Node.NOREADY.name(), 1L);
                    updateRelation(instance.getId(), nodeMap.get(nodeId), nodeId, Status.Node.NOREADY.name());
                }
            }
        }
        // 任务被取消
        if (cancel) {
            runningNode = runningNode - 1;
            waitingNode = 0;
            handleFinish(Status.Node.CANCELED, null);
            return false;
        }
        NodeControl control = JSONObject.parseObject(taskNode.getControl(), NodeControl.class);
        if (null == control) {
            control = new NodeControl();
        }
        if (taskResult.isSuccess()) {
            // 任务执行成功，但是成功暂停
            if (control.isSuspensible()) {
                runningNode = runningNode - 1;
                waitingNode = waitingNode - 1;
                // 执行任务暂停
                suspend(taskNode);
                sendMessage(taskNode, taskResult);
                log.info("任务【{}】的节点任务【{}】执行结束后成功暂停！暂停中...", instance.getName(), taskNode.getName());
                return false;
            }
        } else {
            log.info("任务【{}】的节点任务【{}】执行失败！开始处理节点执行失败逻辑！", instance.getName(), taskNode.getName());
            // log.info("节点【{}】可以失败跳过，任务继续执行下一节点...", taskNode.getName());
            if (!control.isSkippable()) {
                // 如果当前节点后面全是判断节点节点，则不处理，如果存在任务节点则执行失败
                // log.info("节点【{}】不允许失败跳过，执行任务失败逻辑", taskNode.getName());
                List<TaskNode> nodes = nextNodes.get(taskNode.getNodeId());
                if (ListTool.isEmpty(nodes)) {
                    failed = true;
                } else {
                    for (TaskNode node : nodes) {
                        if (!node.getModuleType().contains("ctrl.judge")) {
                            failed = true;
                            break;
                        }
                    }
                }
                if (failed) {
                    runningNode = runningNode - 1;
                    waitingNode = 0;
                    handleFinish(Status.Node.FAILED, null);
                    return false;
                }
            }
        }
        runningNode = runningNode - 1;
        waitingNode = waitingNode - 1;
        return true;
    }

    /**
     * 执行实时任务
     */
    private void executeRealTask() {
        if (null == instance) {
            log.warn("任务出现异常无法执行");
            JobResult jobResult =
                new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "任务出现异常无法执行")), job);
            taskResult = new TaskResult(false, "任务出现异常无法执行", jobResult, job.getTaskId());
            return;
        }
        // log.info("执行实时任务 【{}】...", instance.getName());
        List<TaskNode> firstNodes = new ArrayList<TaskNode>();
        if (single) {
            firstNodes.addAll(taskNodes);
        } else if (noEdges) {
            for (TaskNode node : taskNodes) {
                if (Status.Node.READY.name().equals(node.getStatus())) {
                    firstNodes.add(node);
                }
            }
        } else {
            GeneralResult<List<TaskNode>> taskNodeResult = nodeSelector.selectFirst();
            if (taskNodeResult.isFailed() || taskNodeResult.getData().isEmpty()) {
                GeneralResult<List<TaskNode>> endNodeResult = nodeSelector.selectLast();
                if (endNodeResult.isFailed()) {
                    log.error("获取任务【{}】的节点任务失败！", instance.getName());
                    JobResult jobResult =
                        new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "参数格式异常")), job);
                    taskResult = new TaskResult(false, "参数格式异常", jobResult, job.getTaskId());
                    failed = true;
                    handleFinish(Status.Node.FAILED, null);
                    return;
                }
                if (ListTool.isEmpty(endNodeResult.getData())) {
                    Boolean hasReady = nodeService.hasReady(instance.getId());
                    if (null != hasReady && !hasReady) {
                        handleFinish(Status.Node.SUCCESS, null);
                    } else {
                        log.error("获取任务【{}】的节点任务失败！", instance.getName());
                        JobResult jobResult =
                            new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "参数格式异常")), job);
                        taskResult = new TaskResult(false, "参数格式异常", jobResult, job.getTaskId());
                        failed = true;
                        handleFinish(Status.Node.FAILED, null);
                    }
                    return;
                }
                taskNodeResult = endNodeResult;
            }
            firstNodes.addAll(taskNodeResult.getData());
        }
        handleSuspend();
        if (cancel) {// 处理任务暂停状态时，执行取消按钮，后面所有任务节点不被执行
            waitingNode = 0;
            handleFinish(Status.Node.CANCELED, null);
            taskResult = new TaskResult(false, "task is canceled", instance.getUuid());
            sendMessage(null, taskResult);
            return;
        }
        runningNode = runningNode + firstNodes.size();
        for (TaskNode node : firstNodes) {
            node.setStatus(Status.Node.RUNNING.name());
            this.nodeService.modify(node);
            updateRelation(instance.getId(), node.getId(), node.getNodeId(), Status.Node.RUNNING.name());
            node.setContext(getContext(instance.getContext()));
            node.setInstanceName(instance.getName());
            //一个一个节点去处理
            NodeExecutor nodeExecutor = new NodeExecutor(node, this, jobClient, preLinks, resultMap);
            if (firstNodes.size() == 1) {
                log.info("first nodes is single,don't create thread to execute");
                nodeExecutor.run();
            } else {
                this.nodeEexecutorPool.submit(nodeExecutor);
            }
        }
        // log.info("节点任务构造成功，提交执行");
        while (!finished) {
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
                log.error("Thread sleep error:", e.getMessage());
            }
        }
        log.info("任务【{}】 执行结果：【{}】", instance.getName(), taskResult.isSuccess());
    }

    /**
     * 执行下一节点
     */
    private void execNext(TaskNode currentNode) {
        if (failed) {// 当任务节点执行失败且没有失败跳过时，任务执行失败
            handleFinish(Status.Node.FAILED, null);
            return;
        }
        handleSuspend();
        if (cancel) {// 处理任务暂停状态时，执行取消按钮，后面所有任务节点不被执行
            waitingNode = 0;
            handleFinish(Status.Node.CANCELED, null);
            return;
        }
        if (runningNode <= 0 && waitingNode <= 0) {
            handleFinish(Status.Node.SUCCESS, null);
            return;
        }
        List<TaskNode> readyNodes = null;
        if (null != currentNode && monolayer) {
            log.error("当前任务【{}】除了游离和一个开始执行节点之外，其他节点没有出度只有开始节点入度", instance.getName());
            if (!nextNodes.containsKey(currentNode.getNodeId())) {
                return;
            }
            readyNodes = nextNodes.get(currentNode.getNodeId());
        } else {
            GeneralResult<List<TaskNode>> nodeResult = nodeSelector.selectNext();
            if (nodeResult.isFailed()) {
                // 获取下一节点失败
                log.error("获取任务【{}】剩余待执行节点任务失败 ，失败原因【{}】", instance.getName(), nodeResult.getMessage());
                failed = true;
                handleFinish(Status.Node.FAILED, null);
                return;
            }
            readyNodes = nodeResult.getData();
        }
        if (!ListTool.isEmpty(readyNodes)) {
            runningNode += readyNodes.size();
            for (TaskNode node : readyNodes) {
                node.setStatus(Status.Node.RUNNING.name());
                nodeService.modify(node);
                updateRelation(instance.getId(), node.getId(), node.getNodeId(), Status.Node.RUNNING.name());
                node.setContext(getContext(instance.getContext()));
                node.setInstanceName(instance.getName());
                NodeExecutor nodeExecutor = new NodeExecutor(node, this, jobClient, preLinks, resultMap);
                this.nodeEexecutorPool.submit(nodeExecutor);
            }
            return;
        }
        log.info("任务【{}】未获取到下一节点，检查任务是否执行完成", instance.getName());
        GeneralResult<List<TaskNode>> endNodeResult = nodeSelector.selectLast();
        if (endNodeResult.isFailed()) {
            failed = true;
            handleFinish(Status.Node.FAILED, null);
            return;
        }
        // 获取到结束节点，说明任务执行完
        List<TaskNode> endNodes = endNodeResult.getData();
        if (endNodes.isEmpty()) {
            // Boolean hasReady = nodeService.hasReady(instance.getId());
            // if (null != hasReady && !hasReady) {
            // handleFinish(Status.Node.SUCCESS, null);
            // }
            if (failed) {
                handleFinish(Status.Node.FAILED, null);
            } else if (waitingNode <= 0) {
                handleFinish(Status.Node.SUCCESS, null);
            }
            return;
        }
        runningNode += endNodes.size();
        for (TaskNode endNode : endNodes) {
            if (Category.Node.STOP.name().equalsIgnoreCase(endNode.getCategory())) {
                runningNode = runningNode - 1;
                waitingNode = waitingNode - 1;
                handleFinish(Status.Node.SUCCESS, endNode);
            } else {
                endNode.setStatus(Status.Node.RUNNING.name());
                nodeService.modify(endNode);
                updateRelation(instance.getId(), endNode.getId(), endNode.getNodeId(), Status.Node.RUNNING.name());
                endNode.setContext(getContext(instance.getContext()));
                endNode.setInstanceName(instance.getName());
                NodeExecutor nodeExecutor = new NodeExecutor(endNode, this, jobClient, preLinks, resultMap);
                this.nodeEexecutorPool.submit(nodeExecutor);
            }
        }
    }

    private RequestContext getContext(RequestContext context) {
        if (null == context) {
            return new RequestContext(RequestContext.Catalog.Manager, 1L, null, null, null);
        }
        return context;
    }

    /**
     * 取消任务
     */
    public void cancel() {
        if (cancel) {
            return;
        }
        cancel = true;
        suspendable = false;
        instance.setGmtModify(new Date());
        instance.setStatus(Status.Inst.CANCELING.name());
        instanceService.modify(instance);
        if (runningNode <= 0) {
            waitingNode = 0;
            handleFinish(Status.Node.CANCELED, null);
        }
        taskResult = new TaskResult(false, "task is canceled", instance.getUuid());
        sendMessage(null, taskResult);
    }

    /**
     * 暂停任务
     */
    public void suspend(TaskNode taskNode) {
        if (suspendable && null == taskNode) {
            return;
        }
        suspendable = true;
        if (null != taskNode) {
            if (!Status.Node.SUSPENDED.name().equals(taskNode.getStatus())) {
                // 说明此时为暂停某个任务节点
                log.info("更新任务【{}】的节点任务【{}】状态", instance.getName(), taskNode.getName());
                taskNode.setStatus(Status.Node.SUSPENDED.name());
                taskNode.setGmtModify(new Date());
                nodeService.modify(taskNode);
            }
            suspendingNodes.add(taskNode);
        }
        if (!Status.Node.SUSPENDED.name().equals(instance.getStatus())) {
            instance.setStatus(Status.Inst.SUSPENDED.name());
            instance.setGmtModify(new Date());
            instanceService.modify(instance);
        }
    }

    /**
     * 激活任务
     */
    public void active(Long nodeId) {
        if (suspendingNodes.isEmpty() && !suspendable) {
            return;
        }
        if (null != nodeId) {
            // 此时为激活单个任务节点
            TaskNode taskNode = nodeService.get(nodeId);
            log.info("激活任务节点【{}】", taskNode.getName());
            taskNode.setStatus(Status.Node.SUCCESS.name());
            taskNode.setGmtModify(new Date());
            nodeService.modify(taskNode);
            updateRelation(instance.getId(), nodeId, taskNode.getNodeId(), taskNode.getStatus());
            TaskNode suspendingNode = null;
            for (TaskNode node : suspendingNodes) {
                if (node.getId().equals(nodeId)) {
                    suspendingNode = node;
                    break;
                }
            }
            if (null != suspendingNode) {
                suspendingNodes.remove(suspendingNode);
            }
            if (suspendingNodes.isEmpty()) {
                suspendable = false;
                instance.setStatus(Status.Inst.RUNNING.name());
                instance.setGmtModify(new Date());
            }
        } else {
            // 此时为激活整个任务
            log.info("激活任务【{}】", instance.getName());
            instance.setStatus(Status.Inst.RUNNING.name());
            instance.setGmtModify(new Date());
            suspendable = false;
            for (TaskNode node : suspendingNodes) {
                node.setStatus(Status.Node.SUCCESS.name());
                node.setGmtModify(new Date());
                nodeService.modify(node);
                updateRelation(instance.getId(), node.getId(), node.getNodeId(), Status.Node.SUCCESS.name());
            }
            suspendingNodes = new ArrayList<>();
        }
        instanceService.modify(instance);
        synchronized (this) {
            execNext(null);
        }
    }

    /**
     * 处理定时任务
     */
    private void executeCronTask() {
        response = jobClient.submitJob(job);
        log.info("定时任务【】提交结果：【{}】", job.getParam(Common.NAME), response.getMessage());
        taskResult = new TaskResult(response.isSuccess(), response.getMessage(), job.getTaskId());
    }

    /**
     * 处理任务结束逻辑，可以同时处理任务结束的逻辑，也可以处理任务出现异常时候的各种逻辑
     */
    private void handleFinish(Status.Node status, TaskNode endNode) {
        // 处理节点
        if (null != endNode && Category.Node.STOP.name().equalsIgnoreCase(endNode.getCategory())) {
            modifyNode(endNode, status);
        }
        // 当还有节点还在运行或者没有在运行的节点，但是有等待运行的节点且没有出现失败
        if (runningNode > 0 || (waitingNode > 0 && !failed)) {
            return;
        }
        // 当已经做过结果处理和消息通知
        if (hasAction) {
            return;
        }
        hasAction = true;
        status = failed ? Status.Node.FAILED : status;
        // log.info("处理编排任务结束动作 {}", status.name());
        switch (status) {
            case SUCCESS:
                instance.setGmtFinish(new Date());
                instance.setStatus(status.name());
                instance.setGmtModify(new Date());
                taskResult = new TaskResult(true, "success",
                    new JobResult(true, JSONObject.toJSONString(new GeneralResult<>(true, "success")), job),
                    job.getTaskId());
                this.instanceService.modify(instance);
                runningNode = 0;
                log.info("任务【{}】执行完成 ", instance.getName());
                break;
            default:
                // 处理其他异常情况
                instance.setGmtFinish(new Date());
                instance.setStatus(status.name());
                instance.setGmtModify(new Date());
                taskResult = new TaskResult(false, "false",
                    new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "failed")), job),
                    job.getTaskId());
                this.instanceService.modify(instance);
                // 处理没有执行的节点
                // List<TaskNode> undoNodes = nodeService.getUndoneNodes(instance.getId());
                // if (!ListTool.allNull(undoNodes)) {
                // for (TaskNode undoNode : undoNodes) {
                // modifyNode(undoNode, status);
                // }
                // }
                break;
        }
        // 当定时任务执行完后，切回等待状态
        if (instance.getCode().contains(Catalog.Task.SCHEDULED.getValue())) {
            GeneralResult<?> result = CronUtils.isValid(instance.getCron());
            if (result.isFailed()) {
                this.taskService.updateStatus(instance.getTaskId(), Status.Inst.END.name(), instance.getCreatorId());
                scheduler.removeCronTask(job.getTaskId());
            } else {
                this.taskService.updateStatus(instance.getTaskId(), Status.Inst.WAITTING.name(),
                    instance.getCreatorId());
            }
            this.sendMessage();
        }
        // 说明没有正在执行的任务节点,此时才可以认为任务处理完成了
        handleNotify(status);
        this.finished = true;
    }

    /**
     * 处理任务结果通知逻辑
     */
    private void handleNotify(Status.Node status) {
        // log.info("处理任务执行结果消息发送逻辑...");
        String subject = "任务【" + instance.getName() + "】执行结束通知";
        String content = "任务【" + instance.getName() + "】：";
        switch (status) {
            case SUCCESS:
                content += "执行成功，具体情况请查看任务执行详细信息";
                break;
            case CANCELED:
                content += "被取消，具体情况请查看任务执行详细信息";
                break;
            case FAILED:
                content += "执行失败，具体情况请查看任务执行详细信息";
                break;
            default:
                break;
        }
        notice(subject, content);
    }

    /**
     * 任务执行完发送消息
     */
    private void notice(String subject, String content) {
        Boolean emailEnabled = instance.getEmailEnabled() == null ? false : instance.getEmailEnabled();
        Boolean notifyEnabled = Boolean.valueOf(redisTemplate.opsForValue().get("TaskNotifyEnabled"));
        JSONObject object = new JSONObject();
        object.put("context", getContext(instance.getContext()));
        object.put("subject", subject);
        object.put("content", content);
        object.put("sendMsg", true);
        object.put("sendEmail", emailEnabled && notifyEnabled);
        String exchange = BoCloudExchange.Default.DIRECT.getExchange();
        amqpTemplate.convertAndSend(exchange, "event.notice.queue", object);
        if (!StringUtils.isEmpty(job.getParam("QUEUE"))) {
            JSONObject result = new JSONObject();
            result.put("service", job.getParam("SERVICE"));
            List<TaskNode> nodes = nodeService.getNodes(instance.getId());
            instance.setNodes(nodes);
            result.put("instance", instance);
            // log.info("节点任务执行完发送消息到队列：{}, 消息内容：{}", job.getParam("QUEUE"), result);
            amqpTemplate.convertAndSend(job.getParam("QUEUE"), result);
        }
    }

    /**
     * 处理任务中断逻辑
     */
    private void handleSuspend() {
        while ((suspendable || suspendingNodes.size() > 0) && !cancel) {
            // 当前任务执行处于暂停状态，需要外界人工介入才可以继续执行下一步骤
            try {
                Thread.sleep(10L);
            } catch (InterruptedException e) {
                log.error("任务【{}]】暂停发生中断异常:", instance.getName(), e.getMessage());
                Thread.currentThread().interrupt();
            }
        }
        // log.info("所有任务节点暂停结束！任务继续执行...");
    }

    /**
     * 复杂任务执行器初始化
     */
    private void init() {
        if (!StringUtils.isEmpty(job.getParam(PJOB))) {
            // 处理定时任务回传过来的任务
            Task task = JSONObject.parseObject(job.getParam(DATA), Task.class);
            GeneralResult<Task> result = taskService.detail(task.getId());
            if (result.isFailed() || null == result.getData()) {
                log.warn("定时任务【{}】实例化失败，任务可能已经不存在", job.getParam(Common.NAME));
                return;
            }
            Task objTask = result.getData();
            if (!Status.Inst.WAITTING.name().equalsIgnoreCase(objTask.getStatus())) {
                log.warn("定时任务【{}】实例化失败，任务处在非等待状态", job.getParam(Common.NAME));
                return;
            }
            taskService.updateStatus(task.getId(), Status.Inst.RUNNING.name(), 1L);
            // 切回运行状态
            this.sendMessage();
            RequestContext context = new RequestContext();
            context.setTarget(task.getMenderId());
            context.setTenant(task.getTenantId());
            context.setCatalog(task.getCatalog());
            instance = taskService.initInstance(task.getId(), "CRON", context);
            if (null == instance) {
                log.warn("定时任务【{}】实例化失败", job.getParam(Common.NAME));
                return;
            }
            nodeSelector = new TaskNodeSelector(instance.getId(), nodeService);
            TaskCache.instance().putTaskExecutor(instance.getUuid(), this);
            instance.setStatus(Status.Inst.RUNNING.name());
            job.setTaskId(instance.getUuid());
        } else {
            if (Catalog.Task.SCHEDULED.getValue().equals(job.getParam(TYPE))) {
                return;
            }
            instance = JSONObject.parseObject(job.getParam(DATA), TaskInstance.class);
            this.nodeSelector = new TaskNodeSelector(instance.getId(), nodeService);
            if (instance.getStatus().equals(Status.Inst.SUSPENDED.name())) {
                // 处理任务容灾时容灾的任务为暂停状态。
                this.suspendable = true;
                List<TaskNode> tempNodes = nodeService.getNodes(instance.getId());
                for (TaskNode node : tempNodes) {
                    if (null == node.getStatus()) {
                        continue;
                    }
                    if (node.getStatus().equalsIgnoreCase(Status.Node.SUSPENDED.name())) {
                        suspendingNodes.add(node);
                    }
                }
            } else if (instance.getStatus().equals(Status.Inst.CANCELING.name())) {
                this.cancel = true;
            } else {
                instance.setStatus(Status.Inst.RUNNING.name());
            }
        }
        instance.setGmtModify(new Date());
        instance.setGmtStart(new Date());
        instance.setExecuteHost(service.getInstance().replaceAll(":", "_"));
        instanceService.modify(instance);
        // log.info("复杂任务【{}】执行器初始化完成！", instance.getName());
        TaskGraph graph = JSONObject.parseObject(instance.getGraph(), TaskGraph.class);
        ingoreNodes = NodeHelper.getIngoreNodes(graph.getNodes());
        preLinks = NodeHelper.getPreLinks(graph.getEdges(), ingoreNodes);
        nextEdges = NodeHelper.getNextEdges(graph.getEdges(), ingoreNodes);
        nodeMap = Maps.newHashMap();
        taskNodes = instance.getNodes();
        if (ListTool.isEmpty(taskNodes)) {
            taskNodes = nodeService.getNodes(instance.getId());
        }
        for (TaskNode node : taskNodes) {
            nodeMap.put(node.getNodeId(), node.getId());
            if (Status.Node.READY.name().equals(node.getStatus())) {
                waitingNode = waitingNode + 1;
            }
            if (Status.Node.SUCCESS.name().equals(node.getStatus())
                || Status.Node.FAILED.name().equals(node.getStatus())
                || Status.Node.SUSPENDED.name().equals(node.getStatus())) {
                JobResult jobResult = new JobResult();
                jobResult.setMsg(node.getOutput());
                TaskResult result = new TaskResult(!Status.Node.FAILED.name().equals(node.getStatus()), "", jobResult,
                    instance.getUuid(), node);
                resultMap.put(node.getNodeId(), result);
            }
        }
        nextNodes = NodeHelper.getNextNodes(graph.getEdges(), ingoreNodes, taskNodes);

        noEdges = ListTool.isEmpty(graph.getEdges());
        single = taskNodes.size() == 1;
        monolayer = nextNodes.size() == 1;
    }

    private void modifyNode(TaskNode taskNode, Status.Node status) {
        taskNode.setCost(0.0);
        taskNode.setGmtStart(new Date());
        taskNode.setGmtFinish(new Date());
        taskNode.setGmtModify(new Date());
        taskNode.setStatus(status.name());
        switch (status) {
            case SUCCESS:
                taskNode.setOutput(JSONObject.toJSONString(new GeneralResult<>(true, "执行成功！")));
                break;
            case CANCELED:
                taskNode.setOutput(JSONObject.toJSONString(new GeneralResult<>(false, "Task is canceled")));
                break;
            case FAILED:
                taskNode.setOutput(JSONObject.toJSONString(new GeneralResult<>(false, "因其他节点执行失败，导致该节点未能成功执行")));
                break;
            default:
                taskNode.setOutput(JSONObject.toJSONString(new GeneralResult<>(false, status.name())));
                break;
        }
        this.nodeService.modify(taskNode);
        this.updateRelation(instance.getId(), taskNode.getId(), taskNode.getNodeId(), status.name());
    }

    private void updateRelation(Long instanceId, Long source, String nodeId, String status) {
        if (noEdges || !nextEdges.containsKey(nodeId)) {
            return;
        }
        this.nodeService.updateRelation(instanceId, source, status);
    }

    /**
     * 处理异常信息
     */
    private void handleException() {
        JobResult jobResult = new JobResult(false, JSONObject.toJSONString(new GeneralResult<>(false, "任务类型错误")), job);
        taskResult = new TaskResult(false, "任务类型错误", jobResult, job.getTaskId());
        handleFinish();
    }

    /**
     * 处理整个任务结束之后的逻辑
     */
    private void handleFinish() {
        this.scheduler.onCompleted(taskResult);
        this.instanceService.analysis(instance.getId());
    }

    private void sendMessage(TaskNode node, TaskResult result) {
        // 发送结果消息
        TaskMessage taskMessage = null;
        if (null == node) {
            taskMessage = new TaskMessage(instance.getId(), null, instance.getCreatorId(), result.getMessage(), null,
                result.isSuccess(), suspendable, cancel);
        } else {
            taskMessage = new TaskMessage(instance.getId(), node.getId(), instance.getCreatorId(), result.getMessage(),
                node.getNodeId(), result.isSuccess(), suspendable, cancel);
        }
        RequestContext context = instance.getContext();
        if ((null != context && RequestContext.Catalog.Manager == context.getCatalog())) {
            OperateResult operateResult = new OperateResult(result.isSuccess(), "", null, taskMessage,
                OperateResult.OperateCategory.TaskResult, context);
            publisher.publish(operateResult);
        }
    }

    private void sendMessage() {
        // 定时任务完成后发送消息
        OperateResult result =
            new OperateResult(true, "", "task.cron", new TaskMessage(), OperateResult.OperateCategory.TaskResult, null);
        publisher.publish(result);
    }

    /**
     * 构造方法
     */
    public TaskExecutor(Job job, TaskProgressPublisher publisher, JobClient jobClient, TaskService taskService,
        TaskNodeService nodeService, TaskInstanceService instanceService, JobScheduler scheduler,
        ExecutorService nodeEexecutorPool, AmqpTemplate amqpTemplate, CurrentService service,
        NodeModuleService moduleService, StringRedisTemplate redisTemplate) {
        super();
        this.job = job;
        this.publisher = publisher;
        this.jobClient = jobClient;
        this.taskService = taskService;
        this.nodeService = nodeService;
        this.scheduler = scheduler;
        this.instanceService = instanceService;
        this.nodeEexecutorPool = nodeEexecutorPool;
        this.amqpTemplate = amqpTemplate;
        this.service = service;
        this.redisTemplate = redisTemplate;
    }
}
