package com.bocloud.tse.scheduler.failover;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.springframework.amqp.core.AmqpTemplate;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.cmp.boot.model.BoCloudService;
import com.bocloud.tse.entity.TaskInstance;
import com.bocloud.tse.entity.TaskNode;
import com.bocloud.tse.enums.Status;
import com.bocloud.tse.intf.TaskInstanceService;
import com.bocloud.tse.intf.TaskNodeService;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.utils.ListTool;
import com.megatron.framework.core.CurrentService;
import com.megatron.framework.core.RegistryService;
import com.megatron.framework.core.Service;

import lombok.extern.slf4j.Slf4j;

/**
 * 任务容灾处理逻辑，用来在服务启动之后自动处理异常宕机的任务容灾
 */
@Slf4j
public class TaskFailoverExecutor {

    /**
     * 是否为服务本身的容灾
     */
    private boolean selfFailover;
    private CurrentService service;
    private TaskNodeService nodeService;
    private RegistryService registryService;
    private TaskInstanceService instanceService;
    private List<String> serviceInstances = null;
    private AmqpTemplate amqpTemplate;

    public void execute() {
        // 获取执行中的且执行主机为当前服务实例的任务
        String serviceHost = service.getInstance().replace(":", "_");
        String executeHost = null;
        String excludeHost = null;
        if (selfFailover) {
            executeHost = serviceHost;
        } else {
            excludeHost = serviceHost;
        }
        fetchServices();
        GeneralResult<List<TaskInstance>> instanceResult =
            instanceService.listExceptionInstance(executeHost, excludeHost);
        if (instanceResult.isFailed() || ListTool.allNull(instanceResult.getData())) {
            if (log.isDebugEnabled()) {
                log.debug("无异常任务，容灾结束！");
            }
            return;
        }
        for (TaskInstance instance : instanceResult.getData()) {
            if (!selfFailover && serviceInstances.contains(instance.getExecuteHost())) {
                log.info("任务【{}】 正在被调度引擎【{}】执行，跳过此记录", instance.getName(), instance.getExecuteHost());
                continue;
            }
            instance.setStatus(Status.Inst.FAILED.name());
            instance.setGmtFinish(new Date());
            instance.setGmtModify(new Date());
            instanceService.modify(instance);
            List<TaskNode> nodes = nodeService.getNodes(instance.getId());
            for (TaskNode node : nodes) {
                if (Status.Node.RUNNING.name().equals(node.getStatus())) {
                    node.setStatus(Status.Node.FAILED.name());
                    node.setOutput(JSONObject.toJSONString(new GeneralResult<>(false, "因服务异常挂掉，导致该节点未能成功执行")));
                    node.setGmtFinish(new Date());
                    node.setGmtModify(new Date());
                    if (null == node.getGmtStart()) {
                        node.setCost(0.0);
                        node.setGmtStart(new Date());
                    } else {
                        BigDecimal cost = new BigDecimal(node.getGmtFinish().getTime() - node.getGmtStart().getTime())
                            .divide(new BigDecimal(1000), 3, RoundingMode.HALF_UP);
                        node.setCost(cost.doubleValue());
                    }
                    nodeService.modify(node);
                    nodeService.updateRelation(instance.getId(), node.getId(), Status.Node.FAILED.name());
                }
            }
            JSONObject props = JSONObject.parseObject(instance.getProps());
            if (null != props && props.containsKey("queue")) {
                JSONObject result = new JSONObject();
                result.put("service", props.getString("service"));
                instance.setNodes(nodes);
                result.put("instance", instance);
                amqpTemplate.convertAndSend(props.getString("queue"), result);
            } else if ("compliance".equalsIgnoreCase(instance.getType())
                || "inspection".equalsIgnoreCase(instance.getType()) || "patch".equalsIgnoreCase(instance.getType())) {
                JSONObject result = new JSONObject();
                instance.setNodes(nodes);
                result.put("instance", instance);
                amqpTemplate.convertAndSend("task.job." + instance.getType() + ".result", result);
            }
        }
    }

    /**
     * 获取当前的所有bmp服务实例
     */
    private void fetchServices() {
        serviceInstances = new ArrayList<>();
        List<String> services = registryService.getServices(Service.create(BoCloudService.TSE.getName()));
        if (ListTool.allNull(services)) {
            return;
        }
        for (String service : services) {
            String temp = service.replace("http://", "").replace(":", "_");
            if (!serviceInstances.contains(temp)) {
                serviceInstances.add(temp);
            }
        }
    }

    public void setSelfFailover(boolean selfFailover) {
        this.selfFailover = selfFailover;
    }

    public TaskFailoverExecutor(boolean selfFailover, CurrentService service, TaskNodeService nodeService,
        TaskInstanceService instanceService, RegistryService registryService, AmqpTemplate amqpTemplate) {
        super();
        this.selfFailover = selfFailover;
        this.service = service;
        this.nodeService = nodeService;
        this.instanceService = instanceService;
        this.registryService = registryService;
        this.amqpTemplate = amqpTemplate;
    }

}
