package com.bocloud.tse.scheduler.failover;

import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.entity.Task;
import com.bocloud.tse.entity.TaskInstance;
import com.bocloud.tse.enums.Catalog;
import com.bocloud.tse.enums.TaskParam;
import com.bocloud.tse.intf.TaskInstanceService;
import com.bocloud.tse.intf.TaskNodeService;
import com.bocloud.tse.intf.TaskService;
import com.bocloud.tse.model.Job;
import com.bocloud.tse.scheduler.jobclient.JobClient;
import com.bocloud.tse.scheduler.jobclient.JobSchedulerHandler;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.utils.Common;
import com.megatron.common.utils.ListTool;
import com.megatron.framework.core.CurrentService;
import com.megatron.framework.core.RegistryService;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

/**
 * 任务容灾处理定时器
 */
@Slf4j
@Component
public class TaskFailoverScheduler implements InitializingBean {

    @Autowired
    private TaskInstanceService instanceService;
    @Autowired
    private TaskNodeService nodeService;
    @Autowired
    private TaskService taskService;
    @Autowired
    private CurrentService service;
    @Autowired
    private RegistryService registryService;
    @Autowired
    private LockFactory lockFactory;
    @Autowired
    private AmqpTemplate amqpTemplate;
    @Autowired
    private JobSchedulerHandler jobSchedulerHandler;
    @Autowired
    private JobClient jobClient;
    private TaskFailoverExecutor failoverExecutor = null;

    /**
     * 每隔两分钟执行一次容灾处理 ：每两分钟执行一次
     */
    @Scheduled(cron = "0 0/2 * * * ?")
    protected void execute() throws Exception {
        if (service.isLeader()) {
            this.taskFailover();
        }
    }

    /**
     * 每隔5两分钟执行一次定时任务处理
     */
    @Scheduled(cron = "0 0/5 * * * ?")
    protected void executeSchedule() throws Exception {
        if (service.isLeader()) {
            this.taskScheduleFailover();
        }
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        failoverExecutor =
            new TaskFailoverExecutor(true, service, nodeService, instanceService, registryService, amqpTemplate);
        this.taskFailover();
        failoverExecutor.setSelfFailover(false);
    }

    private void taskFailover() {
        String path = "TASK_FAILOVER_LOCK";
        try (AutoCloseLock lock = lockFactory.getACLock(path)) {
            if (lock.acquire(10, TimeUnit.SECONDS)) {
                failoverExecutor.execute();
            } else {
                log.warn("Get DLock timeout!");
            }
        } catch (Exception e) {
            log.error("任务容灾异常：", e);
        }
    }

    private void taskScheduleFailover() {
        GeneralResult<List<Task>> result = taskService.listSchedule();
        List<Task> tasks = result.getData();
        if (ListTool.isEmpty(tasks)) {
            return;
        }
        for (Task task : tasks) {
            if ("WAITTING".equals(task.getStatus())) {
                continue;
            }
            List<TaskInstance> instances = instanceService.list(task.getId());
            if (ListTool.isEmpty(instances)) {
                taskService.updateStatus(task.getId(), "WAITTING", 1L);
                continue;
            }
            boolean hasRun = false;
            for (TaskInstance instance : instances) {
                if ("RUNNING".equalsIgnoreCase(instance.getStatus())
                    || "SUSPENDED".equalsIgnoreCase(instance.getStatus())) {
                    hasRun = true;
                }
            }
            if (!hasRun) {
                taskService.updateStatus(task.getId(), "WAITTING", 1L);
            }
        }
        for (Task task : tasks) {
            if (jobSchedulerHandler.hasTask(task.getUuid())) {
                continue;
            }
            Job job = buildJob(task, task.getUuid(), task.getName(), Catalog.Task.SCHEDULED.getValue(), task.getCode(),
                task.getIsLoop(), task.getRunTime());
            job.setCronExpression(task.getCron());
            jobClient.submitJob(job);
        }

    }

    private Job buildJob(Object task, String taskId, String taskName, String type, String code, Boolean isLoop,
        Date startTime) {
        Job job = new Job();
        job.setParam(TaskParam.TASKDATA.name(), JSONObject.toJSONString(task));
        job.setParam(TaskParam.TASKTYPE.name(), type);
        job.setTaskId(taskId);
        job.setParam(Common.NAME, taskName);
        job.setParam(TaskParam.TASKCODE.name(), code);
        job.setLoop(isLoop);
        job.setStartTime(startTime);
        return job;
    }

}
