package com.bocloud.tse.scheduler.jobclient;

import java.util.Date;
import java.util.concurrent.TimeUnit;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.config.TriggerTask;
import org.springframework.scheduling.support.CronTrigger;
import org.springframework.scheduling.support.SimpleTriggerContext;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.entity.Task;
import com.bocloud.tse.enums.Catalog;
import com.bocloud.tse.enums.Status;
import com.bocloud.tse.enums.TaskParam;
import com.bocloud.tse.intf.TaskService;
import com.bocloud.tse.model.Job;
import com.bocloud.tse.scheduler.worker.listener.TaskJobListener;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Result;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@Component
public class JobClient {

    // private static final String queue = "task.job.queue";
    @Autowired
    private TaskService taskService;
    @Autowired
    private JobSchedulerHandler jobSchedulerHandler;
    @Autowired
    private LockFactory lockFactory;
    @Autowired
    private TaskJobListener jobListener;

    public Result submitJob(Job job) {
        if (null == job) {
            return new Result(false, "Job参数有误");
        }
        if (Catalog.Task.SCHEDULED.getValue().equalsIgnoreCase(job.getParam(TaskParam.TASKTYPE.name()))) {
            if (job.isLoop()) {
                log.info("添加循环定时任务 -> {}", job.getTaskId());
                jobSchedulerHandler.addTriggerTask(job.getTaskId(),
                    new TriggerTask(() -> submit(job), new CronTrigger(job.getCronExpression())),
                    job.getCronExpression());
            } else {
                log.info("添加一次性定时任务 -> {}", job.getTaskId());
                jobSchedulerHandler.addTimeTask(job.getTaskId(), () -> submit(job), job.getStartTime());
            }
        } else {
            //标准任务处理
            jobListener.submit(job);
        }
        return new Result(true, "Job提交成功");
    }

    public Result cancelJob(String taskId) {
        jobSchedulerHandler.cancelTriggerTask(taskId);
        return new Result(true, "定时任务移除成功");
    }

    private void submit(Job job) {
        String path = Job.class.getSimpleName() + "_" + job.getTaskId();
        try (AutoCloseLock lock = lockFactory.getACLock(path)) {
            Assert.isTrue(lock.acquire(1, TimeUnit.SECONDS), "作业已提交");
            GeneralResult<?> result = taskService.get(job.getTaskId());
            if (result.isFailed() || null == result.getData()) {
                log.info("作业【" + job.getTaskId() + "】已经不存在，移除定时作业");
                cancelJob(job.getTaskId());
                return;
            }
            Task task = (Task)result.getData();
            if (Status.Inst.SUSPENDED.name().equalsIgnoreCase(task.getStatus())) {
                log.info("作业【" + task.getName() + "】已处于挂起状态，移除定时作业");
                cancelJob(job.getTaskId());
                return;
            }
            Date date = null;
            if (job.isLoop()) {
                CronTrigger trigger = new CronTrigger(job.getCronExpression());
                date = trigger.nextExecutionTime(new SimpleTriggerContext());
                if (null == date
                    || (null != task.getGmtExecute() && date.getTime() <= task.getGmtExecute().getTime())) {
                    log.info("作业【" + task.getName() + "】已处于执行中，忽略这次执行");
                    return;
                }
            } else if (null != task.getGmtExecute()) {
                // 一次性任务已经执行过，不用再次执行
                log.info("作业【" + task.getName() + "】已执行结束，忽略这次执行");
                return;
            }
            result = taskService.updateExecute(task.getId(), date == null ? new Date() : date);
            if (result.isFailed()) {
                return;
            }
            Job newJob = JSONObject.parseObject(JSONObject.toJSONString(job), Job.class);
            jobListener.submit(newJob);
            Thread.sleep(3 * 1000L);
        } catch (Exception e) {
            log.info("submit task fail:", e);
        }
    }

}
