package com.bocloud.tse.scheduler.jobclient;

import java.util.Date;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;

import org.springframework.scheduling.SchedulingException;
import org.springframework.scheduling.TaskScheduler;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.scheduling.annotation.SchedulingConfigurer;
import org.springframework.scheduling.config.ScheduledTaskRegistrar;
import org.springframework.scheduling.config.TriggerTask;
import org.springframework.stereotype.Component;

import com.bocloud.tse.scheduler.utils.BeanUtils;

import lombok.extern.slf4j.Slf4j;

/**
 * 平台定时任务调度类
 */
@Slf4j
@Component
@EnableScheduling
public class JobSchedulerHandler implements SchedulingConfigurer {

    private final String FIELD_SCHEDULED_FUTURES = "scheduledTasks";// "scheduledFutures";
    private ScheduledTaskRegistrar taskRegistrar;
    private Set<ScheduledFuture<?>> scheduledFutures = null;
    private Map<String, ScheduledFuture<?>> taskFutures = new ConcurrentHashMap<String, ScheduledFuture<?>>();
    private Map<String, String> taskCrons = new ConcurrentHashMap<String, String>();

    @Override
    public void configureTasks(ScheduledTaskRegistrar taskRegistrar) {
        this.taskRegistrar = taskRegistrar;
    }

    @SuppressWarnings("unchecked")
    private Set<ScheduledFuture<?>> getScheduledFutures() {
        if (scheduledFutures == null) {
            try {
                scheduledFutures =
                    (Set<ScheduledFuture<?>>)BeanUtils.getProperty(taskRegistrar, FIELD_SCHEDULED_FUTURES);
            } catch (NoSuchFieldException e) {
                throw new SchedulingException("not found scheduledFutures field.");
            }
        }
        return scheduledFutures;
    }

    /**
     * 添加任务
     */
    public void addTriggerTask(String taskId, TriggerTask triggerTask, String cron) {
        if (taskFutures.containsKey(taskId)) {
            throw new SchedulingException("the taskId[" + taskId + "] was added.");
        }
        TaskScheduler scheduler = taskRegistrar.getScheduler();
        if (null == scheduler) {
            throw new SchedulingException("the task scheduler is null");
        }
        ScheduledFuture<?> future = scheduler.schedule(triggerTask.getRunnable(), triggerTask.getTrigger());
        getScheduledFutures().add(future);
        taskFutures.put(taskId, future);
        taskCrons.put(taskId, cron);
        log.info("添加周期性定时执行任务成功");
    }

    public void addTimeTask(String taskId, Runnable task, Date startTiem) {
        if (taskFutures.containsKey(taskId)) {
            throw new SchedulingException("the taskId[" + taskId + "] was added.");
        }
        TaskScheduler scheduler = taskRegistrar.getScheduler();
        if (null == scheduler) {
            throw new SchedulingException("the task scheduler is null");
        }
        ScheduledFuture<?> future = scheduler.schedule(task, startTiem);
        getScheduledFutures().add(future);
        taskFutures.put(taskId, future);
        // taskCrons.put(taskId, cron);
        log.info("添加一次性定时执行任务成功");
    }

    /**
     * 取消任务
     */
    public void cancelTriggerTask(String taskId) {
        if (!hasTask(taskId)) {
            return;
        }
        ScheduledFuture<?> future = taskFutures.get(taskId);
        if (future != null) {
            future.cancel(true);
        }
        getScheduledFutures().remove(future);
        taskFutures.remove(taskId);
        taskCrons.remove(taskId);
    }

    /**
     * 重置任务
     */
    public void resetTriggerTask(String taskId, TriggerTask triggerTask, String cron) {
        cancelTriggerTask(taskId);
        addTriggerTask(taskId, triggerTask, cron);
    }

    /**
     * 任务编号
     */
    public Set<String> taskIds() {
        return taskFutures.keySet();
    }

    /**
     * 任务编号
     */
    public String getTaskCron(String taskId) {
        return taskCrons.get(taskId);
    }

    /**
     * 清空当前定时全部定时任务
     */
    public void clearTasks() {
        for (String taskId : taskIds()) {
            this.cancelTriggerTask(taskId);
        }
        taskFutures.clear();
        taskCrons.clear();
    }

    /**
     * 是否存在任务
     */
    public boolean hasTask(String taskId) {
        return this.taskFutures.containsKey(taskId);
    }

    /**
     * 任务调度是否已经初始化完成
     */
    public boolean inited() {
        return this.taskRegistrar != null && this.taskRegistrar.getScheduler() != null;
    }
}
