package com.bocloud.tse.scheduler.listener;

import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.core.AmqpTemplate;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.core.MessageListener;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.model.TaskActionMessage;
import com.bocloud.tse.scheduler.core.JobScheduler;
import com.bocloud.tse.scheduler.core.TaskCache;
import com.megatron.common.model.Result;

import lombok.extern.slf4j.Slf4j;

/**
 * 任务动作监听器，用来监听rabbitMQ上的topic.task.action.${server.id}队列上的实时消息
 */
@Slf4j
@Repository("taskActionListener")
public class TaskActionListener implements MessageListener {

    @Autowired
    private JobScheduler scheduler;
    @Autowired
    private AmqpTemplate amqpTemplate;

    @Override
    public void onMessage(Message message) {
        String content = new String(message.getBody());
        log.info("Get message: {}", content);
        if (!StringUtils.isEmpty(content)) {
            // 消息内容必须有值
            try {
                TaskActionMessage actionMessage = JSONObject.parseObject(content, TaskActionMessage.class);
                // 要求message的三个字段必须有值
                if (null == actionMessage.getAction() || StringUtils.isEmpty(actionMessage.getQueue())
                    || StringUtils.isEmpty(actionMessage.getTaskId())) {
                    log.warn("Task action message is not in right format");
                } else {
                    handleAction(actionMessage);
                }
            } catch (Exception e) {
                log.error("Task Action Message Convert error:", e);
            }
        }
    }

    private void handleAction(TaskActionMessage message) {
        if (TaskCache.instance().hasTaskExecutor(message.getTaskId())) {
            Result taskActionResult = null;
            switch (message.getAction()) {
                case ACTIVE:
                    taskActionResult = scheduler.activeTask(message.getTaskId(), message.getNodeId());
                    break;
                case SUSPEND:
                    taskActionResult = scheduler.suspendTask(message.getTaskId());
                    break;
                case CANCEL:
                    taskActionResult = scheduler.cancelTask(message.getTaskId());
                    break;
                default:
                    log.warn(" Action {} not supported!", message.getAction());
                    break;
            }
            log.info("Task action {} result:{}", message.getAction(), JSONObject.toJSONString(taskActionResult));
            if (null != taskActionResult) {
                sendMessage(message.getQueue(), taskActionResult);
            }
        } else {
            log.info("Task [{}] is not in this server!ignore the message", message.getTaskId());
        }
    }

    private void sendMessage(String queue, Object message) {
        this.amqpTemplate.convertAndSend(queue, message);
    }
}
