package com.bocloud.tse.scheduler.utils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.entity.TaskNode;
import com.bocloud.tse.enums.Category;
import com.bocloud.tse.model.TaskResult;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Result;
import com.megatron.common.utils.ListTool;

/**
 * 节点之间依赖处理类
 */
public class NodeDependencyHandler {

    private TaskNode node;
    private Map<String, TaskResult> resultMap;
    private List<String> preNodes;
    private JSONObject input;
    private JSONObject fronts;

    public NodeDependencyHandler() {
        super();
    }

    public NodeDependencyHandler(TaskNode node, Map<String, TaskResult> resultMap) {
        super();
        this.node = node;
        this.resultMap = resultMap;
        init();
    }

    public NodeDependencyHandler(TaskNode node, Map<String, TaskResult> resultMap, List<String> preNodes) {
        super();
        this.node = node;
        this.resultMap = resultMap;
        this.preNodes = preNodes;
        init();
    }

    /**
     * 初始化
     */
    private void init() {
        // log.info("获取任务节点【{}】前置依赖节点输入信息【{}】", node.getUuid(), node.getInput());
        input = JSONObject.parseObject(node.getInput());
        if (null == input) {
            input = new JSONObject();
        }
        fronts = input.getJSONObject("fronts");
        if (null == fronts) {
            fronts = new JSONObject();
        }
        fronts.put("context", node.getContext());
    }

    /**
     * 依赖处理入口，处理资源类和软件类节点输入输出依赖
     */
    public void handle() {
        if (ListTool.isEmpty(preNodes)) {
            input.put("fronts", fronts);
            node.setParams(input.toJSONString());
            return;
        }
        if (node.getModuleType().contains(Category.Node.MAINTAIN.getValue())) {
            // 处理运维依赖关系
            handleMaintain();
        }
        // 参数依赖处理完后，任务节点执行处理所需要的全部参数放入params当中
        node.setParams(input.toJSONString());
    }

    /**
     * 得到判断节点的判断结果
     */
    public Result getJudgeResult() {
        try {
            List<JudgeCondition> conditions =
                JSONArray.parseArray(fronts.getString("conditions"), JudgeCondition.class);
            if (ListTool.isEmpty(conditions)) {
                return new Result(false, "false");
            }
            Collections.sort(conditions, new Comparator<JudgeCondition>() {
                public int compare(JudgeCondition first, JudgeCondition second) {
                    if (null == first || second == null) {
                        return 0;
                    }
                    return first.getOrder().compareTo(second.getOrder());
                }
            });
            boolean preResult = false;
            boolean lastResult = false;
            String preLink = null;
            int index = 0;
            for (JudgeCondition condition : conditions) {
                boolean currentResult = false;
                TaskResult taskResult = resultMap.get(condition.getId());
                GeneralResult<?> result =
                    JSONObject.parseObject(taskResult.getJobResult().getMsg(), GeneralResult.class);
                if ("result".equalsIgnoreCase(condition.getSource())) {
                    if (result.isSuccess() == Boolean.valueOf(condition.getValue())) {
                        currentResult = true;
                    }
                } else {
                    JSONObject data = JSONObject.parseObject(JSONObject.toJSONString(result.getData()));
                    currentResult = calculate(data, condition);
                }
                if (index == 0) {
                    if ("OR".equalsIgnoreCase(condition.getLink())) {
                        if (currentResult) {
                            lastResult = true;
                            break;
                        }
                    } else {
                        preResult = currentResult;
                        preLink = condition.getLink();
                    }
                } else {
                    if (null == preLink) {
                        if ("OR".equalsIgnoreCase(condition.getLink())) {
                            if (currentResult) {
                                lastResult = true;
                                break;
                            }
                        } else {
                            preResult = currentResult;
                            preLink = condition.getLink();
                        }
                    } else {
                        if ("OR".equalsIgnoreCase(condition.getLink())) {
                            if (preResult && currentResult) {
                                lastResult = true;
                                break;
                            }
                            preResult = false;
                            preLink = null;
                        } else {
                            preResult = preResult && currentResult ? true : false;
                            preLink = condition.getLink();
                        }
                    }
                }
                index++;
                if (index == conditions.size()) {
                    lastResult = preResult;
                    break;
                }
            }
            return new Result(true, String.valueOf(lastResult));
        } catch (Exception e) {
            return new Result(false, "false");
        }
    }

    private boolean calculate(JSONObject data, JudgeCondition condition) {
        // TODO 处理后期判断节点的前置节点输出内容的比较和运算
        return false;
    }

    /**
     * 处理运维类任务节点的目标机器依赖
     */
    private void handleMaintain() {
        handScriptResultDepend();
    }

    /**
     * 处理当前节点对前置脚本节点输出的依赖
     */
    private void handScriptResultDepend() {
        if (StringUtils.isEmpty(fronts.getString("params"))) {
            return;
        }
        List<JSONObject> params = JSONObject.parseArray(fronts.getString("params"), JSONObject.class);
        for (JSONObject param : params) {
            if ("PREOUTPUT".equalsIgnoreCase(param.getString("source"))) {
                // 参数解析替换
                String paramKey = param.getString("paramKey");
                String dependNodeId = param.getString("nodeId");
                TaskResult result = resultMap.get(dependNodeId);
                if (result.isFailed()) {
                    throw new RuntimeException("获取前置节点输出结果失败，前置节点执行异常！");
                }
                Map<String, String> msgMap = parasLog(result.getJobResult().getMsg());
                param.put("value", msgMap.get(paramKey));
            }
        }
        fronts.put("params", params);
        input.put("fronts", fronts);
    }

    /**
     * 将输出的日志解析并返回脚本日志的格式化数据
     */
    private Map<String, String> parasLog(String logString) throws RuntimeException {
        Map<String, String> logMap = new HashMap<>();
        JSONObject resultjson = JSONObject.parseObject(logString);
        if (resultjson.getBoolean("failed")) {
            throw new RuntimeException("脚本执行结果解析失败：前置节点返回执行失败" + resultjson);
        }
        JSONArray resultdata = resultjson.getJSONArray("data");
        if (resultdata.isEmpty()) {
            throw new RuntimeException("脚本执行结果解析失败：前置节点返回执行失败" + resultjson);
        }
        JSONObject object = JSONObject.parseObject(JSON.toJSONString(resultdata.get(0)));
        String message = object.getString("message");
        String[] logs = message.split("[\n]|[\r]");
        for (String log : logs) {
            if (log.trim().length() == 0) {
                continue;
            }
            int colonIndex = log.indexOf(":");
            if (colonIndex < 0) {
                continue;
            }
            String key = log.substring(0, colonIndex);
            switch (key) {
                case "script_run_status":
                    String script_run_status = log.substring(colonIndex + 1).trim();
                    if (null == script_run_status) {
                        throw new RuntimeException("脚本执行结果解析失败：script_run_status不能为\"" + script_run_status + "\"");
                    }
                    logMap.put("status", script_run_status);
                    break;
                case "inspect_result":
                    String inspect_result = log.substring(colonIndex + 1).trim();
                    logMap.put("result", inspect_result);
                    Map<String, String> resultKv = getResultKv(inspect_result);
                    logMap.putAll(resultKv);
                    break;
                default:
                    break;
            }
        }

        if (logMap.get("status") == null) {
            throw new RuntimeException("脚本执行结果解析失败：script_run_status不能为空");
        }
        if ((logMap.get("status").equals("0") || logMap.get("status").equals("2"))) {
            if (logMap.get("result") == null) {
                throw new RuntimeException("脚本执行结果解析失败：script_run_status为0或2时，inspect_result不能为空");
            }
        }
        return logMap;
    }

    /**
     * 解析result
     */
    private Map<String, String> getResultKv(String scriptResult) {
        try {
            Map<String, String> resultMap = new HashMap<>();
            List<String> messageGroups = new ArrayList<>();
            String[] messages = scriptResult.split("[|][|][|][|]");
            for (String s : messages) {
                messageGroups.add(s);
            }
            for (String result : messageGroups) {
                String key = "";
                String value = "";
                String[] param = result.split("====");
                if (param.length < 2) { // 对应的值为空
                    key = param[0];
                } else {
                    key = param[0];
                    value = param[1];// 脚本执行结果的值
                }
                resultMap.put(key, value);
            }
            return resultMap;
        } catch (Exception e) {
            throw new RuntimeException("脚本执行结果解析失败：scriptResult不能为" + scriptResult);
        }
    }

}