package com.bocloud.tse.scheduler.worker;

import java.util.*;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import com.bocloud.tse.constant.CacheConstant;
import com.bocloud.tse.constant.ExecuteTypeConstant;
import com.bocloud.tse.core.exception.NodeFormatException;
import com.bocloud.tse.model.HostResult;
import com.bocloud.tse.model.SaltCacheModel;
import com.bocloud.tse.model.SaltModel;
import com.bocloud.tse.model.WorkerHost;
import com.bocloud.tse.scheduler.core.DataCache;
import com.bocloud.tse.scheduler.worker.executor.RemoteSaltStack;
import com.bocloud.tse.scheduler.worker.executor.RemoteWorker;
import com.bocloud.tse.scheduler.worker.utils.ResultMessageHandler;
import com.megatron.common.utils.IDFactory;
import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.rabbit.core.RabbitAdmin;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.util.Assert;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.cmp.boot.model.BoCloudService;
import com.megatron.common.encrypt.AESEncryptor;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.utils.ListTool;
import com.megatron.common.utils.MapTools;
import com.megatron.framework.core.Service;
import com.megatron.framework.http.core.HttpMethod;
import com.megatron.framework.http.model.RemoteService;
import com.megatron.framework.http.model.ServiceParam;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

/**
 * 远程Worker调用任务分发器，用来对worker请求进行参数校验和请求分片调度。
 */
@Slf4j
public class WorkerDispatcher {

    private static final String URL = "/api/gaia/v1/job/submit";
    private static final Service SERVICE = Service.create(BoCloudService.GAIA.getName());
    private Map<String, Object> headers = MapTools.simpleMap("Content-Type", "application/json");
    private ExecutorService executor = null;
    private CompletionService<GeneralResult<?>> completionService = null;
    private String code = null;

    private RabbitAdmin rabbitAdmin;
    private LockFactory lockFactory;
    private StringRedisTemplate redisTemplate;
    private boolean fixed;

    public WorkerDispatcher() {
        super();
    }

    public WorkerDispatcher(RabbitAdmin rabbitAdmin, StringRedisTemplate redisTemplate, LockFactory lockFactory,
        String code) {
        this.rabbitAdmin = rabbitAdmin;
        this.redisTemplate = redisTemplate;
        this.lockFactory = lockFactory;
        this.code = code;
        this.fixed = getResultQueueFixed();
    }

    /**
     * worker任务分发执行
     */
    public GeneralResult<?> execute(List<Map<String, Object>> params, Integer timeout, String nodeId, String instance)
        throws Exception {
        log.info("start to dispatch task! node job uuid is [{}]", nodeId);
        if (null == params || params.isEmpty()) {
            log.error("remote worker param is empty!");
            throw new NodeFormatException("worker param is illegel!");
        }
        String executeType = String.valueOf(params.get(0).get("executeType"));
        List<String> batchNos = new ArrayList<>();
        Map<String, List<WorkerHost>> targetMap = new HashMap<>();
        for (Map<String, Object> param : params) {
            if (param.containsKey("noworkergroup")) {
                log.error("exists hosts without groups");
                return GeneralResult.FAILED("存在没有对应分组worker的机器");
            }
            String key = null;
            if (ExecuteTypeConstant.SALTSTACK.equalsIgnoreCase(executeType)) {
                // 已经包含前缀了
                key = String.valueOf(param.get("saltId"));
            } else {
                key = CacheConstant.WORKERGROUP_ + param.get("groupId");
            }
            if (param.containsKey("groupId") || param.containsKey("saltId")){
                String value = redisTemplate.opsForValue().get(key);
                if (StringUtils.isEmpty(value)) {
                    log.error("[" + key + "] don't have some workers to execute");
                    return GeneralResult.FAILED("不存在对应执行worker");
                }
            }
            param.put("fixed", fixed);
            if (fixed) {
                String uuid = IDFactory.instance().uuid();
                batchNos.add(uuid);
                targetMap.put(uuid, WorkerHost.parseTargets(JSONObject.toJSONString(param.get("targets"))));
                param.put("batch_no", uuid);
                param.put("fixedQueue", "task.result.queue." + instance);
            }
        }
        List<HostResult> hostResults = new ArrayList<>();
        int corePoolSize = params.size() + 1;
        executor = new ThreadPoolExecutor(corePoolSize, corePoolSize, 30, TimeUnit.SECONDS,
            new LinkedBlockingQueue<>(corePoolSize), (Runnable runnable) -> {
                return new Thread(runnable);
            });
        completionService = new ExecutorCompletionService<>(executor);
        for (Map<String, Object> param : params) {
            if (ExecuteTypeConstant.SALTSTACK.equals(executeType)) {
                this.submitSalt(param, timeout, nodeId);
            } else {
                //执行任务
                this.submitAnsible(param, timeout, nodeId);
            }
        }
        int finished = 0;
        int workerNum = params.size();
        // 处理任务执行的结果
        GeneralResult<?> result = null;
        boolean hasError = false;
        if (fixed && !ExecuteTypeConstant.SALTSTACK.equals(executeType)) {
            // 等待worker固定任务结果队列结果汇总处理
            Long startTime = System.currentTimeMillis();
            Long useTime = 0L;
            while (finished < workerNum && useTime < timeout * 1000) {
                Iterator<String> iterator = batchNos.iterator();
                while (iterator.hasNext()) {
                    String batchNo = iterator.next();
                    String content = DataCache.get(batchNo);
                    if (null == content) {
                        continue;
                    }
                    // 移除缓存
                    DataCache.remove(batchNo);
                    iterator.remove();
                    // 处理任务结果
                    try {
                        result = ResultMessageHandler.handleResult(JSONObject.parseObject(content, GeneralResult.class),
                            instance, targetMap.get(batchNo));
                        if (result.isFailed()) {
                            // 单次任务请求处理失败！
                            hasError = true;
                        }
                    } catch (Exception e) {
                        hasError = true;
                        log.error("callback data transfer exception:", e);
                        result = ResultMessageHandler.handleError("返回数据格式异常!", instance, targetMap.get(batchNo));
                    }
                    hostResults.addAll(buildHostResult(result));
                    finished++;
                }
                if (finished < workerNum) {
                    this.sleep(300L);
                }
                // 获取结果时间
                useTime = System.currentTimeMillis() - startTime;
            }
            // 超时未完成
            if (finished < workerNum) {
                hasError = true;
                for (String batchNo : batchNos) {
                    result = ResultMessageHandler.handleError("获取任务结果消息超时!", instance, targetMap.get(batchNo));
                    hostResults.addAll(buildHostResult(result));
                }
            }
        } else {
            while (finished < workerNum) {
                try {
                    result = completionService.take().get();
                    if (result.isFailed()) {
                        // 单次任务请求处理失败！
                        hasError = true;
                    }
                    hostResults.addAll(buildHostResult(result));
                } catch (Exception e) {
                    log.error("单次请求worker任务执行异常", e);
                    hasError = true;
                }
                finished++;
            }
        }
        // TODO 这里需要处理失败重试机制，后续优化
        if (null != executor) {
            executor.shutdown();
        }
        if (hasError) {
            log.error("node job [{}] execute failed", nodeId);
            if (ListTool.isEmpty(hostResults)) {
                hostResults.addAll(buildHostResult(params, false, "任务执行失败"));
            }
            return new GeneralResult<>(false, hostResults, "任务执行失败");
        }
        // 没有失败任务，直接返回处理结果
        log.info("node job [{}] execute success", nodeId);
        return new GeneralResult<>(true, hostResults, "任务执行成功！");
    }

    private void submitSalt(Map<String, Object> param, Integer timeout, String nodeId) throws Exception {
        String errorMessage = null;
        String key = String.valueOf(param.get("saltId"));
        SaltModel instance = getSaltInstance(key);
        if (null == instance) {
            log.error("get salt worker redis cache error!");
            errorMessage = "获取执行worker失败，可能未配置可执行的worker";
            instance = new SaltModel();
        }
        // 对密码进行解密
        instance.setPswd(new AESEncryptor().decrypt(instance.getPswd(), null));
        RemoteSaltStack task = new RemoteSaltStack(
            instance.getProtocol() + "://" + instance.getHost() + ":" + instance.getPort(), instance.getUser(),
            instance.getPswd(), param, timeout, errorMessage, nodeId, instance.getHost() + ":" + instance.getPort(),getTaskSaltMinionQueueLimit());
        completionService.submit(task);
    }

    private void submitAnsible(Map<String, Object> param, Integer timeout, String nodeId) throws Exception {
        String instance = null;
        String errorMessage = null;
        RemoteWorker task = null;
        if ("task.maintain.hostallow".equalsIgnoreCase(code) || "task.maintain.gaia".equalsIgnoreCase(code)) {
            instance = WorkerLoadPolicy.getInstance(CacheConstant.SUPER_WORKERS, redisTemplate, lockFactory);
            if (null == instance) {
                log.error("get super worker redis cache error!");
                errorMessage = "获取执行worker失败，可能未配置可执行的超级worker";
            }
        } else {
            param.put("node_id", nodeId);
            if (param.containsKey("groupId")) {
                String key = CacheConstant.WORKERGROUP_ + param.get("groupId").toString();
                //根据配置的策略获取对应的worker实例
                instance = WorkerLoadPolicy.getInstance(key, redisTemplate, lockFactory);
            }
            if (null == instance) {
                log.error("get worker redis cache error!");
                errorMessage = "获取执行worker失败，可能未配置可执行的worker";
            }
        }
        if ("RABBITMQ".equalsIgnoreCase(getExecutionMethod())) {
            param.put("instance", null != instance ? (instance.indexOf(":") > -1 ? instance.split(":")[0] : instance) : null);
            // 通过mq发送
            task = new RemoteWorker(new ServiceParam(param, headers), rabbitAdmin, timeout, null, errorMessage, nodeId,
                instance, fixed);
        } else {
       String url = "http://" + instance + URL;
        RemoteService service =
            new RemoteService(SERVICE, new ServiceParam(param, headers), true, url, HttpMethod.OPERATE);
         task = new RemoteWorker(service, rabbitAdmin, timeout, null, false, errorMessage, nodeId, instance, fixed);
        }
        completionService.submit(task);
    }

    private SaltModel getSaltInstance(String key) {
        try (AutoCloseLock lock = lockFactory.getACLock(key)) {
            String value = redisTemplate.opsForValue().get(key);
            SaltCacheModel loadInfo = JSONObject.parseObject(value, SaltCacheModel.class);
            if (null == loadInfo || ListTool.isEmpty(loadInfo.getSaltModels())) {
                log.error("获取worker负载缓存失败:");
                return null;
            }
            Assert.isTrue(lock.acquire(300, TimeUnit.SECONDS), "请求超时");
            List<SaltModel> workers = loadInfo.getSaltModels();
            int position = loadInfo.getPosition();
            if (position >= workers.size()) {
                position = 0;
            }
            SaltModel instance = workers.get(position);
            position++;
            loadInfo.setPosition(position);
            redisTemplate.opsForValue().set(key, JSONObject.toJSONString(loadInfo));
            return instance;
        } catch (Exception e) {
            log.error("获取worker负载缓存失败:", e);
            return null;
        }
    }

    private List<HostResult> buildHostResult(GeneralResult<?> result) {
        List<HostResult> hostResults = new ArrayList<>();
        if (null != result.getData()) {
            try {
                hostResults = JSONArray.parseArray(JSONObject.toJSONString(result.getData()), HostResult.class);
            } catch (Exception e) {
                log.error("build host result error!", e);
            }
        }
        return hostResults;
    }

    private List<HostResult> buildHostResult(List<Map<String, Object>> params, Boolean success, String message) {
        List<HostResult> hostResults = new ArrayList<>();
        for (Map<String, Object> param : params) {
            if (null == param.get("targets")) {
                continue;
            }
            try {
                List<HostResult> results =
                    JSONArray.parseArray(JSONObject.toJSONString(param.get("targets")), HostResult.class);
                for (HostResult result : results) {
                    result.setSuccess(success);
                    result.setMessage(message);
                }
                hostResults.addAll(results);
            } catch (Exception e) {
                log.error("build host result error!", e);
            }
        }
        return hostResults;
    }

    /**
     * 获取salt minion执行任务时排队的上限
     *
     * @return worker执行结果返回
     */
    private Integer getTaskSaltMinionQueueLimit() {
        String limit = redisTemplate.opsForValue().get("TaskSaltMinionQueueLimit");
        return StringUtils.isNotBlank(limit) ? Integer.parseInt(limit) : null;
    }
    /**
     * 获取worker执行方式（目前仅支持RABBITMQ，HTTP）
     * 
     * @return worker执行方式
     */
    private String getExecutionMethod() {
        return redisTemplate.opsForValue().get("TaskWorkDistributeMethod");
    }

    /**
     * 获取worker执行结果返回的方式,默认是通过临时队列返回
     * 
     * @return worker执行结果返回
     */
    private boolean getResultQueueFixed() {
        String fixed = redisTemplate.opsForValue().get("TaskWorkResultQueueFixed");
        return StringUtils.isNotBlank(fixed) && "true".equalsIgnoreCase(fixed.trim());
    }

    private void sleep(long time) {
        try {
            Thread.sleep(time);
        } catch (InterruptedException e) {
            log.error("Thread [{}] InterruptedException", Thread.currentThread().getName());
        }
    }}
