package com.bocloud.tse.scheduler.worker;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.util.Assert;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.model.WorkerLoadInfo;
import com.bocloud.tse.model.WorkerMetricInfo;
import com.bocloud.tse.model.WorkerWeightInfo;
import com.megatron.common.utils.ListTool;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

/**
 * worker负载策略工具类
 */
@Slf4j
public class WorkerLoadPolicy {

    public static String getInstance(String key, StringRedisTemplate template, LockFactory factory) {
        try {
            //value = {"excludeWorkers":[],"position":0,"workers":["10.20.12.65:18091"]}
            String value = template.opsForValue().get(key);
            WorkerLoadInfo loadInfo = JSONObject.parseObject(value, WorkerLoadInfo.class);
            if (null == loadInfo || ListTool.isEmpty(loadInfo.getWorkers())) {
                log.error("获取worker负载缓存失败:");
                return null;
            }
            if (1 > loadInfo.getWorkers().size() - loadInfo.getExcludeWorkers().size()) {
                log.error("获取worker负载缓存失败:被排除的worker数量已过载！");
                return null;
            }
            LoadPolicy policy = getPolicy(template.opsForValue().get("TaskWokerLoadPolicy"));
            String instance = null;
            switch (policy) {
                case Random:
                    instance = random(loadInfo, key);
                    break;
                case Dispersed:
                    // 因为zk上worker当前任务数不是实时的，需要防止任务在zk更新前，全部落到同一个worker
                    // 设置任务数差值，只有worker之间任务数差值满足条件才返回实例
                    instance = dispersed(key, template, 5, factory);
                    break;
                case Weight:
                    instance = weight(loadInfo, template, key);
                    break;
                default:
                    instance = polling(key, template, factory);
                    break;
            }
            return instance;
        } catch (Exception e) {
            log.error("获取worker负载缓存失败:", e);
            return null;
        }
    }

    /**
     * 按照轮询规则获取worker实例
     */
    private static String polling(String key, StringRedisTemplate template, LockFactory factory) {
        try (AutoCloseLock lock = factory.getACLock(key)) {
            Assert.isTrue(lock.acquire(300, TimeUnit.SECONDS), "请求超时");
            WorkerLoadInfo info = JSONObject.parseObject(template.opsForValue().get(key), WorkerLoadInfo.class);
            List<String> workers = getWorkers(info, key);
            int position = info.getPosition();
            if (position >= workers.size()) {
                position = 0;
            }
            String instance = workers.get(position);
            position++;
            info.setPosition(position);
            template.opsForValue().set(key, JSONObject.toJSONString(info));
            return instance;
        } catch (Exception e) {
            log.error("获取worker负载缓存失败:", e);
            return null;
        }
    }

    /**
     * 按照分散规则获取worker实例
     */
    private static String dispersed(String key, StringRedisTemplate template, int diff, LockFactory factory) {
        try (AutoCloseLock lock = factory.getACLock(key)) {
            Assert.isTrue(lock.acquire(300, TimeUnit.SECONDS), "请求超时");
            WorkerLoadInfo info = JSONObject.parseObject(template.opsForValue().get(key), WorkerLoadInfo.class);
            List<String> workers = getWorkers(info, key);
            List<WorkerMetricInfo> metrics = new ArrayList<>();
            workers.stream().forEach(worker -> {
                String value = template.opsForValue().get("workmetric_" + worker);
                log.info("worker【{}】执行任务数【{}】", worker, value);
                if (StringUtils.isEmpty(value) || "null".equals(value)) {
                    value = "0";
                }
                metrics.add(new WorkerMetricInfo(worker, Integer.parseInt(value)));
            });
            if (metrics.isEmpty()) {
                return null;
            }
            List<WorkerMetricInfo> list = metrics.stream()
                .sorted(Comparator.comparingInt(WorkerMetricInfo::getTaskNumber)).collect(Collectors.toList());
            // 判断worker当前任务数相差最大的差值，满足条件后才走分散规则
            WorkerMetricInfo firstMetric = list.get(0);
            WorkerMetricInfo lastMetric = list.get(list.size() - 1);
            if (lastMetric.getTaskNumber() - firstMetric.getTaskNumber() >= diff) {
                return firstMetric.getInstance();
            }
            log.info("worker差值小于【{}】,不满足分散规则，走轮询策略", diff);
            int position = info.getPosition();
            if (position >= workers.size()) {
                position = 0;
            }
            String instance = workers.get(position);
            position++;
            info.setPosition(position);
            template.opsForValue().set(key, JSONObject.toJSONString(info));
            return instance;
        } catch (Exception e) {
            log.error("获取worker负载缓存失败:", e);
            return null;
        }
    }

    /**
     * 根据权重规则获取worker实例
     */
    private static String weight(WorkerLoadInfo loadInfo, StringRedisTemplate template, String key) {
        List<String> workers = getWorkers(loadInfo, key);
        AtomicInteger totalWeight = new AtomicInteger();
        List<WorkerWeightInfo> weights = new ArrayList<>();
        workers.stream().forEach(worker -> {
            // 获取权重信息
            String value = template.opsForValue().get("workweight_" + worker);
            log.info("worker【{}】权重【{}】", worker, value);
            if (null == value || "null".equals(value)) {
                return;
            }
            totalWeight.addAndGet(Integer.parseInt(value));
            weights.add(new WorkerWeightInfo(worker, totalWeight.get()));
        });
        return weightSection(new Random().nextInt(totalWeight.get()), weights);
    }

    /**
     * 根据权重区间获取worker实例
     */
    private static String weightSection(int random, List<WorkerWeightInfo> weights) {
        for (int i = 0; i < weights.size(); i++) {
            if (rangeInDefined(random, weights.get(i).getWeight(), weights.get(i + 1).getWeight())) {
                return weights.get(i).getInstance();
            }
        }
        return null;
    }

    /**
     * 判断某个值是否在某个区间内(区间左闭右开,包含所在的起始值和结束值)
     */
    private static boolean rangeInDefined(int current, int min, int max) {
        return current != max && Math.max(min, current) == Math.min(current, max);
    }

    /**
     * 通过随机规则获取worker实例
     */
    private static String random(WorkerLoadInfo info, String key) {
        List<String> workers = getWorkers(info, key);
        return workers.get(new Random().nextInt(workers.size()));
    }

    /**
     * 过滤排除的worker列表
     *
     * @param info
     * @param key
     * @return
     */
    private static List<String> getWorkers(WorkerLoadInfo info, String key) {
        List<String> workers = info.getWorkers();
        List<String> excludeWorkers = info.getExcludeWorkers();
        if (excludeWorkers.isEmpty()) {
            return info.getWorkers();
        }
        Map<String, String> excludeMap =
            excludeWorkers.stream().collect(Collectors.toMap(Function.identity(), Function.identity()));
        return workers.stream().filter(worker -> !excludeMap.containsKey(worker)).collect(Collectors.toList());
    }

    private static LoadPolicy getPolicy(String value) {
        if (StringUtils.isEmpty(value)) {
            return LoadPolicy.Polling;
        }
        for (LoadPolicy policy : LoadPolicy.values()) {
            if (policy.name().equals(value)) {
                return policy;
            }
        }
        return LoadPolicy.Polling;
    }

    enum LoadPolicy {
        Random, Polling, Dispersed, Weight
    }

}
