package com.bocloud.tse.scheduler.worker.executor;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.TypeReference;
import com.bocloud.tse.core.domain.WorkerScript;
import com.bocloud.tse.core.domain.args.maintain.FileArg;
import com.bocloud.tse.core.domain.args.software.*;
import com.bocloud.tse.core.module.Module;
import com.bocloud.tse.model.HostResult;
import com.bocloud.tse.model.StateApplyResultModel;
import com.bocloud.tse.model.WorkerHost;
import com.bocloud.tse.scheduler.worker.utils.SaltClientCache;
import com.bocloud.tse.scheduler.worker.utils.SaltClientModel;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.reflect.TypeToken;
import com.megatron.common.encrypt.AESEncryptor;
import com.megatron.common.encrypt.Encryptor;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.utils.MapTools;
import com.suse.salt.netapi.AuthModule;
import com.suse.salt.netapi.calls.LocalAsyncResult;
import com.suse.salt.netapi.calls.LocalCall;
import com.suse.salt.netapi.calls.modules.State;
import com.suse.salt.netapi.client.AsyncHttpClient;
import com.suse.salt.netapi.client.SaltClient;
import com.suse.salt.netapi.client.impl.HttpAsyncClientImpl;
import com.suse.salt.netapi.datatypes.AuthMethod;
import com.suse.salt.netapi.datatypes.PasswordAuth;
import com.suse.salt.netapi.datatypes.Token;
import com.suse.salt.netapi.datatypes.target.MinionList;
import com.suse.salt.netapi.errors.SaltError;
import com.suse.salt.netapi.parser.JsonParser;
import com.suse.salt.netapi.results.Result;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.client.config.CookieSpecs;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.conn.ssl.NoopHostnameVerifier;
import org.apache.http.conn.ssl.TrustAllStrategy;
import org.apache.http.impl.nio.client.CloseableHttpAsyncClient;
import org.apache.http.impl.nio.client.HttpAsyncClients;
import org.apache.http.ssl.SSLContexts;
import org.springframework.util.CollectionUtils;

import javax.net.ssl.SSLContext;
import java.net.URI;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.util.*;
import java.util.concurrent.Callable;

@Slf4j
public class RemoteSaltStack implements Callable<GeneralResult<?>> {

    private static final Integer DEFAULT_TIMEOUT = 1000 * 60 * 10;

    private static final Integer DEFAULT_QUEUE_LIMIT = 20;
    private AsyncHttpClient httpClient;
    private String server;
    private SaltClient client;
    private AuthMethod method;
    private Map<String, Object> params;
    private Integer timeout;
    private String user;
    private String pswd;
    private String errorMessage;
    private String nodeId;
    private List<WorkerHost> targets;
    private String instance;

    private Integer queueLimit;

    public RemoteSaltStack(String server, String user, String pswd, Map<String, Object> params, Integer timeout,
                           String errorMessage, String nodeId, String instance, Integer queueLimit) {
        this.server = server;
        this.params = params;
        if (null == timeout) {
            this.timeout = DEFAULT_TIMEOUT;
        } else {
            this.timeout = timeout * 1000;
        }
        this.user = user;
        this.pswd = pswd;
        this.errorMessage = errorMessage;
        this.nodeId = nodeId;
        this.instance = instance;
        this.queueLimit = queueLimit;
        init();
    }

    @Override
    public GeneralResult<?> call() throws Exception {
        if (StringUtils.isNotEmpty(errorMessage)) {
            log.error("task execute fail:" + errorMessage);
            return handleError(errorMessage);
        }
        try {
            log.info("start node job [{}] call remote worker [{}],timeout is [{}]ms", nodeId, server, timeout);
            if (params.containsKey("script")) {
                return this.scriptBySls();
            }
            Module<?> module = JSON.parseObject(JSON.toJSONString(params.get("module")), Module.class);
            if ("bocloud_copy".equals(module.getName())) {
                return this.copyBySls();
            }
            if ("install_tomcat".equals(module.getName())) {
                return this.installTomcatBySls();
            }
            if ("install_java".equals(module.getName())) {
                return this.installJDK();
            }
            if ("install_jenkins".equals(module.getName())) {
                return this.installJenkins();
            }
            if ("install_apache".equals(module.getName())) {
                return this.installApacheBySls();
            }
            if ("install_gitlab".equals(module.getName())) {
                return this.installGitlabBySls();
            }
            if ("install_nginx".equals(module.getName())) {
                return this.installNginxBySls();
            }
            return new GeneralResult<>(false, "Unsupported");
        } catch (Exception e) {
            log.error("remote salt stack execute error:", e);
            return new GeneralResult<>(false, "Unsupported");
        }
    }

    /**
     * 安装、配置 nginx
     *
     * @return
     */
    private GeneralResult<?> installNginxBySls() {
        boolean success = true;
        List<Map<String, Object>> hostResults = new ArrayList<>();
        Map<WorkerHost, String> minions = new HashMap<>();
        targets.forEach(target -> minions.put(target, target.getSaltMinionId()));
        if (!hostResults.isEmpty()) {
            success = false;
        }
        Module<NginxArg> module = JSON.parseObject(JSON.toJSONString(params.get("module")), new TypeReference<Module<NginxArg>>() {
        });
        NginxArg arg = module.getArgs();
        Map<String, Object> params = new HashMap<>();
        params.put("nginx", JSONObject.parseObject(JSONObject.toJSONString(arg)));
        Optional<Map<String, Object>> pillar = Optional.of(params);
        List<String> mods = new ArrayList<>();
        mods.add("nginx.install_nginx");
        long start = System.currentTimeMillis();
        //mods：对应sls文件 ， pillar：执行状态文件需要的数据
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务install nginx结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务install nginx结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取install nginx结果失败");
        }
        Map<WorkerHost, String> sucMinions = new HashMap<>();
        success = processJobResults(hostResults, cmdResults, minions, start, success, null);
        log.info("nginx.install_nginx::::::" + success);
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }

    private GeneralResult<?> handleError(String message) {
        List<HostResult> hostResults = new ArrayList<>();
        for (WorkerHost target : targets) {
            hostResults.add(new HostResult(false, message, target.getHost(), target.getRawName(), target.getPort(),
                    target.getUser(), instance));
        }
        return new GeneralResult<>(false, hostResults, message);
    }

    /**
     * 通过state执行脚本
     *
     * @return 执行结果
     */
    private GeneralResult<?> scriptBySls() {
        boolean success = true;
        WorkerScript script = JSON.parseObject(JSON.toJSONString(params.get("script")), WorkerScript.class);
        List<Map<String, Object>> hostResults = new ArrayList<>();
        Map<WorkerHost, String> minions = new HashMap<>();
        targets.forEach(target -> minions.put(target, target.getSaltMinionId()));
        if (!hostResults.isEmpty()) {
            success = false;
        }
        String contents = script.getContent();
        Map<String, Object> params = MapTools.simpleMap("type", "STATE");
        contents = processContents(contents);
        // 脚本处理
        log.info("contents:{}", contents);
        params.put("contents", contents);
        if ("SHELL".equalsIgnoreCase(script.getType())) {
            params.put("name", "script_" + UUID.randomUUID() + ".sh");
        } else if ("PYTHON".equalsIgnoreCase(script.getType())) {
            params.put("name", "script_" + UUID.randomUUID() + ".py");
        } else if ("BAT".equalsIgnoreCase(script.getType())) {
            params.put("name", "script_" + UUID.randomUUID() + ".bat");
        } else if ("PERL".equalsIgnoreCase(script.getType())) {
            params.put("name", "script_" + UUID.randomUUID() + ".pl");
        } else if ("PLAYBOOK".equalsIgnoreCase(script.getType())) {
            params.put("name", "script_" + UUID.randomUUID() + ".yml");
        } else if ("POWERSHELL".equalsIgnoreCase(script.getType())) {
            params.put("name", "script_" + UUID.randomUUID() + ".ps1");
        }

        params.put("params", script.getParams());
        Optional<Map<String, Object>> pillar = Optional.of(params);
        List<String> mods = new ArrayList<>();
        mods.add("script");

        long start = System.currentTimeMillis();
        //mods：对应sls文件 ， pillar：执行状态文件需要的数据
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务script结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务script结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取script结果失败");
        }
        Map<WorkerHost, String> sucMinions = new HashMap<>();
        success = processJobResults(hostResults, cmdResults, minions, start, success, null);
        log.info("script::::::" + success);
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }

    private void init() {
        if (null != params) {
            this.targets = JSON.parseArray(JSON.toJSONString(params.get("targets")), WorkerHost.class);
        }
        this.method = new AuthMethod(new PasswordAuth(user, pswd, AuthModule.PAM));
        SaltClientModel model = SaltClientCache.getInstance().getClient(server);
        if (null == model) {
            SSLContext sslContext;
            try {
                sslContext = SSLContexts.custom().loadTrustMaterial(null, TrustAllStrategy.INSTANCE).build();
            } catch (NoSuchAlgorithmException | KeyManagementException | KeyStoreException e) {
                log.warn("Custom SSLContext Error: {}. Use Default SSLContext", e.getMessage(), e);
                sslContext = SSLContexts.createDefault();
            }

            RequestConfig requestConfig = RequestConfig.custom().setConnectionRequestTimeout(10000)
                    .setConnectTimeout(10000).setSocketTimeout(20 * 60 * 1000).setCookieSpec(CookieSpecs.STANDARD).build();
            CloseableHttpAsyncClient httpAsyncClient = HttpAsyncClients.custom().setDefaultRequestConfig(requestConfig)
                    .setSSLContext(sslContext).setSSLHostnameVerifier(NoopHostnameVerifier.INSTANCE).setMaxConnTotal(100)
                    .setMaxConnPerRoute(100).build();
            httpAsyncClient.start();

            this.httpClient = new HttpAsyncClientImpl(httpAsyncClient);
            this.client = new SaltClient(URI.create(server), httpClient);
            SaltClientCache.getInstance().update(server, new SaltClientModel(client, httpClient));
        } else {
            this.httpClient = model.getHttpClient();
            this.client = model.getClient();
        }
    }

    private static String processContents(String contents) {
        List<String> list = new ArrayList<>();
        list.add(contents);
        String str = JSONObject.toJSONString(list);
        // 需要的字符串是以双引号开头的字符串，截掉两边的[]
        return str.substring(1, str.length() - 1);
    }

    private GeneralResult<?> copyBySls() {
        boolean success = true;
        List<Map<String, Object>> hostResults = new ArrayList<>();
        Map<WorkerHost, String> minions = new HashMap<>();
        targets.forEach(target -> minions.put(target, target.getSaltMinionId()));
        if (!hostResults.isEmpty()) {
            success = false;
        }
        Module<FileArg> module =
                JSON.parseObject(JSON.toJSONString(params.get("module")), new TypeReference<Module<FileArg>>() {
                });
        String[] srcFiles = module.getArgs().getSrc().split(",");
        Map<String, Object> params = new HashMap<>();
        String destPath = module.getArgs().getDest();
        params.put("dest", destPath);
        List<Map<String, Object>> file = new ArrayList<>();
        for (String src : srcFiles) {
            Map<String, Object> tempParams = new HashMap<>();
            tempParams.put("src", "salt://" + src);
            tempParams.put("fileName", src.substring(src.lastIndexOf("/") + 1));
            file.add(tempParams);
        }
        params.put("files", file);
        //pillar:{"files":[{"fileName":"/opt/test/test.txt","src":"salt:///tmp/20230612/974aaf5ddfc94e44ad6d6316e53b37de/test.txt"}],"dest":"/opt/test"}
        Optional<Map<String, Object>> pillar = Optional.of(params);
        List<String> mods = new ArrayList<>();
        mods.add("file");
        // mods.add(module.getName());
        long start = System.currentTimeMillis();
        //mods：对应sls文件 ， pillar：执行状态文件需要的数据
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务copy结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务copy结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取copy结果失败");
        }
        Map<WorkerHost, String> sucMinions = new HashMap<>();
        success = processJobResults(hostResults, cmdResults, minions, start, success, null);
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }

    private GeneralResult<?> installTomcatBySls() {
        boolean success = true;
        List<Map<String, Object>> hostResults = new ArrayList<>();
        Map<WorkerHost, String> minions = new HashMap<>();
        targets.forEach(target -> minions.put(target, target.getSaltMinionId()));
        if (!hostResults.isEmpty()) {
            success = false;
        }
        Module<TomcatArg> module =
                JSON.parseObject(JSON.toJSONString(params.get("module")), new TypeReference<Module<TomcatArg>>() {
                });
        TomcatArg arg = module.getArgs();
        Map<String, Object> params = new HashMap<>();
        params.put("args", JSONObject.parseObject(JSONObject.toJSONString(arg)));
        Optional<Map<String, Object>> pillar = Optional.of(params);
        List<String> mods = new ArrayList<>();
        mods.add("tomcat.install_tomcat");
        long start = System.currentTimeMillis();
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务install tomcat结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务install tomcat结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取install tomcat结果失败");
        }
        success = processJobResults(hostResults, cmdResults, minions, start, success, null);
        log.info("end install tomcat targets:{} , args:{} , success:{}", targets, arg, success);
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }

    private GeneralResult<?> installApacheBySls() {
        boolean success = true;
        List<Map<String, Object>> hostResults = new ArrayList<>();
        Map<WorkerHost, String> minions = new HashMap<>();
        targets.forEach(target -> minions.put(target, target.getSaltMinionId()));
        if (!hostResults.isEmpty()) {
            success = false;
        }
        Module<ApacheArg> module =
                JSON.parseObject(JSON.toJSONString(params.get("module")), new TypeReference<Module<ApacheArg>>() {
                });
        ApacheArg arg = module.getArgs();
        log.info("start install apache targets:{} args:{}", targets, arg);
        Map<String, Object> params = new HashMap<>();
        params.put("args", JSONObject.parseObject(JSONObject.toJSONString(arg)));
        Optional<Map<String, Object>> pillar = Optional.of(params);
        List<String> mods = new ArrayList<>();
        mods.add("apache.install_apache");
        long start = System.currentTimeMillis();
        Map<WorkerHost, String> sucMinions = new HashMap<>();
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务install apache结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务install apache结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取install jdk结果失败");
        }
        success = processJobResults(hostResults, cmdResults, minions, start, success, sucMinions);
        if (!CollectionUtils.isEmpty(sucMinions)) {
            //apache安装成功的，配置apache
            GeneralResult<List<Map<String, Object>>> result = configApacheBySls(pillar, sucMinions, start);
            hostResults.addAll(result.getData());
            if (success && result.isFailed()) {
                success = false;
            }
        }
        log.info("end install apache targets:{}  , args:{} , success:{}", targets, arg, success);
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }

    private GeneralResult<?> installGitlabBySls() {
        boolean success = true;
        List<Map<String, Object>> hostResults = new ArrayList<>();
        Map<WorkerHost, String> minions = new HashMap<>();
        targets.forEach(target -> minions.put(target, target.getSaltMinionId()));
        if (!hostResults.isEmpty()) {
            success = false;
        }
        Module<GitlabArg> module =
                JSON.parseObject(JSON.toJSONString(params.get("module")), new TypeReference<Module<GitlabArg>>() {
                });
        GitlabArg arg = module.getArgs();
        log.info("start install gitlab targets:{} args:{}", targets, arg);
        //密码解密
        if (StringUtils.isNotEmpty(arg.getGitlabRootPassword())) {
            Encryptor encryptor = new AESEncryptor();
            arg.setGitlabRootPassword(encryptor.decrypt(arg.getGitlabRootPassword(), null));
        }
        Map<String, Object> params = new HashMap<>();
        params.put("args", JSONObject.parseObject(JSONObject.toJSONString(arg)));
        Optional<Map<String, Object>> pillar = Optional.of(params);
        List<String> mods = new ArrayList<>();
        mods.add("gitlab.install_gitlab");
        long start = System.currentTimeMillis();
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务install gitlab结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务install gitlab结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取install gitlab结果失败");
        }
        success = processJobResults(hostResults, cmdResults, minions, start, success, null);
        log.info("end install gitlab targets:{} , args:{} , success:{}", targets, arg, success);
        return new GeneralResult<>(success, hostResults, "Finished the job");

    }

    public Map<String, Map<String, JSONObject>> getJobs(String jid) {
        // 获取minions
        Token token = client.login(user, pswd, AuthModule.PAM).toCompletableFuture().join();
        URI uri = URI.create(server).resolve("/jobs/" + jid);
        Map<String, Object> response =
                httpClient.get(uri, Collections.singletonMap("X-Auth-Token", token.getToken()), JsonParser.MAP)
                        .toCompletableFuture().join();
        log.info("response---------------:{}", JSON.toJSONString(response));
        Map<String, Map<String, JSONObject>> result = JSON.parseObject(JSONObject.toJSONString(response.get("return")), new TypeReference<Map<String, Map<String, JSONObject>>>() {
        });
        return result;

    }

    /**
     * 配置apache
     *
     * @param pillar     参数
     * @param sucMinions 执行的minions
     * @param start      执行的开始时间
     * @return 操作结果
     */
    private GeneralResult<List<Map<String, Object>>> configApacheBySls(Optional<Map<String, Object>> pillar, Map<WorkerHost, String> sucMinions, Long start) {
        boolean success = true;
        List<Map<String, Object>> hostResults = new ArrayList<>();
        List<String> mods = new ArrayList<>();
        mods.add("apache.config_apache");
        log.info("start config apache targets:{}", targets);
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(sucMinions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务config apache结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务config apache结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取config apache结果失败");
        }
        success = processJobResults(hostResults, cmdResults, sucMinions, start, success, null);
        log.info("end config apache targets:{} , success:{}", targets, success);
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }


    /**
     * 安装jdk
     *
     * @return 操作结果
     */
    private GeneralResult<?> installJDK() {
        Map<WorkerHost, String> minions = new HashMap<>();
        targets.forEach(target -> minions.put(target, target.getSaltMinionId()));
        Module<JdkArg> module =
                JSON.parseObject(JSON.toJSONString(params.get("module")), new TypeReference<Module<JdkArg>>() {
                });
        JdkArg arg = module.getArgs();
        Map<String, Object> params = new HashMap<>();
        params.put("args", JSONObject.parseObject(JSONObject.toJSONString(arg)));
        Optional<Map<String, Object>> pillar = Optional.of(params);
        return installJDKBySls(pillar, minions, System.currentTimeMillis(), null, null);
    }

    /**
     * 安装Jenkins
     *
     * @return 操作结果
     */
    private GeneralResult<?> installJenkins() {
        Map<WorkerHost, String> sucMinions = new HashMap<>();
        List<Map<String, Object>> hostResults = new ArrayList<>();
        long start = System.currentTimeMillis();
        //构建jdk的参数安装jdk
        boolean success = true;
        Map<WorkerHost, String> minions = new HashMap<>();
        targets.forEach(target -> minions.put(target, target.getSaltMinionId()));
        Map<String, Object> jdkArg = new HashMap(16);
        jdkArg.put("version", "8u162");
        jdkArg.put("set_java_home", true);
        jdkArg.put("java_depended", true);
        Map<String, Object> jdkParams = new HashMap<>();
        jdkParams.put("args", JSONObject.parseObject(JSONObject.toJSONString(jdkArg)));
        Optional<Map<String, Object>> jdkPillar = Optional.of(jdkParams);
        if (installJDKBySls(jdkPillar, minions, start, hostResults, sucMinions).isFailed()) {
            //部分成功也是失败
            success = false;
        }
        //如果有执行成功的服务器，这部分服务器安装jenkins
        if (!CollectionUtils.isEmpty(sucMinions)) {
            Module<JenkinsArg> module = JSON.parseObject(JSON.toJSONString(params.get("module")), new TypeReference<Module<JenkinsArg>>() {
            });
            JenkinsArg jenkinsArg = module.getArgs();

            //密码解密
            if (StringUtils.isNotEmpty(jenkinsArg.getJenkinsAdminPassword())) {
                Encryptor encryptor = new AESEncryptor();
                jenkinsArg.setJenkinsAdminPassword(encryptor.decrypt(jenkinsArg.getJenkinsAdminPassword(), null));
            }

            Map<String, Object> jenkinsParams = new HashMap<>();
            jenkinsParams.put("args", JSONObject.parseObject(JSONObject.toJSONString(jenkinsArg)));
            Optional<Map<String, Object>> jenkinsPillar = Optional.of(jenkinsParams);
            minions.clear();
            minions.putAll(sucMinions);
            if (success && installJenkinsBySls(jenkinsPillar, sucMinions, start, hostResults, null).isFailed()) {
                success = false;
            }
        }
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }

    /**
     * 通过sls安装jdk
     *
     * @param pillar      sls参数
     * @param sucMinions  执行成功的服务器
     * @param start       任务开始时间
     * @param hostResults 结果列表
     * @param minions     需要执行的服务器
     * @return 任务结果
     */
    private GeneralResult<List<Map<String, Object>>> installJenkinsBySls(Optional<Map<String, Object>> pillar, Map<WorkerHost, String> minions, Long start, List<Map<String, Object>> hostResults, Map<WorkerHost, String> sucMinions) {
        boolean success = true;
        if (null == hostResults) {
            hostResults = new ArrayList<>();
        }
        List<String> mods = new ArrayList<>();
        mods.add("jenkins.install_jenkins");
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务install jenkins结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务install jenkins结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取install jenkins结果失败");
        }
        success = processJobResults(hostResults, cmdResults, minions, start, success, sucMinions);
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }


    /**
     * 通过sls文件安装jdk+jce
     *
     * @param pillar     安装的参数
     * @param sucMinions 执行成的服务器信息
     * @param start      开始时间
     * @param minions    需要执行的服务器
     * @return 安装结果
     */
    private GeneralResult<?> installJDKBySls(Optional<Map<String, Object>> pillar, Map<WorkerHost, String> minions, Long start, List<Map<String, Object>> hostResults, Map<WorkerHost, String> sucMinions) {
        boolean success = true;
        if (null == hostResults) {
            hostResults = new ArrayList<>();
        }
        Map<WorkerHost, String> successMinions = new HashMap<>(16);
        if (null != sucMinions) {
            successMinions = sucMinions;
        }
        List<String> mods = new ArrayList<>();
        mods.add("java.install_jdk");

        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务install jdk结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务install jdk结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取install jdk结果失败");
        }
        success = processJobResults(hostResults, cmdResults, minions, start, success, successMinions);
        if (!CollectionUtils.isEmpty(successMinions)) {
            //jdk安装成功的，安装jce
            minions.clear();
            minions.putAll(successMinions);
            GeneralResult<List<Map<String, Object>>> result = installJceBySls(pillar, minions, start, hostResults, sucMinions);
            hostResults.addAll(result.getData());
            if (success && result.isFailed()) {
                success = false;
            }
        }
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }

    /**
     * 安装jce
     *
     * @param pillar     参数
     * @param sucMinions 执行的minions
     * @param start      执行的开始时间
     * @return 操作结果
     */
    private GeneralResult<List<Map<String, Object>>> installJceBySls(Optional<Map<String, Object>> pillar, Map<WorkerHost, String> minions, Long start, List<Map<String, Object>> hostResults, Map<WorkerHost, String> sucMinions) {
        boolean success = true;
        if (null == hostResults) {
            hostResults = new ArrayList<>();
        }
        List<String> mods = new ArrayList<>();
        mods.add("java.install_jce");
        Optional<LocalAsyncResult<Map<String, State.ApplyResult>>> optionalResult = getLocalCall(mods, pillar)
                .withTimeouts(Optional.of(timeout), Optional.of(timeout))
                .callAsync(client, new MinionList(new ArrayList<>(minions.values())), method).toCompletableFuture().join();
        //获取异步结果
        Map<String, Map<String, JSONObject>> cmdResults = new HashMap<>();
        if (null != optionalResult && null != optionalResult.get()) {
            cmdResults = getResultByJid(optionalResult.get().getJid(), start);
            if (null == cmdResults) {
                return new GeneralResult<>(false, hostResults, "任务install jce结果获取失败");
            } else if (cmdResults.isEmpty()) {
                return new GeneralResult<>(false, hostResults, "任务install jce结果获取超时");
            }
        } else {
            return new GeneralResult<>(false, hostResults, "获取install jce结果失败");
        }
        success = processJobResults(hostResults, cmdResults, minions, start, success, sucMinions);
        return new GeneralResult<>(success, hostResults, "Finished the job");
    }

    /**
     * 对结果进行处理
     *
     * @param hostResults 处理后结果列表
     * @param cmdResults  执行结果原始数据
     * @param minions     执行的服务器
     * @param start       开始时间
     * @param success     是否成功
     * @param sucMinions  执行成功的服务器（如果不为null,执行成功的服务器结果不会放入hostResults）
     * @return 执行是否成功
     */
    private boolean processResults(List<Map<String, Object>> hostResults, Map<String, Result<Map<String, State.ApplyResult>>> cmdResults, Map<WorkerHost, String> minions,
                                   Long start, boolean success, Map<WorkerHost, String> sucMinions) {
        long cost = System.currentTimeMillis() - start;
        if (null != sucMinions) {
            sucMinions.clear();
        }
        for (Map.Entry<WorkerHost, String> minion : minions.entrySet()) {
            Result<Map<String, State.ApplyResult>> cmdResult = cmdResults.get(minion.getValue());
            log.info("cmdResult:{}", cmdResult);

            boolean resultFlag = true;
            String stdout = "";
            String stderr = "";
            String changes = "";
            Optional<SaltError> error = cmdResult.error();
            if (error.equals(Optional.empty())) {
                Map<String, State.ApplyResult> resultMap = cmdResult.result().get();
                for (Map.Entry<String, State.ApplyResult> result : resultMap.entrySet()) {
                    log.info("salt stack sls key:【{}】", result.getKey());
                    if (result.getKey().indexOf("_ignore") > 0) {
                        //忽略执行结果
                        log.info("ignore this option key result:{}", result.getKey());
                    } else if (result.getKey().startsWith("file_|")) {
                        // 文件执行结果
                        if (!result.getValue().isResult()) {
                            resultFlag = result.getValue().isResult();
                            // 错误信息:
                            stderr += result.getValue().getComment() + "\n";
                        } else {
                            stdout += result.getValue().getComment() + "\n";
                            JsonElement element = result.getValue().getChanges();
                            JsonObject jsonObject = element.getAsJsonObject();
                            if (null != jsonObject && jsonObject.has("diff")) {
                                changes += jsonObject.get("diff").getAsString() + "\n";
                            }
                        }
                    } else if (result.getKey().startsWith("cmd_|")) {
                        // 脚本执行结果
                        JsonElement element = result.getValue().getChanges();
                        JsonObject obj = element.getAsJsonObject();
                        if (!result.getValue().isResult()) {
                            resultFlag = result.getValue().isResult();
                            // 错误信息
                            if (null != obj && null != obj.get("stderr")) {
                                stderr += obj.get("stderr").getAsString() + "\n";
                            } else {
                                stderr += result.getValue().getComment() + "\n";
                            }
                        } else {
                            if (null != obj && null != obj.get("stdout")) {
                                stdout += obj.get("stdout").getAsString() + "\n";
                            } else {
                                stdout += result.getValue().getComment() + "\n";
                            }
                        }
                    } else {
                        // 其他
                        JsonElement element = result.getValue().getChanges();
                        JsonObject obj = element.getAsJsonObject();
                        if (!result.getValue().isResult()) {
                            resultFlag = result.getValue().isResult();
                            // 错误信息
                            if (null != obj && null != obj.get("stderr")) {
                                stderr += obj.get("stderr").getAsString() + "\n";
                            } else {
                                stderr += result.getValue().getComment() + "\n";
                            }
                        } else {
                            if (null != obj && null != obj.get("stdout")) {
                                stdout += obj.get("stdout").getAsString() + "\n";
                            } else {
                                stdout += result.getValue().getComment() + "\n";
                            }
                        }
                    }
                }
            } else {
                resultFlag = false;
                SaltError saltError = error.get();
                stderr = saltError.toString();
            }
            Map<String, Object> hostResult = JSON.parseObject(JSON.toJSONString(minion.getKey()));
            hostResult.put("cost", cost);
            hostResult.put("success", resultFlag);
            if (resultFlag) {
                hostResult.put("message", stdout);
                hostResult.put("changes", changes);
            } else {
                success = false;
                hostResult.put("message", stderr);
            }
            cmdResult.error().map(Objects::toString).ifPresent(msg -> hostResult.put("message", msg));
            if (null == hostResult.get("message") || StringUtils.isEmpty(hostResult.get("message").toString())) {
                hostResult.put("message", resultFlag ? "执行成功" : "执行失败");
            }
            hostResult.put("workerInstance", instance);
            if (null != sucMinions && resultFlag) {
                //执行成功的minions，用于执行下一个sls
                sucMinions.put(minion.getKey(), minion.getValue());
            } else {
                hostResults.add(hostResult);
            }
        }

        if (cmdResults.values().stream().map(Result::error).anyMatch(Optional::isPresent)) {
            success = false;
        }
        return success;
    }


    private boolean processJobResults(List<Map<String, Object>> hostResults, Map<String, Map<String, JSONObject>> cmdResults, Map<WorkerHost, String> minions,
                                      Long start, boolean success, Map<WorkerHost, String> sucMinions) {
        long cost = System.currentTimeMillis() - start;
        for (Map.Entry<WorkerHost, String> minion : minions.entrySet()) {
            boolean resultFlag = true;
            String stderr = "";
            String stdout = "";
            String changes = "";
            //判断是否存在是队列已满
            String minionResultStr = JSONObject.toJSONString(cmdResults.get(minion.getValue()));
            if (minionResultStr.indexOf("The function \\\"state.apply\\\" is running") > 0) {
                resultFlag = false;
                stderr = "当前服务器任务队列已满，清稍后执行";
            } else {
                Map<String, JSONObject> cmdResult = cmdResults.get(minion.getValue());
                log.info("cmdResult:{}", cmdResult);
                if (null != cmdResult) {
                    for (String key : cmdResult.keySet()) {
                        JSONObject obj = cmdResult.get(key);
                        StateApplyResultModel model = JSONObject.parseObject(obj.toJSONString(), StateApplyResultModel.class);
                        log.info("salt stack sls key:【{}】,success:【{}】", key, model.isResult());
                        if (key.startsWith("file_|")) {
                            // 文件执行结果
                            if (!model.isResult()) {
                                resultFlag = model.isResult();
                                // 错误信息:
                                stderr += model.getComment() + "\n";
                            } else if (key.indexOf("_ignore") > 0) {
                                //忽略执行结果
                                log.info("ignore this option key result:{}", model.isResult());
                            } else {
                                stdout += model.getComment() + "\n";
                                JSONObject jsonObject = model.getChanges();
                                if (null != jsonObject && jsonObject.containsKey("diff")) {
                                    changes += jsonObject.getString("diff") + "\n";
                                }
                            }
                        } else if (key.startsWith("cmd_|")) {
                            // 脚本执行结果
                            JSONObject jsonObject = model.getChanges();
                            if (!model.isResult()) {
                                resultFlag = model.isResult();
                                // 错误信息
                                if (null != jsonObject && null != jsonObject.getString("stderr")) {
                                    stderr += jsonObject.getString("stderr") + "\n";
                                } else if (key.indexOf("_ignore") > 0) {
                                    //忽略执行结果
                                    log.info("ignore this option key result:{}", model.isResult());
                                } else {
                                    stderr += model.getComment() + "\n";
                                }
                            } else {
                                if (null != jsonObject && null != jsonObject.getString("stdout")) {
                                    stdout += jsonObject.getString("stdout") + "\n";
                                } else if (key.indexOf("_ignore") > 0) {
                                    //忽略执行结果
                                    log.info("ignore this option key result:{}", model.isResult());
                                } else {
                                    stdout += model.getComment() + "\n";
                                }
                            }
                        } else {
                            // 其他
                            // 脚本执行结果
                            JSONObject jsonObject = model.getChanges();
                            if (!model.isResult()) {
                                resultFlag = model.isResult();
                                // 错误信息
                                if (null != jsonObject && null != jsonObject.get("stderr")) {
                                    stderr += jsonObject.getString("stderr") + "\n";
                                } else if (key.indexOf("_ignore") > 0) {
                                    //忽略执行结果
                                    log.info("ignore this option key result:{}", model.isResult());
                                } else {
                                    stderr += model.getComment() + "\n";
                                }
                            } else {
                                if (null != jsonObject && null != jsonObject.get("stdout")) {
                                    stdout += jsonObject.getString("stdout") + "\n";
                                } else if (key.indexOf("_ignore") > 0) {
                                    //忽略执行结果
                                    log.info("ignore this option key result:{}", model.isResult());
                                } else {
                                    stdout += model.getComment() + "\n";
                                }
                            }
                        }
                    }
                } else {
                    resultFlag = false;
                    stderr = "获取主机任务结果失败";
                }
            }
            Map<String, Object> hostResult = JSON.parseObject(JSON.toJSONString(minion.getKey()));
            hostResult.put("cost", cost);
            hostResult.put("success", resultFlag);
            if (resultFlag) {
                hostResult.put("message", stdout);
                hostResult.put("changes", changes);
            } else {
                success = false;
                hostResult.put("message", stderr);
            }
            if (null == hostResult.get("message") || StringUtils.isEmpty(hostResult.get("message").toString())) {
                hostResult.put("message", resultFlag ? "执行成功" : "执行失败");
            }
            hostResult.put("workerInstance", instance);
            if (null != sucMinions && resultFlag) {
                //执行成功的minions，用于执行下一个sls
                sucMinions.put(minion.getKey(), minion.getValue());
            } else {
                hostResults.add(hostResult);
            }
        }
        return success;
    }


    /**
     * 获取异步结果
     *
     * @param jid   任务id
     * @param start 开始时间
     * @return 返回结果
     * @throws Exception 异常
     */
    private Map<String, Map<String, JSONObject>> getResultByJid(String jid, Long start) {
        Map<String, Map<String, JSONObject>> resultMap = new HashMap<>();
        try {
            int i = 0;
            while (true) {
                Thread.sleep(10000);
                //任务未结束
                if (System.currentTimeMillis() - start > timeout) {
                    log.error("获取任务结果超时:{}", jid);
                    return resultMap;
                }
                log.info("------------第{}次获取任务结果------------", ++i);
                resultMap = getJobs(jid);
                if (!resultMap.isEmpty()) {
                    return resultMap;
                }
            }
        } catch (Exception e) {
            log.error("获取任务结果失败:{}", jid);
            return null;
        }

    }

    private LocalCall<Map<String, State.ApplyResult>> getLocalCall(List<String> mods, Optional<Map<String, Object>> pillar) {
        Map<String, Object> kwargs = new LinkedHashMap();
        kwargs.put("mods", mods);
        pillar.ifPresent((p) -> {
            kwargs.put("pillar", p);
        });
        // 任务调用时，设置的queue信息会覆盖minion的设置。如果不设置值会使用minion的配置
        if (null != this.queueLimit) {
            if (0 == this.queueLimit) {
                //设置为0，不使用队列
                kwargs.put("queue", false);
            } else if (this.queueLimit > 0) {
                //大于0，设置队列排队的数量
                kwargs.put("queue", this.queueLimit);
            }
        }
        return new LocalCall("state.apply", Optional.empty(), Optional.of(kwargs), new TypeToken<Map<String, State.ApplyResult>>() {
        });
    }
}
