package com.bocloud.tse.scheduler.worker.executor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;

import com.megatron.framework.http.model.ServiceParam;
import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.core.Binding;
import org.springframework.amqp.core.Binding.DestinationType;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.core.Queue;
import org.springframework.amqp.rabbit.core.RabbitAdmin;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.cmp.boot.model.BoCloudExchange;
import com.bocloud.tse.core.exception.NodeFormatException;
import com.bocloud.tse.model.HostResult;
import com.bocloud.tse.model.WorkerHost;
import com.megatron.common.enums.PostDataFormat;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.utils.IDFactory;
import com.megatron.common.utils.ListTool;
import com.megatron.common.utils.MapTools;
import com.megatron.framework.http.model.RemoteService;

import lombok.extern.slf4j.Slf4j;

/**
 * 远端配置任务执行线程，用来支持任务执行能分发的对某个worker的请求操作
 */
@Slf4j
public class RemoteWorker implements Callable<GeneralResult<?>> {

    private static final String EXCHANGE = BoCloudExchange.Default.DIRECT.getExchange();
    private static final Map<String, Object> arguments = MapTools.simpleMap("x-ha-policy", "all");
    private static final Integer DEFAULT_TIMEOUT = 60 * 10;
    private static final String QUEUE = "queue";
    private static final String TIMEOUT = "timeout";
    private List<WorkerHost> targets;

    // 默认超时时间是10分钟
    private Integer timeout;
    // worker返回任务结果的队列
    private String callback;
    private boolean sync = false;
    private RemoteService service;
    private RabbitAdmin rabbitAdmin;
    private String errorMessage;
    private String nodeId;
    // 执行worker实例
    private String instance;

    private boolean fixed;


    public RemoteWorker(RemoteService service, RabbitAdmin rabbitAdmin, Integer timeout, String callback, boolean sync,
                        String errorMessage, String nodeId, String instance, boolean fixed) throws Exception {
        this.service = new WorkerRemoteService(service.getService(), service.getParam(), true, service.getUrl(),
                service.getMethod());
        this.rabbitAdmin = rabbitAdmin;
        this.timeout = null == timeout ? DEFAULT_TIMEOUT : timeout;
        this.callback = callback;
        this.sync = sync;
        this.errorMessage = errorMessage;
        this.nodeId = nodeId;
        this.instance = instance;
        this.fixed = fixed;
        init();
    }

    /**
     * 构建通过Mq下发的remoteWorker
     */
    public RemoteWorker(ServiceParam param, RabbitAdmin rabbitAdmin, Integer timeout, String callback,
                        String errorMessage, String nodeId, String instance, boolean fixed) throws Exception {
        this.service = new WorkerRemoteMQService(instance, param, rabbitAdmin);
        this.rabbitAdmin = rabbitAdmin;
        this.timeout = null == timeout ? DEFAULT_TIMEOUT : timeout;
        this.callback = callback;
        this.sync = false;
        this.errorMessage = errorMessage;
        this.nodeId = nodeId;
        this.instance = instance;
        this.fixed = fixed;
        init();
    }


    @Override
    public GeneralResult<?> call() throws Exception {
        if (StringUtils.isNotEmpty(errorMessage)) {
            log.error("task execute fail:" + errorMessage);
            return handleError(errorMessage);
        }
        return sync ? handleSync() : handleAsync();
    }

    /**
     * 处理同步调用
     */
    private GeneralResult<?> handleSync() {
        log.info("Sync node job [{}] call remote worker [{}] and timeout is:[{}]ms", nodeId, service.getUrl(), timeout);
        return handleResult(service.invoke());
    }

    /**
     * 处理异步调用
     */
    @SuppressWarnings("unchecked")
    private GeneralResult<?> handleAsync() {
        if (StringUtils.isEmpty(callback)) {
            this.callback = IDFactory.instance().uuid();
        }
        log.info("Async node job [{}] call remote worker [{}],timeout is [{}]s,callback is [{}]", nodeId,
                service.getUrl(), timeout, callback);
        Map<String, Object> params = (Map<String, Object>) service.getParam().getParams();
        params.put(QUEUE, this.callback);
        service.getParam().setParams(params);
        if (fixed) {
            return invoke();
        }
        GeneralResult<?> result;
        try {
            // 声明worker结果的存放队列并将其绑定到指定的exchange:bocloud.direct.exchange
            rabbitAdmin.declareQueue(new Queue(this.callback, false, false, true, arguments));
            rabbitAdmin
                    .declareBinding(new Binding(this.callback, DestinationType.QUEUE, EXCHANGE, this.callback, arguments));
            GeneralResult<?> invoke = invoke();
            if (invoke.isFailed()) {
                return invoke;
            }
            // 这里说明请求成功发送到worker上，接下来可以等待结果消息了
            Message message = rabbitAdmin.getRabbitTemplate().receive(callback, (timeout + 60) * 1000);
            if (null == message) {
                log.error("get callback from remote worker by queue [{}] timeout", callback);
                result = handleError("获取任务结果消息超时！");
            } else {
                String content = new String(message.getBody(), "UTF-8");
                log.info("get node job [{}] result from remote worker by queue [{}]", nodeId, callback);
                if (log.isDebugEnabled()) {
                    log.debug("callback from remote worker is:\n[{}]", content);
                }
                try {
                    result = handleResult(JSONObject.parseObject(content, GeneralResult.class));
                } catch (Exception e) {
                    log.error("callback data transfer exception:", e);
                    result = handleError("返回数据格式异常!");
                }
            }
        } catch (Exception e) {
            log.error("等待Worker【{}】返回任务结果消息异常:", service.getUrl(), e);
            result = handleError(e.getMessage());
        } finally {
            rabbitAdmin.deleteQueue(this.callback);
        }
        if (result.isFailed()) {
            log.error("node job [{}] failed result [{}] by queue [{}]", nodeId, JSONObject.toJSONString(result),
                    callback);
        }
        return result;
    }


    private GeneralResult<?> invoke() {
        log.info("start node job [{}] request callback is [{}]", nodeId, callback);
        // log.info("request params is [{}]", JSONObject.toJSON(service.getParam().getParams()));
        GeneralResult<?> invoke = service.invoke();
        if (invoke.isFailed()) {
            log.error("invoke remote worker [{}] failed for: [{}]", service.getUrl(), invoke.getMessage());
            return handleError("任务下发请求异常，请检查worker机器网络状况或者worker状态，请求路径为：" + service.getUrl());
        }
        log.info("end node job [{}] request callback is [{}]", nodeId, callback);
        if (!invoke.isFailed()) {
            // invoke成功，通过返回message里面的状态最终判断是否成功。 失败就直接处理，不用等待消息结果
            JSONObject message = JSONObject.parseObject(invoke.getMessage());
            boolean issuccess = message.getBoolean("success");
            if (!issuccess) {
                log.error("invoke remote worker [{}] failed for: [{}]", service.getUrl(), invoke.getMessage());
                return handleError(message.getString("message"));
            }
        }
        return invoke;
    }

    /**
     * 构建消息
     */
    private GeneralResult<?> handleResult(GeneralResult<?> invoke) {
        if (null == invoke.getData()) {
            String message = "返回结果异常：未返回任何机器任务信息；worker可能处于异常状态，未进行任务执行";
            return handleError(message);
        }
        List<HostResult> results = buildHostResult(invoke);
        if (ListTool.isEmpty(results)) {
            String message = "返回结果异常：未返回任何机器任务信息；worker可能处于异常状态，未进行任务执行";
            for (WorkerHost target : targets) {
                results.add(new HostResult(false, message, target.getHost(), target.getRawName(), target.getPort(),
                        target.getUser(), instance));
            }
            return new GeneralResult<>(false, results, invoke.getMessage());
        }
        for (HostResult result : results) {
            result.setWorkerInstance(instance);
        }
        return new GeneralResult<>(invoke.isSuccess(), results, invoke.getMessage());
    }

    private List<HostResult> buildHostResult(GeneralResult<?> result) {
        List<HostResult> results = new ArrayList<>();
        if (null != result.getData()) {
            try {
                results = JSONArray.parseArray(JSONObject.toJSONString(result.getData()), HostResult.class);
            } catch (Exception e) {
                log.error("build host result error!", e);
            }
        }
        return results;
    }

    private GeneralResult<List<HostResult>> handleError(String message) {
        List<HostResult> results = new ArrayList<>();
        for (WorkerHost target : targets) {
            results.add(new HostResult(false, message, target.getHost(), target.getRawName(), target.getPort(),
                    target.getUser(), instance));
        }
        return new GeneralResult<>(false, results, message);
    }

    @SuppressWarnings("unchecked")
    private void init() throws Exception {
        Map<String, Object> params = (Map<String, Object>) service.getParam().getParams();
        if (null == params) {
            params = new HashMap<String, Object>();
        }
        params.put(TIMEOUT, timeout);
        service.getParam().setParams(params);
        service.setDataFormat(PostDataFormat.RAW);
        targets = WorkerHost.parseTargets(JSONObject.toJSONString(params.get("targets")));
        if (targets.isEmpty()) {
            throw new NodeFormatException("目标机器参数信息格式错误");
        }
    }

}
