package com.bocloud.tse.scheduler.worker.maintain;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.springframework.amqp.rabbit.core.RabbitAdmin;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.util.StringUtils;

import com.alibaba.druid.DbType;
import com.alibaba.druid.sql.SQLUtils;
import com.alibaba.druid.sql.ast.SQLStatement;
import com.alibaba.druid.util.JdbcConstants;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.core.annotation.Worker;
import com.bocloud.tse.core.domain.front.maintain.DBFront;
import com.bocloud.tse.core.domain.input.maintain.DBInput;
import com.bocloud.tse.core.exception.NodeFormatException;
import com.bocloud.tse.core.utils.DBDriver;
import com.bocloud.tse.core.utils.DBTool;
import com.bocloud.tse.scheduler.worker.AbstractWorker;
import com.megatron.common.encrypt.AESEncryptor;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.Result;
import com.megatron.framework.http.core.ServiceFactory;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@Worker(name = "task.maintain.database")
public class DatabaseWorker extends AbstractWorker {

    /**
     * 超时时间,默认超时时间为10分钟
     */
    private Integer timeout = 1000 * 60 * 10;
    /**
     * 数据库连接字符串
     */
    private String url;
    /**
     * 数据库操作参数
     */
    private DBFront param;

    /**
     * 数据库连接
     */
    private Connection connection;

    public DatabaseWorker(String code, String nodeId, RabbitAdmin rabbitAdmin, StringRedisTemplate redisTemplate,
        ServiceFactory serviceFactory, LockFactory lockFactory, Integer timeout) {
        super(code, nodeId, rabbitAdmin, redisTemplate, serviceFactory, lockFactory, timeout);
    }

    @Override
    public void init(Object param) throws NodeFormatException {
        if (param instanceof DBInput) {
            DBInput input = (DBInput)param;
            this.param = input.getFronts();
            this.url = buildConnectionUrl(this.param.getJdbcUrl());
        } else {
            throw new NodeFormatException("database task params illegel!");
        }
    }

    @Override
    public GeneralResult<?> call(String instanceId) {
        GeneralResult<?> result;
        if (loadDriver().isFailed()) {
            log.error("database driver [{}] load exception ! ", param.getDriver());
            result = buildErrorResult("无法加载到驱动,驱动参数为" + param.getDriver());
        } else {
            try {
                List<String> scriptList = splitScript(param.getContent());
                List<String> scriptResult = new ArrayList<>();
                ExecutorService executor = Executors.newSingleThreadExecutor();
                // 开启一个子线程执行sql任务，当前线程进行超时等待
                Future<GeneralResult<?>> future = getNodeResultFuture(scriptList, scriptResult, executor);
                // 处理future的结果
                result = getNodeResult(future, executor);
            } catch (Exception e) {
                log.error("database executor exception:", e);
                result = buildErrorResult("无法加载到驱动,驱动参数为" + param.getDriver());
            }
        }
        return result;
    }

    private GeneralResult<?> getNodeResult(Future<GeneralResult<?>> future, ExecutorService executor) {
        ExecutorService timeoutExecutor = Executors.newSingleThreadExecutor();
        try {
            // 超时时间+10秒之后，查看结果,如果没有结果，数据库任务可能已经超时
            return future.get(Long.valueOf(timeout) + 10000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("线程异常", e);
            return buildErrorResult("出现异常错误");
        } catch (ExecutionException e) {
            log.error("执行task错误", e);
            return buildErrorResult("执行task错误");
        } catch (TimeoutException e) {
            // 超时
            future.cancel(true);
            try {
                // 尝试终止连接
                connection.abort(timeoutExecutor);
                return buildErrorResult("taks已经取消,请等待数据回滚");
            } catch (SQLException e1) {
                log.error("终端连接失败", e1);
            }
            try {
                // 如果终止连接失败，可能是jdbc驱动没有实现终止的方法,此时只能阻塞，然后等待操作完回滚了
                connection.rollback();
            } catch (SQLException e1) {
                log.error("超时回滚出错", e);
                return buildErrorResult("执行sql失败！");
            }
            return buildErrorResult("taks超时取消");
        } finally {
            closeConnection();
            executor.shutdown();
            timeoutExecutor.shutdown();
        }
    }

    /**
     * 执行业务方法
     */
    private Future<GeneralResult<?>> getNodeResultFuture(List<String> scriptList, List<String> scriptResult,
        ExecutorService executor) {
        return executor.submit(new Callable<GeneralResult<?>>() {
            @Override
            public GeneralResult<?> call() throws Exception {
                if (scriptList.isEmpty()) {
                    log.error("database sctipt is empty!");
                    return buildErrorResult("script脚本有误");
                }
                PreparedStatement pstmt = null;
                boolean haveUpdateOp = false;
                try {
                    // 开启连接
                    connection = DriverManager.getConnection(url, param.getUsername(),
                        new AESEncryptor().decrypt(param.getPassword(), null));
                } catch (SQLException se) {
                    log.error("build database connection error!", se);
                    return buildErrorResult("无法建立连接,url参数为" + url);
                }
                try {
                    // 设置自动提交关闭
                    connection.setAutoCommit(false);
                    for (String toExecute : scriptList) {
                        pstmt = connection.prepareStatement(toExecute);
                        pstmt.setQueryTimeout(timeout);
                        // oracle使用此配置有效，mysql无效
                        String resultStr = null;
                        if (isSelect(toExecute) || isShow(toExecute)) {
                            ResultSet resultSet = pstmt.executeQuery();
                            resultStr = resultSetToJson(resultSet);
                        } else {
                            haveUpdateOp = true;
                            int updateCount = pstmt.executeUpdate();
                            resultStr = "受影响的行数为" + updateCount;
                        }
                        scriptResult.add(resultStr);
                        closePrepareStatement(pstmt);
                    }
                    connection.commit();
                } catch (Exception e) {
                    log.error("执行sql失败！", e);
                    try {
                        if (haveUpdateOp) {
                            connection.rollback();
                        }
                    } catch (Exception e1) {
                        log.error("尝试数据库回滚失败", e1);
                        return buildErrorResult("调用数据回滚失败");
                    }
                    return buildErrorResult(e.getMessage());
                } finally {
                    closePrepareStatement(pstmt);
                    closeConnection();
                }
                return buildSuccessResult("script执行成功", scriptResult);
            }
        });
    }

    /**
     * 关闭数据库连接
     */
    private void closeConnection() {
        if (connection != null) {
            try {
                connection.close();
            } catch (SQLException e) {
                log.error("连接关闭失败", e);
            }
        }
    }

    /**
     * 关闭preparedstatement
     */
    private void closePrepareStatement(PreparedStatement pstmt) {
        if (pstmt != null) {
            try {
                pstmt.close();
            } catch (SQLException e) {
                log.error("pstmt关闭失败", e);
            }
        }
    }

    /**
     * 创建错误结果
     */
    private GeneralResult<?> buildErrorResult(String message) {
        return new GeneralResult<>(false, message, "failed");
    }

    /**
     * 创建成功结果
     */
    private GeneralResult<?> buildSuccessResult(String message, List<String> results) {
        return new GeneralResult<>(true, results, message);
    }

    /**
     * 生成数据库连接串
     */
    private String buildConnectionUrl(String url) throws NodeFormatException { // 此方法可能会用
        String prefix = "";
        try {
            switch (DBDriver.valueOf(param.getDatabase().toUpperCase())) {
                case MYSQL:
                    prefix = "jdbc:mysql://";
                    break;
                case ORACLE:
                    prefix = "jdbc:oracle:thin:@";
                    break;
                case SQLSERVER_MS:
                    prefix = "jdbc:sqlserver://";
                    break;
                case SQLSERVER_JTS:
                    prefix = "jdbc:jtds:sqlserver://";
                    break;
                default:
                    break;
            }
            if (!StringUtils.hasText(prefix)) {
                log.error("jdbc url is illegel!");
                throw new NodeFormatException("数据库操作参数错误!");
            }
        } catch (Exception e) {
            log.error("jdbc url is illegel!", e);
            throw new NodeFormatException("数据库操作参数错误!");
        }
        return prefix + url;
    }

    /**
     * 判断是否查询操作
     */
    private boolean isSelect(String script) {
        return script.toLowerCase().trim().startsWith("select");
    }

    /**
     * 判断是否查询操作
     */
    private boolean isShow(String script) {
        return script.toLowerCase().trim().startsWith("show");
    }

    /**
     * ResultSet转Json
     */
    public String resultSetToJson(ResultSet rs) throws SQLException {
        // json数组
        JSONArray array = new JSONArray();
        // 获取列数
        ResultSetMetaData metaData = rs.getMetaData();
        int columnCount = metaData.getColumnCount();
        // 遍历ResultSet中的每条数据
        while (rs.next()) {
            JSONObject jsonObj = new JSONObject();
            // 遍历每一列
            for (int i = 1; i <= columnCount; i++) {
                String columnName = metaData.getColumnLabel(i);
                String value = rs.getString(columnName);
                jsonObj.put(columnName, value);
            }
            array.add(jsonObj);
        }
        return array.toJSONString();
    }

    /**
     * 对脚本进行分析拆分
     */
    private List<String> splitScript(String script) throws Exception {
        List<String> retList = new ArrayList<>();
        String driver = param.getDBDriver().getDriver();
        DbType dbType = null;
        if (DBTool.isMySql(driver)) {
            dbType = JdbcConstants.MYSQL;
        } else if (DBTool.isOracle(driver)) {
            dbType = JdbcConstants.ORACLE;
        } else if (DBTool.isSqlServer(driver)) {
            dbType = JdbcConstants.SQL_SERVER;
        }
        // 格式化输出
        List<SQLStatement> stmtList = SQLUtils.parseStatements(script, dbType);
        // 遍历解析出的sql语句
        for (int i = 0; i < stmtList.size(); i++) {
            SQLStatement stmt = stmtList.get(i);
            String sql = stmt.toString();
            if (StringUtils.hasText(sql)) {
                if (isSelect(sql)) {
                    // 查询操作的话加上分页
                    sql = sql.replaceAll("([;| ]+)$", " ");
                    if (DBTool.isMySql(driver)) {
                        sql = DBTool.buildMysqlPageSql(sql, 0, 10);
                    } else if (DBTool.isOracle(driver)) {
                        sql = DBTool.buildOraclePageSql(sql, 0, 10);
                    } else if (DBTool.isSqlServer(driver)) {
                        // 暂时只支持sqlserver2012
                        sql = DBTool.buildSqlServer2012PageSql(sql, 0, 10);
                    }
                }
                retList.add(sql);
            }
        }
        return retList;
    }

    private Result loadDriver() {
        try {
            // 加载MySql的驱动类
            Class.forName(param.getDBDriver().getDriver());
        } catch (ClassNotFoundException e) {
            log.error("找不到驱动程序类 ，加载驱动失败！", e);
            return new Result(false, "找不到驱动程序类 ，加载驱动失败！");
        } catch (Exception e) {
            log.error("get driver exception {}!", e.getMessage());
            return new Result(false, "找不到驱动程序类 ，加载驱动失败！");
        }
        return new Result(true, "load success!");
    }

}