package com.bocloud.tse.scheduler.worker.maintain;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.rabbit.core.RabbitAdmin;
import org.springframework.data.redis.core.StringRedisTemplate;

import com.bocloud.tse.constant.CacheConstant;
import com.bocloud.tse.constant.ExecuteTypeConstant;
import com.bocloud.tse.core.domain.TargetGroup;
import com.bocloud.tse.core.domain.input.Target;
import com.bocloud.tse.core.exception.NodeFormatException;
import com.bocloud.tse.model.WorkerHost;
import com.bocloud.tse.scheduler.worker.AbstractWorker;
import com.bocloud.tse.scheduler.worker.WorkerDispatcher;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.megatron.common.enums.OSCategory;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.utils.ListTool;
import com.megatron.framework.http.core.ServiceFactory;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

@Slf4j
public abstract class MaintainWorker extends AbstractWorker {

    /**
     * 目标主机数量经验值
     */
    int expNum = 100;
    WorkerDispatcher dispatcher;
    List<List<WorkerHost>> targets;
    Map<String, List<List<WorkerHost>>> targetsMap;
    List<Map<String, Object>> paramList;

    public MaintainWorker(String code, String nodeId, RabbitAdmin rabbitAdmin, StringRedisTemplate redisTemplate,
        ServiceFactory serviceFactory, LockFactory lockFactory, Integer timeout) {
        super(code, nodeId, rabbitAdmin, redisTemplate, serviceFactory, lockFactory, timeout);
        this.targets = new ArrayList<>();
        this.targetsMap = Maps.newHashMap();
        this.dispatcher = new WorkerDispatcher(rabbitAdmin, redisTemplate, lockFactory, code);
        this.paramList = Lists.newArrayList();
    }

    abstract void buildParam() throws NodeFormatException;

    @Override
    protected GeneralResult<?> call(String instanceId) {
        GeneralResult<?> result;
        try {
            // 任务分发并获取整体结果
            result = dispatcher.execute(paramList, getTimeout(), getNodeId(), instanceId);
        } catch (Exception e) {
            result = GeneralResult.FAILED(e.getMessage());
        }
        return result;
    }

    /**
     * 主机分片
     *
     * @param target
     */
    public void burst(Target target) throws NodeFormatException {
        if (this.getCode().equalsIgnoreCase("task.maintain.hostallow") || this.getCode().equalsIgnoreCase("task.maintain.gaia")) {
            GeneralResult<TargetGroup> targetGroupResult = groupTarget(target.getData());
            if (targetGroupResult.isSuccess()) {
                TargetGroup targetGroup = targetGroupResult.getData();
                this.targets.addAll(splitTarget(targetGroup.getLinuxs()));
                this.targets.addAll(splitTarget(targetGroup.getWindows()));
            }
        } else {
            GeneralResult<TargetGroup> targetGroupResult = groupTarget(target.getData());
            if (targetGroupResult.isSuccess()) {
                TargetGroup targetGroup = targetGroupResult.getData();
                // this.targets.addAll(splitTarget(targetGroup.getLinuxs()));
                // this.targets.addAll(splitTarget(targetGroup.getWindows()));
                List<WorkerHost> linuxs = targetGroup.getLinuxs();
                if (!ListTool.isEmpty(linuxs)) {
                    handleTargetsMap(target.getExecuteType(), linuxs);
                }
                List<WorkerHost> windows = targetGroup.getWindows();
                if (!ListTool.isEmpty(windows)) {
                    handleTargetsMap(target.getExecuteType(), windows);
                }
            }
        }
    }

    private void handleTargetsMap(String executeType, List<WorkerHost> hosts) {
        if (ExecuteTypeConstant.SALTSTACK.equalsIgnoreCase(executeType)) {
            Map<String, List<WorkerHost>> saltMap = new HashMap<>();
            for (WorkerHost host : hosts) {
                String key = CacheConstant.SALTMASTER_ + host.getSaltId();
                if (saltMap.containsKey(key)) {
                    saltMap.get(key).add(host);
                } else {
                    List<WorkerHost> list = new ArrayList<>();
                    list.add(host);
                    saltMap.put(key, list);
                }
            }
            Iterator<String> saltIterator = saltMap.keySet().iterator();
            while (saltIterator.hasNext()) {
                String key = saltIterator.next();
                if (targetsMap.containsKey(key)) {
                    targetsMap.get(key).addAll(splitTarget(saltMap.get(key)));
                } else {
                    targetsMap.put(key, splitTarget(saltMap.get(key)));
                }
            }
            return;
        }

        /* segmengMap：根据机器网段进行分组
         * key:segment_{host对应的网段id}
         * value:同一网段下的所有host
         */
        Map<String, List<WorkerHost>> segmengMap = new HashMap<>();
        for (WorkerHost host : hosts) {
            String key = CacheConstant.SEGMENT_ + host.getSegmentId();
            if (segmengMap.containsKey(key)) {
                segmengMap.get(key).add(host);
            } else {
                List<WorkerHost> list = new ArrayList<>();
                list.add(host);
                segmengMap.put(key, list);
            }
        }
        Iterator<String> iterator = segmengMap.keySet().iterator();
        /* groupMap:根据机器网段分组进行分组
         * key: 网段分组id
         * value: 网段分组下对应的所有机器集合
         */
        Map<String, List<WorkerHost>> groupMap = new HashMap<>();
        List<WorkerHost> nogroupList = new ArrayList<>();
        while (iterator.hasNext()) {
            String key = iterator.next();
            String value = this.getRedisTemplate().opsForValue().get(key);
            if (StringUtils.isEmpty(value)) {
                nogroupList.addAll(segmengMap.get(key));
                continue;
            }
            if (groupMap.containsKey(value)) {
                groupMap.get(value).addAll(segmengMap.get(key));
            } else {
                List<WorkerHost> list = new ArrayList<>();
                list.addAll(segmengMap.get(key));
                groupMap.put(value, list);
            }
        }
        if (!ListTool.isEmpty(nogroupList)) {
            groupMap.put("noworkergroup", nogroupList);
        }
        Iterator<String> groupIterator = groupMap.keySet().iterator();
        while (groupIterator.hasNext()) {
            String key = groupIterator.next();
            if (targetsMap.containsKey(key)) {
                targetsMap.get(key).addAll(splitTarget(groupMap.get(key)));
            } else {
                /* targetsMap
                 * key:网段分组id
                 * value: 网段分组下对应的所有机器集合
                 */
                targetsMap.put(key, splitTarget(groupMap.get(key)));
            }
        }
    }

    /**
     * 将目标机器按照操作系统类型分为Windows和Linux两组
     */
    private GeneralResult<TargetGroup> groupTarget(List<WorkerHost> targets) throws NodeFormatException {
        List<WorkerHost> linuxs = Lists.newArrayList();
        List<WorkerHost> windows = Lists.newArrayList();
        if (ListTool.isEmpty(targets)) {
            throw new NodeFormatException("目标机器为空!");
        }
        for (WorkerHost target : targets) {
            if (OSCategory.LINUX.name().equalsIgnoreCase(target.getCategory())) {
                linuxs.add(target);
            } else if (OSCategory.WINDOWS.name().equalsIgnoreCase(target.getCategory())) {
                windows.add(target);
            } else {
                log.error("未知操作系统，放置到Linux操作系统中");
                linuxs.add(target);
            }
        }
        return new GeneralResult<TargetGroup>(true, new TargetGroup(windows, linuxs), "success");
    }

    /**
     * 根据每个worker处理目标机器数的经验值划分每个worker处理的目标机器IP列表
     */
    private List<List<WorkerHost>> splitTarget(List<WorkerHost> targets) {
        List<List<WorkerHost>> targetList = new ArrayList<>(16);
        int step;
        if (this.getCode().equalsIgnoreCase("task.maintain.recovery")) {
            step = 1;
        } else {
            step = expNum;
        }
        for (int i = 0; i < targets.size(); i = i + step) {
            List<WorkerHost> targetArray = new ArrayList<>(16);
            for (int j = i; j < i + step && j < targets.size(); j++) {
                targetArray.add(targets.get(j));
            }
            targetList.add(targetArray);
        }
        return targetList;
    }

    public int getExpNum() {
        return expNum;
    }

    public void setExpNum(int expNum) {
        this.expNum = expNum;
    }

    public List<List<WorkerHost>> getTargets() {
        return targets;
    }

    public void processBecome(Target target) {
        if (null == target) {
            return;
        }
        if (!ListTool.isEmpty(targets)) {
            for (List<WorkerHost> hosts : this.targets) {
                processBecome(target, hosts);
            }
            return;
        }
        Iterator<String> iterator = targetsMap.keySet().iterator();
        while (iterator.hasNext()) {
            String key = iterator.next();
            List<List<WorkerHost>> list = targetsMap.get(key);
            for (List<WorkerHost> hosts : list) {
                processBecome(target, hosts);
            }
        }

    }

    private void processBecome(Target target, List<WorkerHost> hosts) {
        Boolean become = null != target.getBecome() ? target.getBecome() : target.getSudo();
        for (WorkerHost host : hosts) {
            Boolean hostBecome =
                null != host.getBecome() ? host.getBecome() : (null != host.getSudo() ? host.getSudo() : become);
            if (null == hostBecome || !hostBecome) {
                host.setBecome(false);
                host.setBecomeMethod(null);
                host.setBecomeUser(null);
                host.setBecomePass(null);
                continue;
            }
            host.setBecome(true);
            host.setBecomeMethod(host.getBecomeMethod());
            // 游离的使用自身的becomeUser，becomePass;非游离的会使用统一的账号密码填充账号密码的缺失项
            if (!host.isFree()) {
                host.setBecomeUser(null != host.getBecomeUser() ? host.getBecomeUser() : target.getBecomeUser());
                host.setBecomePass(null != host.getBecomePass() ? host.getBecomePass() : target.getBecomePass());
            }
        }
    }
}