package com.bocloud.tse.scheduler.worker.maintain;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;

import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.rabbit.core.RabbitAdmin;
import org.springframework.data.redis.core.StringRedisTemplate;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.constant.ExecuteTypeConstant;
import com.bocloud.tse.core.annotation.Worker;
import com.bocloud.tse.core.domain.WorkerScript;
import com.bocloud.tse.core.domain.front.maintain.ScriptFront;
import com.bocloud.tse.core.domain.input.Target;
import com.bocloud.tse.core.domain.input.maintain.ScriptInput;
import com.bocloud.tse.core.exception.NodeFormatException;
import com.bocloud.tse.model.WorkerHost;
import com.megatron.common.encrypt.AESEncryptor;
import com.megatron.common.encrypt.Encryptor;
import com.megatron.common.utils.ListTool;
import com.megatron.common.utils.MapTools;
import com.megatron.framework.http.core.ServiceFactory;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@Worker(name = "task.maintain.script")
public class ScriptWorker extends MaintainWorker {

    private ScriptInput input;

    /**
     * 构造方法
     *
     * @param code
     * @param rabbitAdmin
     * @param serviceFactory
     * @param timeout
     */
    public ScriptWorker(String code, String nodeId, RabbitAdmin rabbitAdmin, StringRedisTemplate redisTemplate,
        ServiceFactory serviceFactory, LockFactory lockFactory, Integer timeout) {
        super(code, nodeId, rabbitAdmin, redisTemplate, serviceFactory, lockFactory, timeout);
    }

    @Override
    protected void init(Object param) throws NodeFormatException {
        if (param instanceof ScriptInput) {
            this.input = (ScriptInput)param;
            Target target = input.getTarget();
            // 目标主机切片
            burst(target);
            // 构建请求参数
            buildParam();
        } else {
            log.error("Script worker param illegal!");
            throw new NodeFormatException("Worker Param illegal!");
        }
    }

    /**
     * 参数构建
     */
    @Override
    void buildParam() {
        if (null == input) {
            return;
        }
        //task.maintain.script
        String code = input.getCode();
        if (StringUtils.isBlank(code)) {
            code = "";
        }
        // 处理become相关参数
        processBecome(input.getTarget());
        if (code.contains("task.maintain.custom")) {
            buildUnInternalNodeParam();
        } else {
            //构建对应任务节点参数
            buildInternalNodeParam();
        }

    }

    private void buildUnInternalNodeParam() {
        Encryptor encryptor = new AESEncryptor();
        ScriptFront front = input.getFronts();
        List<Map<String, Object>> scripts = front.getScripts();
        List<JSONObject> params = JSONObject.parseArray(front.getParams(), JSONObject.class);
        StringBuffer sb = new StringBuffer();
        String content = scripts.get(0).get("content").toString();
        String category = scripts.get(0).get("category").toString();
        if (StringUtils.isNotBlank(content) && StringUtils.isNotBlank(front.getParams())) {
            if (!ListTool.isEmpty(params)) {
                for (JSONObject param : params) {
                    // 处理脚本参数
                    Boolean internal = param.getBoolean("internal") == null ? true : param.getBoolean("internal");
                    if (internal) {
                        content = content.replaceAll("\\$\\{\\{" + param.getString("key") + "\\}\\}",
                            Matcher.quoteReplacement(param.getString("value")));
                    } else if (!internal) {
                        String value = "";
                        if ("ENCRYPT".equalsIgnoreCase(param.getString("type"))) {
                            // 加密参数,需要解密
                            value = encryptor.decrypt(param.getString("value"), null);
                        } else {
                            value = param.getString("value");
                        }
                        sb.append(" ").append(value);
                    }
                }
            }
        }
        content = content + sb.toString();
        String executeType = this.input.getTarget().getExecuteType();
        Integer timeout = this.getTimeout();
        Long asyncTimeout = front.getAsyncTimeout();
        boolean async = null != front.getAsync() ? front.getAsync() : false;
        if (async) {
            if (null == asyncTimeout || asyncTimeout <= 0) {
                asyncTimeout = timeout >= 1800 ? timeout * 5L / 6 : 1200L;
            }
            if (asyncTimeout > timeout) {
                asyncTimeout = timeout * 1L;
            }
        }
        WorkerScript script = new WorkerScript(category, null, content, null, async, asyncTimeout);
        if (!ListTool.isEmpty(targets)) {
            for (List<WorkerHost> hosts : this.targets) {
                Map<String, Object> param = MapTools.simpleMap("targets", hosts);
                param.put("script", script);
                param.put("response", "async_playbook");
                param.put("timeout", timeout);
                param.put("executeType", executeType);
                paramList.add(param);
            }
            return;
        }
        Iterator<String> iterator = targetsMap.keySet().iterator();
        while (iterator.hasNext()) {
            String key = iterator.next();
            List<List<WorkerHost>> list = targetsMap.get(key);
            for (List<WorkerHost> hosts : list) {
                Map<String, Object> param = MapTools.simpleMap("targets", hosts);
                param.put("script", script);
                param.put("response", "async_playbook");
                param.put("timeout", timeout);
                param.put("executeType", executeType);
                if (ExecuteTypeConstant.SALTSTACK.equalsIgnoreCase(executeType)) {
                    param.put("saltId", key);
                } else {
                    if ("noworkergroup".equalsIgnoreCase(key)) {
                        param.put(key, key);
                    } else {
                        param.put("groupId", key);
                    }
                }
                paramList.add(param);
            }
        }
    }

    private void buildInternalNodeParam() {
        ScriptFront front = input.getFronts();
        String content = front.getContent();
        if (!StringUtils.isEmpty(front.getParams())) {
            List<JSONObject> params = JSONObject.parseArray(front.getParams(), JSONObject.class);
            if (!ListTool.isEmpty(params)) {
                for (JSONObject param : params) {
                    content = content.replaceAll("\\$\\{\\{" + param.getString("key") + "\\}\\}",
                        Matcher.quoteReplacement(param.getString("value")));
                }
            }
        }
        String executeType = this.input.getTarget().getExecuteType();
        Integer timeout = this.getTimeout();
        Long asyncTimeout = front.getAsyncTimeout();
        boolean async = null != front.getAsync() ? front.getAsync() : false;
        if (async) {
            if (null == asyncTimeout || asyncTimeout <= 0) {
                asyncTimeout = timeout >= 1800 ? timeout * 5L / 6 : 1200L;
            }
            if (asyncTimeout > timeout) {
                asyncTimeout = timeout * 1L;
            }
        }
        WorkerScript script = new WorkerScript(front.getCategory(), null, content, null,
            null != front.getAsync() ? front.getAsync() : false, asyncTimeout);
        if (!ListTool.isEmpty(targets)) {
            for (List<WorkerHost> hosts : this.targets) {
                Map<String, Object> param = MapTools.simpleMap("targets", hosts);
                param.put("script", script);
                param.put("response", "async_playbook");
                param.put("timeout", timeout);
                param.put("executeType", executeType);
                paramList.add(param);
            }
            return;
        }
        Iterator<String> iterator = targetsMap.keySet().iterator();
        while (iterator.hasNext()) {
            String key = iterator.next();
            List<List<WorkerHost>> list = targetsMap.get(key);
            for (List<WorkerHost> hosts : list) {
                Map<String, Object> param = MapTools.simpleMap("targets", hosts);
                param.put("script", script);
                param.put("response", "async_playbook");
                param.put("timeout", timeout);
                param.put("executeType", executeType);
                if (ExecuteTypeConstant.SALTSTACK.equalsIgnoreCase(executeType)) {
                    param.put("saltId", key);
                } else {
                    if ("noworkergroup".equalsIgnoreCase(key)) {
                        param.put(key, key);
                    } else {
                        param.put("groupId", key);
                    }
                }
                paramList.add(param);
            }
        }
    }

}