package com.bocloud.tse.service;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.cmp.boot.model.BoCloudService;
import com.bocloud.tse.entity.TaskWorker;
import com.bocloud.tse.intf.EngineService;
import com.bocloud.tse.model.EngineModule;
import com.bocloud.tse.service.internal.TseInternalService;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.utils.ListTool;
import com.megatron.framework.core.RegistryService;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;

import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * @author lyy
 * @date 2023年04月03日 15:42
 */
@Slf4j
@Service("engineService")
public class EngineServiceImpl implements EngineService {

    @Autowired
    private RegistryService registry;
    @Autowired
    private StringRedisTemplate redisTemplate;
    @Autowired
    private TseInternalService tseInternalService;
    @Autowired
    private LockFactory lockFactory;

    @Value("${spring.cloud.zookeeper.discovery.instance-host}:${server.port}")
    private String serviceInstence;

    @Override
    public GeneralResult<?> services() {
        try {
            List<String> services =
                registry.getServices(com.megatron.framework.core.Service.create(BoCloudService.TSE.getName()));
            if (ListTool.isEmpty(services)) {
                return new GeneralResult<>(false, "任务引擎服务列表为空");
            }
            Collections.sort(services);
            String excludeServices = redisTemplate.opsForValue().get("TaskExcludeServices");
            List<String> instances = new ArrayList<>();
            if (StringUtils.isNotEmpty(excludeServices)) {
                instances = Arrays.asList(excludeServices.split(","));
            }
            List<JSONObject> servers = new ArrayList<>();
            for (String service : services) {
                // service格式为：http://10.7.0.26:18088
                if (service.startsWith("http")) {
                    service = service.split("//")[1];
                }
                JSONObject server = new JSONObject();
                server.put("instance", service);
                server.put("excludeInstance", instances.contains(service.split(":")[0]));
                GeneralResult<JSONObject> result = tseInternalService.getStatus(service);
                if (result.isSuccess()) {
                    JSONObject status = result.getData();
                    server.put("tasks", status.getInteger("RunningTask"));
                    server.put("nodes", status.getInteger("RunningNode"));
                    // 获取监听状态
                    GeneralResult<Boolean> listenerStatus = tseInternalService.getEngineStatus(service);
                    if (listenerStatus.isSuccess()) {
                        server.put("status", listenerStatus.getData());
                    }
                }
                servers.add(server);
            }
            return new GeneralResult<>(true, servers, "获取任务引擎服务列表成功");
        } catch (RuntimeException e) {
            log.error("get cop services error:", e);
            return new GeneralResult<>(false, "获取任务引擎服务列表失败");
        }
    }

    @Override
    public GeneralResult<Void> exclude(String instance, Long target) {
        String path = TaskWorker.class.getSimpleName() + "_TaskExcludeServices";
        try (AutoCloseLock lock = lockFactory.getACLock(path)) {
            Assert.isTrue(lock.acquire(10, TimeUnit.SECONDS), "请求超时");
            String excludeServices = redisTemplate.opsForValue().get("TaskExcludeServices");
            Set<String> instances = new HashSet<>();
            if (StringUtils.isNotEmpty(excludeServices)) {
                instances.addAll(Arrays.asList(excludeServices.split(",")));
            }
            instances.add(instance.split(":")[0]);
            String value = "";
            if (instances.size() > 0) {
                value = String.join(",", instances);
            }
            redisTemplate.opsForValue().set("TaskExcludeServices", value);
            return new GeneralResult<>(true, "引擎排除成功");
        } catch (Exception e) {
            log.error("task exclude fail:", e);
            return new GeneralResult<>(false, "引擎排除失败");
        }
    }

    @Override
    public GeneralResult<Void> recover(String instance, Long target) {
        String path = TaskWorker.class.getSimpleName() + "_TaskExcludeServices";
        try (AutoCloseLock lock = lockFactory.getACLock(path)) {
            Assert.isTrue(lock.acquire(10, TimeUnit.SECONDS), "请求超时");
            String excludeServices = redisTemplate.opsForValue().get("TaskExcludeServices");
            Set<String> instances = new HashSet<>();
            if (StringUtils.isNotEmpty(excludeServices)) {
                instances.addAll(Arrays.asList(excludeServices.split(",")));
                instances.remove(instance.split(":")[0]);
            }
            String value = "";
            if (instances.size() > 0) {
                value = String.join(",", instances);
            }
            redisTemplate.opsForValue().set("TaskExcludeServices", value);
            return new GeneralResult<>(true, "引擎恢复成功");

        } catch (Exception e) {
            log.error("task recover fail:", e);
            return new GeneralResult<>(false, "引擎恢复失败");
        }
    }

    @Override
    public GeneralResult<?> action(EngineModule module) {
        String operate = "操作";
        try {
            String instance = module.getInstance();
            String action = module.getAction();
            // 判断ip与端口好不为当前服务，调用指定ip与端口号
            if (!checkInstance(instance)) {
                log.info("当前服务不匹配:{} {}", instance, serviceInstence);
                // 调用其他服务
                return tseInternalService.listenerAction(instance, action);
            }
            if ("stop".equals(action)) {
                GeneralResult<JSONObject> result = tseInternalService.getStatus(instance);
                if (result.isFailed()) {
                    return result;
                }
                JSONObject status = result.getData();
                Integer task = status.getInteger("RunningTask");
                Integer node = status.getInteger("RunningNode");
                if ((null != task && task > 0) || (null != node && node > 0)) {
                    return new GeneralResult<>(false, "该任务引擎存在运行任务不可关闭");
                }
            }
            operate = "stop".equals(action) ? "关闭" : "start".equals(action) ? "开启" : action;
            return new GeneralResult<>(true, "任务引擎" + operate + "成功");
        } catch (Throwable e) {
            log.info("engine action fail: ", e);
            return GeneralResult.FAILED("任务引擎" + operate + "失败");
        }
    }

    /**
     * 校验实例
     *
     * @param instance
     *            实例信息 ip:port
     * @return 是否是当前实例
     */
    @Override
    public boolean checkInstance(String instance) {
        return this.serviceInstence.equals(instance);
    }

}
