package com.bocloud.tse.service;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.entity.TaskFailRecord;
import com.bocloud.tse.entity.TaskInstance;
import com.bocloud.tse.entity.TaskInstanceJob;
import com.bocloud.tse.entity.TaskNode;
import com.bocloud.tse.enums.Status;
import com.bocloud.tse.enums.Status.Inst;
import com.bocloud.tse.intf.TaskInstanceService;
import com.bocloud.tse.model.HostResult;
import com.bocloud.tse.model.Job;
import com.bocloud.tse.model.WorkerHost;
import com.bocloud.tse.model.WorkerHostResult;
import com.bocloud.tse.repository.TaskFailRecordRepository;
import com.bocloud.tse.repository.TaskInstanceHostRepository;
import com.bocloud.tse.repository.TaskInstanceJobRepository;
import com.bocloud.tse.repository.TaskInstanceRepository;
import com.bocloud.tse.repository.TaskNodeRepository;
import com.google.common.collect.Lists;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.utils.ListTool;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

/**
 * 记录抽象Service接口实现类
 */
@Slf4j
@Service("taskInstanceService")
public class TaskInstanceServiceImpl implements TaskInstanceService {

    @Autowired
    private TaskInstanceRepository taskInstanceRepository;
    @Autowired
    private LockFactory lockFactory;
    @Autowired
    private TaskInstanceJobRepository taskInstanceJobRepository;
    @Autowired
    private TaskNodeRepository taskNodeRepository;
    // @Autowired
    // private TaskHostRepository taskHostRepository;
    @Autowired
    private TaskFailRecordRepository taskFailRecordRepository;
    @Autowired
    private TaskInstanceHostRepository taskInstanceHostRepository;

    @Override
    public GeneralResult<Void> modify(TaskInstance instance) {
        try {
            Boolean result = taskInstanceRepository.update(instance);
            return result ? new GeneralResult<>(true, "修改任务记录成功") : new GeneralResult<>(false, "修改任务记录失败");
        } catch (Exception e) {
            log.error("Modify task record fail:", e);
            return new GeneralResult<>(false, "修改任务记录失败");
        }
    }

    @Override
    public GeneralResult<Void> updateStatus(Long id, String status, Long userId) {
        String path = TaskInstance.class.getSimpleName() + "_" + id;
        boolean result = false;
        TaskInstance record = null;
        try (AutoCloseLock lock = lockFactory.getACLock(path)) {
            Assert.isTrue(lock.acquire(10, TimeUnit.SECONDS), "请求超时");
            record = this.taskInstanceRepository.query(id);
            Assert.notNull(record, "数据不存在！");
            result = this.taskInstanceRepository.updateStatus(id, status, userId);
        } catch (Exception e) {
            log.error("Update task instance status fail:", e);
            return new GeneralResult<>(false, "状态更新失败");
        }
        return result ? new GeneralResult<>(true, "状态更新成功") : new GeneralResult<>(false, "状态更新失败");
    }

    @Override
    public GeneralResult<List<TaskInstance>> listByExecHost(Inst status, String executeHost) {
        try {
            List<TaskInstance> list = taskInstanceRepository.listByHost(status, executeHost, null);
            return new GeneralResult<>(true, list, "任务实例查询成功");
        } catch (Exception e) {
            log.error("list task instance is error :", e);
            return new GeneralResult<>(false, new ArrayList<>(), "任务实例查询失败");
        }
    }

    @Override
    public GeneralResult<List<TaskInstance>> listExceptionInstance(String executeHost, String excludeHost) {
        try {
            List<TaskInstance> list = taskInstanceRepository.listExpInstance(executeHost, excludeHost);
            return new GeneralResult<>(true, list, "任务实例查询成功");
        } catch (Exception e) {
            log.error("list exception task instance is error :", e);
            return new GeneralResult<>(false, new ArrayList<>(), "异常任务实例查询失败");
        }
    }

    @Override
    public List<TaskInstance> list(Long taskId) {
        try {
            return taskInstanceRepository.listByTask(taskId);
        } catch (Exception e) {
            log.error("list task instance is error :", e);
            return null;
        }
    }

    @Override
    public GeneralResult<?> putJob(Job job) {
        try {
            TaskInstanceJob object = new TaskInstanceJob(job.getTaskId(), JSONObject.toJSONString(job));
            taskInstanceJobRepository.save(object);
            return new GeneralResult<>(true, "添加任务记录Job成功");
        } catch (Exception e) {
            log.error("Create task instance job fail:", e);
            return new GeneralResult<>(false, "添加任务记录Job失败");
        }
    }

    @Override
    public List<Job> getJobs() {
        String path = TaskInstanceJob.class.getSimpleName();
        try (AutoCloseLock lock = lockFactory.getACLock(path)) {
            Assert.isTrue(lock.acquire(30, TimeUnit.SECONDS), "请求超时");
            int count = taskInstanceJobRepository.count();
            if (count == 0) {
                return new ArrayList<>();
            }
            int rows = 0;
            if (count <= 30) {
                rows = 3;
            } else if (count <= 50) {
                rows = 5;
            } else if (count <= 100) {
                rows = 10;
            } else if (count <= 200) {
                rows = 20;
            } else if (count <= 300) {
                rows = 30;
            } else if (count <= 500) {
                rows = 50;
            } else {
                rows = 100;
            }
            List<TaskInstanceJob> jobs = taskInstanceJobRepository.list(rows);
            List<Job> list = new ArrayList<>();
            List<Long> ids = new ArrayList<>();
            for (TaskInstanceJob job : jobs) {
                list.add(JSONObject.parseObject(job.getJob(), Job.class));
                ids.add(job.getId());
            }
            if (!ids.isEmpty()) {
                taskInstanceJobRepository.delete(ids);
            }
            return list;
        } catch (Exception e) {
            log.error("get instance job fail:", e);
            return new ArrayList<>();
        }
    }

    @Override
    public List<Job> getAllJobs() {
        try {
            List<TaskInstanceJob> jobs = taskInstanceJobRepository.list();
            List<Job> list = new ArrayList<>();
            for (TaskInstanceJob job : jobs) {
                list.add(JSONObject.parseObject(job.getJob(), Job.class));
            }
            return list;
        } catch (Exception e) {
            log.error("list task instance job fail:", e);
            return null;
        }
    }

    @Override
    public void analysis(Long instanceId) {
        try {
            TaskInstance instance = taskInstanceRepository.query(instanceId);
            if (null == instance) {
                log.warn("task instance analysis fail: instance not exist");
                return;
            }
            // 保存任务执行链路数据
            List<TaskNode> nodes = taskNodeRepository.listByInstance(instanceId);
            for (TaskNode node : nodes) {
                if (StringUtils.isEmpty(node.getOutput())) {
                    continue;
                }
                GeneralResult<?> result = JSONObject.parseObject(node.getOutput(), GeneralResult.class);
                if (null == result.getData()) {
                    continue;
                }
                List<HostResult> hostResults =
                    JSONArray.parseArray(JSONObject.toJSONString(result.getData()), HostResult.class);
                if (ListTool.isEmpty(hostResults)) {
                    continue;
                }
                for (HostResult hostResult : hostResults) {
                    taskInstanceHostRepository.updateByInstanceIds(instance.getId(), hostResult.getHost(),
                        instance.getExecuteHost().replaceAll("_", ":"), hostResult.getWorkerInstance(), node.getId());
                }
            }
            // 查询任务结果
            if (!Status.Inst.FAILED.name().equals(instance.getStatus())) {
                // 成功，删除所有错误记录
                taskFailRecordRepository.deleteByInstanceId(instanceId);
            } else {
                // 查询所有节点的错误记录
                List<TaskFailRecord> records = taskFailRecordRepository.list(instanceId);
                Map<String, TaskFailRecord> recordMap = new HashMap<>();
                records.stream().forEach(record -> {
                    recordMap.put(record.getNodeId() + "_" + record.getHost(), record);
                });
                // 对节点结果进行处理
                List<TaskFailRecord> delRecords = new ArrayList<>();
                for (TaskNode node : nodes) {
                    if (!Status.Inst.FAILED.name().equals(node.getStatus()) || StringUtils.isEmpty(node.getOutput())) {
                        continue;
                    }
                    // 节点失败
                    GeneralResult<?> result = JSONObject.parseObject(node.getOutput(), GeneralResult.class);
                    if (StringUtils.isEmpty(result.getMessage()) || result.getMessage().contains("因其他节点执行失败")) {
                        continue;
                    }
                    JSONObject input = JSONObject.parseObject(node.getInput());
                    JSONObject target = input.getJSONObject("target");
                    List<WorkerHost> targets = null;
                    if (null != target) {
                        targets = JSONObject.parseArray(target.getString("data"), WorkerHost.class);
                    }
                    targets = null == targets ? new ArrayList<>() : targets;
                    List<WorkerHostResult> list =
                        JSONArray.parseArray(JSONObject.toJSONString(result.getData()), WorkerHostResult.class);
                    List<WorkerHostResult> results = Lists.newArrayList();
                    if (ListTool.isEmpty(list)) {
                        for (WorkerHost host : targets) {
                            results.add(new WorkerHostResult(host.getId(), result.isSuccess(), host.getHost(),
                                host.getUser(), result.getMessage()));
                        }
                    } else {
                        results.addAll(list);
                    }
                    results.stream().forEach(object -> {
                        String host = object.getHost();
                        TaskFailRecord record = recordMap.get(node.getId() + "_" + host);
                        if (!object.isSuccess()) {
                            // 当前设备作业执行失败，更新或者新增错误记录
                            if (null == record) {
                                record = new TaskFailRecord(null, instanceId, instance.getName(), node.getId(),
                                    node.getName(), host, object.getMessage(), new Date(), new Date());
                                taskFailRecordRepository.save(record);
                            } else {
                                record.setMessage(object.getMessage());
                                record.setGmtModify(new Date());
                                taskFailRecordRepository.update(record);
                                recordMap.remove(node.getId() + "_" + host);
                            }
                        }
                    });
                }
                delRecords.addAll(recordMap.values());
                delRecords.forEach(record -> {
                    taskFailRecordRepository.delete(record.getId());
                });
            }
        } catch (Exception e) {
            log.error("task instance analysis fail: {}", e);
        }
    }

}
