package com.bocloud.tse.service;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;
import org.springframework.util.StringUtils;

import com.alibaba.fastjson.JSONObject;
import com.bocloud.tse.entity.CommandInstance;
import com.bocloud.tse.entity.Identification;
import com.bocloud.tse.entity.NodeRelation;
import com.bocloud.tse.entity.Task;
import com.bocloud.tse.entity.TaskHost;
import com.bocloud.tse.entity.TaskInstance;
import com.bocloud.tse.entity.TaskInstanceHost;
import com.bocloud.tse.entity.TaskNode;
import com.bocloud.tse.enums.AuthenMode;
import com.bocloud.tse.enums.Catalog;
import com.bocloud.tse.enums.Category;
import com.bocloud.tse.enums.Status;
import com.bocloud.tse.intf.CommandService;
import com.bocloud.tse.intf.TaskService;
import com.bocloud.tse.model.NodeConfig;
import com.bocloud.tse.model.TaskEdge;
import com.bocloud.tse.model.TaskGraph;
import com.bocloud.tse.repository.CommandInstanceRepository;
import com.bocloud.tse.repository.IdentificationRepository;
import com.bocloud.tse.repository.NodeRelationRepository;
import com.bocloud.tse.repository.TaskHostRepository;
import com.bocloud.tse.repository.TaskInstanceHostRepository;
import com.bocloud.tse.repository.TaskInstanceRepository;
import com.bocloud.tse.repository.TaskNodeRepository;
import com.bocloud.tse.repository.TaskRepository;
import com.bocloud.tse.utils.NodeHelper;
import com.google.common.collect.Maps;
import com.megatron.common.model.GeneralResult;
import com.megatron.common.model.RequestContext;
import com.megatron.common.utils.IDFactory;
import com.megatron.common.utils.ListTool;
import com.megatron.framework.lock.AutoCloseLock;
import com.megatron.framework.lock.LockFactory;

import lombok.extern.slf4j.Slf4j;

/**
 * 任务抽象Service接口实现类
 */
@Slf4j
@Service("taskService")
public class TaskServiceImpl implements TaskService {

    @Autowired
    private LockFactory lockFactory;
    @Autowired
    private TaskRepository taskRepository;
    @Autowired
    private TaskInstanceRepository taskInstanceRepository;
    @Autowired
    private TaskNodeRepository taskNodeRepository;
    @Autowired
    private NodeRelationRepository nodeRelationRepository;
    @Autowired
    private IdentificationRepository identificationRepository;
    @Autowired
    private CommandService commandService;
    @Autowired
    private CommandInstanceRepository commandInstanceRepository;
    @Autowired
    private TaskInstanceHostRepository taskInstanceHostRepository;
    @Autowired
    private TaskHostRepository taskHostRepository;

    @Override
    public GeneralResult<Task> detail(Long id) {
        try {
            Task task = this.taskRepository.query(id);
            if (null == task) {
                return new GeneralResult<>(false, "任务不存在");
            }
            return new GeneralResult<>(true, task, "查询任务详细成功");
        } catch (Exception e) {
            log.error("Query task detail fail:", e);
            return new GeneralResult<>(false, e.getMessage());
        }
    }

    @Override
    public TaskInstance initInstance(Long taskId, String mode, RequestContext context) {
        Task task = null;
        try {
            task = this.taskRepository.query(taskId);
            Assert.notNull(task, "数据不存在！");
            // 创建任务实例
            Long userId = context.getTarget();
            TaskInstance instance = new TaskInstance(userId, userId, userId, task.getName(), Status.Inst.CREATED.name(),
                task.getRemark(), task.getCode(), task.getGraph().toString(), task.getCron(), context.getTenant(),
                context.getCatalog());
            instance.setType("standard");
            instance.setUuid(IDFactory.instance().uuid());
            instance.setName(instance.getName() + "_" + instance.getUuid().substring(0, 8));
            instance.setTaskId(taskId);
            if (!StringUtils.hasText(mode)) {
                mode = "WEB";
            }
            instance.setExecuteMode(mode);
            instance.setTenantId(context.getTenant());
            instance.setIsBackup(false);
            instance.setProps(JSONObject.toJSONString(context));
            instance.setEmailEnabled(null == task.getEmailEnabled() ? false : task.getEmailEnabled());
            this.taskInstanceRepository.save(instance);
            List<Long> commandIds = new ArrayList<>();
            // 解析任务节点
            TaskGraph graph = JSONObject.parseObject(instance.getGraph(), TaskGraph.class);
            List<TaskNode> nodes = graph.getNodes();
            Map<String, Long> nodeMap = Maps.newHashMap();
            boolean isBackup = true;
            List<TaskInstanceHost> instanceHosts = new ArrayList<>();
            for (TaskNode node : nodes) {
                NodeConfig config = JSONObject.parseObject(node.getConfig(), NodeConfig.class);
                if (null == config || config.isIgnore()) {
                    continue;
                }
                String code = node.getModuleType();
                if (!code.contains("start") && !code.contains("stop") && !code.contains("backup")) {
                    isBackup = false;
                }
                if (code.startsWith("task")) {
                    node.setCatalog(Catalog.Node.TASK.name());
                    if (code.contains("maintain")) {
                        node.setCategory(Category.Node.MAINTAIN.name());
                    } else if (code.contains("software")) {
                        node.setCategory(Category.Node.SOFTWARE.name());
                    }
                } else {
                    node.setCatalog(Catalog.Node.CTRL.name());
                    if (code.contains("start")) {
                        node.setCategory(Category.Node.START.name());
                    } else if (code.contains("stop")) {
                        node.setCategory(Category.Node.STOP.name());
                    } else if (code.contains("judge")) {
                        node.setCategory(Category.Node.JUDGE.name());
                    }
                }
                List<TaskInstanceHost> taskInstanceHosts = new ArrayList<>();
                if (!code.contains("ctrl")) {
                    JSONObject input = JSONObject.parseObject(node.getInput());
                    JSONObject target = input.getJSONObject("target");
                    if (null != target) {
                        List<JSONObject> hosts = JSONObject.parseArray(target.getString("data"), JSONObject.class);
                        if (ListTool.isEmpty(hosts)) {
                            hosts = new ArrayList<>();
                        }
                        for (JSONObject host : hosts) {
                            // 记录任务实例和主机关系
                            TaskInstanceHost taskInstanceHost = new TaskInstanceHost();
                            taskInstanceHost.setHost(host.getString("host"));
                            taskInstanceHost.setInstanceId(instance.getId());
                            taskInstanceHosts.add(taskInstanceHost);
                            if (null == host.getLong("id")) {
                                continue;
                            }
                            TaskHost taskHost = taskHostRepository.query(host.getLong("id"));
                            if (null == taskHost) {
                                continue;
                            }
                            if("SaltStack".equalsIgnoreCase(target.getString("executeType"))) {
                                if (null == taskHost.getSaltId()) {
                                    log.warn("服务器("+taskHost.getName()+")缺少saltId");
                                    return null;
                                }
                                host.put("saltId", taskHost.getSaltId());
                                //必须拥有saltMinionId,salt返回的执行结果key是minionId
                                if (null == taskHost.getSaltMinionId()) {
                                    log.warn("服务器("+taskHost.getName()+")缺少saltMinionId");
                                    return null;
                                }
                                host.put("saltMinionId", taskHost.getSaltMinionId());
                            }
                        }
                        if (AuthenMode.BATCH.name().equalsIgnoreCase(target.getString("authenMode"))) {
                            Identification ident = identificationRepository.query(target.getLong("identId"));
                            if (null == ident) {
                                this.taskInstanceRepository.remove(instance.getId(), userId);
                                log.warn("认证账户为空，任务实例化失败");
                                return null;
                            }
                            for (JSONObject host : hosts) {
                                host.put("user", ident.getUsername());
                                host.put("pasd", ident.getPassword());
                            }
                        } else if (AuthenMode.ACCOUNT.name().equalsIgnoreCase(target.getString("authenMode"))) {
                            for (JSONObject host : hosts) {
                                Long identId = host.getLong("identId");
                                if (null == identId) {
                                    log.warn("认证账户为空，任务实例化失败");
                                    return null;
                                }
                                Identification ident = identificationRepository.query(identId);
                                if (null == ident) {
                                    this.taskInstanceRepository.remove(instance.getId(), userId);
                                    log.warn("认证账户为空，任务实例化失败");
                                    return null;
                                }
                                host.put("user", ident.getUsername());
                                host.put("pasd", ident.getPassword());
                            }
                        }
                        target.put("data", hosts);
                        input.put("target", target);
                        node.setInput(input.toString());
                    }
                }
                node.setUuid(IDFactory.instance().uuid());
                node.setName(node.getName() + "_" + node.getUuid().substring(0, 8));
                node.preSave(userId, userId, instance.getId(), Status.Node.READY.name());
                node.setNodeId(config.getId());
                this.taskNodeRepository.save(node);
                if (!taskInstanceHosts.isEmpty()) {
                    taskInstanceHosts.forEach(taskInstanceHost -> {
                        taskInstanceHost.setNodeId(node.getId());
                        taskInstanceHost.setNodeName(node.getName());
                    });
                    instanceHosts.addAll(taskInstanceHosts);
                }
                nodeMap.put(node.getNodeId(), node.getId());

                if (node.getModuleType().contains("maintain.script")) {
                    List<Long> commands = commandService.getCommands(node.getInput());
                    for (Long command : commands) {
                        if (!commandIds.contains(command)) {
                            commandIds.add(command);
                        }
                    }
                }
            }
            if (isBackup) {
                instance.setIsBackup(isBackup);
                taskInstanceRepository.update(instance);
            }
            // 解析任务节点关系
            List<TaskEdge> edges = graph.getEdges();
            Map<String, String> ingoreNodes = NodeHelper.getIngoreNodes(nodes);
            if (!ListTool.isEmpty(edges)) {
                List<NodeRelation> relations = new ArrayList<>();
                for (TaskEdge edge : edges) {
                    if (edge.isIgnore()) {
                        continue;
                    }
                    String source = edge.getSource();
                    source = ingoreNodes.containsKey(source) ? ingoreNodes.get(source) : source;
                    String target = edge.getTarget();
                    target = ingoreNodes.containsKey(target) ? ingoreNodes.get(target) : target;
                    NodeRelation relation = new NodeRelation(nodeMap.get(source), nodeMap.get(target), instance.getId(),
                        Status.Node.READY.name(), edge.isDicision());
                    relations.add(relation);
                }
                if (!relations.isEmpty()) {
                    this.nodeRelationRepository.batchSave(relations);
                }
            }
            // 处理命令与任务实例关系
            List<CommandInstance> commandInstances = new ArrayList<>();
            for (Long commandId : commandIds) {
                commandInstances.add(new CommandInstance(commandId, instance.getId()));
            }
            if (!commandInstances.isEmpty()) {
                commandInstanceRepository.batchSave(commandInstances, 100);
            }
            // 处理任务实例链路数据
            if (!instanceHosts.isEmpty()) {
                taskInstanceHostRepository.batchSave(instanceHosts, 100);
            }
            taskInstanceRepository.updateTime(instance.getId());
            instance.setNodes(nodes);
            instance.setContext(context);
            return instance;
        } catch (Exception e) {
            log.error("initInstance [{}] exception !", taskId, e);
            return null;
        }
    }

    @Override
    public GeneralResult<Void> updateStatus(Long id, String status, Long userId) {
        String path = Task.class.getSimpleName() + "_" + id;
        try (AutoCloseLock lock = lockFactory.getACLock(path)) {
            Assert.isTrue(lock.acquire(10, TimeUnit.SECONDS), "请求超时");
            taskRepository.updateStatus(id, status, userId);
            return new GeneralResult<>(true, "修改任务状态成功");
        } catch (Exception e) {
            log.error("update task status fail:", e);
            return new GeneralResult<>(false, "修改任务状态失败");
        }
    }

    @Override
    public GeneralResult<Task> get(String uuid) {
        try {
            Task task = this.taskRepository.get(uuid);
            Assert.notNull(task, "数据不存在！");
            return new GeneralResult<>(true, task, "查询任务详细成功");
        } catch (Exception e) {
            log.error("Query task detail fail:", e);
            return new GeneralResult<>(false, e.getMessage());
        }
    }

    @Override
    public GeneralResult<Void> updateExecute(Long id, Date date) {
        String path = Task.class.getSimpleName() + "_" + id;
        try (AutoCloseLock lock = lockFactory.getACLock(path)) {
            Assert.isTrue(lock.acquire(10, TimeUnit.SECONDS), "请求超时");
            taskRepository.updateExecute(id, date);
            return new GeneralResult<>(true, "修改任务执行时间成功");
        } catch (Exception e) {
            log.error("Update task execute time fail:", e);
            return new GeneralResult<>(false, "修改任务执行时间失败");
        }
    }

    @Override
    public GeneralResult<List<Task>> listSchedule() {
        try {
            List<Task> tasks = taskRepository.listSchedule();
            return new GeneralResult<>(true, tasks, "查询定时作业成功");
        } catch (Exception e) {
            log.error("List task fail:", e);
            return new GeneralResult<>(false, "查询定时作业失败");
        }
    }

}
